/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.attached_item;

import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.function.Predicate;

import jakarta.validation.ValidationException;

import org.apache.commons.lang3.StringUtils;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.context.event.EventListener;
import org.springframework.stereotype.Component;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandCreatedEvent;
import de.ozgcloud.command.CommandFailedEvent;
import de.ozgcloud.command.RevokeCommandEvent;
import de.ozgcloud.vorgang.command.CommandService;
import de.ozgcloud.vorgang.command.Order;
import de.ozgcloud.vorgang.command.PersistedCommand;
import de.ozgcloud.vorgang.vorgang.VorgangDeletedEvent;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Component
@RequiredArgsConstructor
@Log4j2
class VorgangAttachedItemEventListener {

	private static final String IS_CREATE_ITEM_ORDER_CONDITION = "{T(de.ozgcloud.vorgang.attached_item.VorgangAttachedItemEventListener).IS_CREATE_ITEM_ORDER.test(event.getSource())}";
	private static final String IS_UPDATE_ITEM_ORDER_CONDITION = "{T(de.ozgcloud.vorgang.attached_item.VorgangAttachedItemEventListener).IS_UPDATE_ITEM_ORDER.test(event.getSource())}";
	private static final String IS_PATCH_ITEM_ORDER_CONDITION = "{T(de.ozgcloud.vorgang.attached_item.VorgangAttachedItemEventListener).IS_PATCH_ITEM_ORDER.test(event.getSource())}";
	private static final String IS_DELETE_ITEM_ORDER_CONDITION = "{T(de.ozgcloud.vorgang.attached_item.VorgangAttachedItemEventListener).IS_DELETE_ITEM_ORDER.test(event.getSource())}";

	public static final Predicate<Command> IS_CREATE_ITEM_ORDER = command -> Order.CREATE_ATTACHED_ITEM.isMeant(command.getOrder());
	public static final Predicate<Command> IS_UPDATE_ITEM_ORDER = command -> Order.UPDATE_ATTACHED_ITEM.isMeant(command.getOrder());
	public static final Predicate<Command> IS_PATCH_ITEM_ORDER = command -> Order.PATCH_ATTACHED_ITEM.isMeant(command.getOrder());
	public static final Predicate<Command> IS_DELETE_ITEM_ORDER = command -> Order.DELETE_ATTACHED_ITEM.isMeant(command.getOrder());

	private final VorgangAttachedItemService service;
	private final VorgangAttachedItemMapper mapper;

	private final ApplicationEventPublisher publisher;

	private final CommandService commandService;

	@EventListener(condition = IS_CREATE_ITEM_ORDER_CONDITION)
	public void createItem(CommandCreatedEvent event) {
		var command = event.getSource();
		validateCreateCommand(command);

		var item = mapper.fill(command.getBodyObject()).toBuilder().id(null).version(0).build();
		service.create(command, item);
	}

	void validateCreateCommand(Command command) {
		if (!StringUtils.equals(command.getRelationId(), command.getVorgangId())) {
			throw new ValidationException("Command invalid creation command: vorgangId and relationId must be equals");
		}
	}

	@EventListener(condition = IS_UPDATE_ITEM_ORDER_CONDITION)
	public void updateItem(CommandCreatedEvent event) {
		var command = event.getSource();
		var filledItem = mapper.fill(command.getBodyObject()).toBuilder().id(command.getRelationId()).build();

		setCommandPreviousState(command.getId(), getVorgangAttachedItem(command.getRelationId()).getItem());

		doUpdate(command, filledItem);
	}

	private void doUpdate(Command command, VorgangAttachedItem filledItem) {
		if (isVersionSet(command)) {
			service.update(command.getId(), filledItem.toBuilder().version(command.getRelationVersion()).build());
		} else {
			service.forceUpdate(command.getId(), filledItem.getId(), filledItem.getItem());
		}
	}

	private boolean isVersionSet(Command command) {
		return Objects.nonNull(command.getRelationVersion()) && command.getRelationVersion() > 0;
	}

	VorgangAttachedItem getVorgangAttachedItem(String id) {
		return service.getById(id);
	}

	@EventListener(condition = IS_PATCH_ITEM_ORDER_CONDITION)
	public void patchItem(CommandCreatedEvent event) {
		var command = event.getSource();
		var item = mapper.fill(command.getBodyObject());

		setCommandPreviousState(command.getId(), getExistingItemMapEntries(command.getRelationId(), item.getItem().keySet()));

		service.patch(command, item.getItem());
	}

	Map<String, Object> getExistingItemMapEntries(String id, Set<String> keys) {
		return getVorgangAttachedItem(id).getItem().entrySet().stream().filter(entry -> keys.contains(entry.getKey()))
				.collect(HashMap::new, (itemMap, entry) -> itemMap.put(entry.getKey(), entry.getValue()), HashMap::putAll);
	}

	void setCommandPreviousState(String commandId, Map<String, Object> itemMap) {
		commandService.setPreviousState(commandId, itemMap);
	}

	@EventListener(condition = IS_DELETE_ITEM_ORDER_CONDITION)
	public void deleteItem(CommandCreatedEvent event) {
		var command = event.getSource();
		try {
			service.markAsDeleteByIdAndVersion(command.getRelationId(), command.getRelationVersion());
			publisher.publishEvent(new VorgangAttachedItemDeletedEvent(command));
		} catch (RuntimeException e) {
			handleException(e, command.getId());
		}
	}

	void handleException(Exception e, String commandId) {
		LOG.error("Command failed with Exception", e);
		publisher.publishEvent(new CommandFailedEvent(commandId, "Command Failed: " + e.getMessage()));
	}

	@EventListener
	public void onVorgangDeleted(VorgangDeletedEvent event) {
		try {
			service.deleteByVorgangId(event.getCommand().getVorgangId());
		} catch (RuntimeException e) {
			LOG.error("Error on deleting Attached Items.", e);
		}
	}

	@EventListener(condition = IS_CREATE_ITEM_ORDER_CONDITION)
	public void onRevokeCreateItem(RevokeCommandEvent event) {
		service.revokeCreate(event.getSource());
	}

	@EventListener(condition = IS_DELETE_ITEM_ORDER_CONDITION)
	public void onRevokeDeleteItem(RevokeCommandEvent event) {
		service.revokeDelete(event.getSource());
	}

	@EventListener(condition = IS_UPDATE_ITEM_ORDER_CONDITION)
	public void onRevokeUpdateItem(RevokeCommandEvent event) {
		var command = event.getSource();
		var persistedCommand = getPersistedCommand(command.getId());
		service.revokeUpdate(command, persistedCommand.getPreviousState());
	}

	@EventListener(condition = IS_PATCH_ITEM_ORDER_CONDITION)
	public void onRevokePatchItem(RevokeCommandEvent event) {
		var command = event.getSource();
		var persistedCommand = getPersistedCommand(command.getId());
		service.revokePatch(command, persistedCommand.getPreviousState());
	}

	private PersistedCommand getPersistedCommand(String commandId) {
		return (PersistedCommand) commandService.getById(commandId);
	}
}
