/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.clientattribute;

import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Stream;

import org.apache.commons.lang3.StringUtils;
import org.mapstruct.Mapper;
import org.mapstruct.Mapping;
import org.mapstruct.MappingConstants;
import org.mapstruct.ValueMapping;

import de.ozgcloud.vorgang.grpc.clientAttribute.GrpcAccessPermission;
import de.ozgcloud.vorgang.grpc.clientAttribute.GrpcClientAttribute;
import de.ozgcloud.vorgang.grpc.clientAttribute.GrpcClientAttributeValue;
import de.ozgcloud.vorgang.grpc.clientAttribute.GrpcClientAttributeValue.ValueCase;

@Mapper
public abstract class ClientAttributeMapper {

	private static final String UNREACHABLE_ERROR_MSG = "Should not be able to reach missing value state";

	@ValueMapping(source = "UNRECOGNIZED", target = MappingConstants.NULL)
	@Mapping(source = "value", target = "stringValue")
	@Mapping(source = "value", target = "intValue")
	@Mapping(source = "value", target = "boolValue")
	@Mapping(source = "value", target = "doubleValue")
	public abstract ClientAttribute mapFromGrpc(GrpcClientAttribute attribute);

	Optional<String> mapStringValue(GrpcClientAttributeValue value) {
		return Optional.ofNullable(value)
				.filter(val -> val.getValueCase() == ValueCase.STRINGVALUE)
				.map(GrpcClientAttributeValue::getStringValue)
				.map(StringUtils::trimToNull).filter(Objects::nonNull);
	}

	Optional<Long> mapLongValue(GrpcClientAttributeValue value) {
		return Optional.ofNullable(value)
				.filter(val -> val.getValueCase() == ValueCase.INTVALUE)
				.map(GrpcClientAttributeValue::getIntValue);
	}

	Optional<Boolean> mapBoolValue(GrpcClientAttributeValue value) {
		return Optional.ofNullable(value)
				.filter(val -> val.getValueCase() == ValueCase.BOOLVALUE)
				.map(GrpcClientAttributeValue::getBoolValue);
	}

	Optional<Double> mapDecimalValue(GrpcClientAttributeValue value) {
		return Optional.ofNullable(value)
				.filter(val -> val.getValueCase() == ValueCase.DOUBLEVALUE)
				.map(GrpcClientAttributeValue::getDoubleValue);
	}

	public List<GrpcClientAttribute.Builder> mapClientAttributesToBuilder(ClientAttributesMap clientAttributesMap) {
		return mapClientAttributes(clientAttributesMap).stream().map(GrpcClientAttribute::toBuilder).toList();
	}

	public List<GrpcClientAttribute> mapClientAttributes(ClientAttributesMap clientAttributesMap) {
		return clientAttributesMap.entrySet().stream()
				.filter(entry -> !entry.getValue().isEmpty())
				.flatMap(entry -> buildGrpcClientAttribute(entry.getKey(), entry.getValue()))
				.toList();
	}

	Stream<GrpcClientAttribute> buildGrpcClientAttribute(String clientName, ClientAttributeMap clientAttributeMap) {
		return clientAttributeMap.entrySet().stream()
				.map(e -> GrpcClientAttribute.newBuilder()
						.setClientName(clientName)
						.setAttributeName(e.getKey())
						.setAccess(GrpcAccessPermission.valueOf(e.getValue().getAccess().toString()))
						.setValue(buildValue(e.getValue())))
				.map(GrpcClientAttribute.Builder::build);

	}

	GrpcClientAttributeValue buildValue(ClientAttribute attribute) {
		if (attribute.getIntValue().isPresent())
			return attribute.getIntValue().map(val -> GrpcClientAttributeValue.newBuilder().setIntValue(val).build())
					.orElseThrow(() -> new IllegalStateException(UNREACHABLE_ERROR_MSG));
		if (attribute.getStringValue().isPresent())
			return attribute.getStringValue().map(val -> GrpcClientAttributeValue.newBuilder().setStringValue(val).build())
					.orElseThrow(() -> new IllegalStateException(UNREACHABLE_ERROR_MSG));
		if (attribute.getDoubleValue().isPresent())
			return attribute.getDoubleValue().map(val -> GrpcClientAttributeValue.newBuilder().setDoubleValue(val).build())
					.orElseThrow(() -> new IllegalStateException(UNREACHABLE_ERROR_MSG));
		if (attribute.getBoolValue().isPresent())
			return attribute.getBoolValue().map(val -> GrpcClientAttributeValue.newBuilder().setBoolValue(val).build())
					.orElseThrow(() -> new IllegalStateException(UNREACHABLE_ERROR_MSG));

		return GrpcClientAttributeValue.newBuilder().build();
	}
}
