/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.errorhandling;

import java.util.UUID;

import org.springframework.security.access.AccessDeniedException;

import de.ozgcloud.common.errorhandling.ExceptionUtil;
import de.ozgcloud.common.errorhandling.TechnicalException;
import io.grpc.Metadata;
import io.grpc.Metadata.Key;
import io.grpc.Status;
import io.grpc.StatusException;
import io.grpc.StatusRuntimeException;
import lombok.extern.log4j.Log4j2;
import net.devh.boot.grpc.server.advice.GrpcAdvice;
import net.devh.boot.grpc.server.advice.GrpcExceptionHandler;

@Log4j2
@GrpcAdvice
public class ExceptionHandler {

	private static final String ERROR_MESSAGE = "Grpc internal server error";

	static final String KEY_ERROR_CODE = "ERROR_CODE";
	static final String KEY_EXCEPTION_ID = "EXCEPTION_ID";

	@GrpcExceptionHandler
	public StatusException handleNotFoundException(NotFoundException e) {
		LOG.warn(ERROR_MESSAGE, e);
		return createStatusException(buildNotFoundStatus(e), buildMetadata(e));
	}

	private Status buildNotFoundStatus(NotFoundException e) {
		return Status.NOT_FOUND.withDescription(e.getMessage()).withCause(e.getCause());
	}

	@GrpcExceptionHandler
	public StatusException handleFunctionalException(FunctionalException e) {
		LOG.error(ERROR_MESSAGE, e);
		return createStatusException(buildInternalStatus(e), buildMetadata(e));
	}

	private Metadata buildMetadata(FunctionalException e) {
		var metadata = buildMetadataFromParameter(e);
		addErrorCode(metadata, e);
		addExceptionId(metadata, e.getExceptionId());
		return metadata;
	}

	private Metadata buildMetadataFromParameter(FunctionalException e) {
		var metadata = new Metadata();
		e.getParameters().forEach((key, value) -> metadata.put(createKeyOf(key), value));
		return metadata;
	}

	private void addErrorCode(Metadata metadata, FunctionalException e) {
		metadata.put(createKeyOf(KEY_ERROR_CODE), e.getErrorCode());
	}

	@GrpcExceptionHandler
	public StatusException handleTechnicalException(TechnicalException e) {
		LOG.error(ERROR_MESSAGE, e);
		return createStatusException(buildInternalStatus(e), buildMetadata(e.getExceptionId()));
	}

	private Status buildInternalStatus(RuntimeException e) {
		return Status.INTERNAL.withDescription(e.getMessage()).withCause(e.getCause());
	}

	@GrpcExceptionHandler
	public StatusRuntimeException passThroughStatusRuntimeException(StatusRuntimeException e) {
		return e;
	}

	@GrpcExceptionHandler
	public StatusException handleRuntimeException(RuntimeException e) {
		var exceptionId = createExceptionId();
		var messageWithExceptionId = ExceptionUtil.formatMessageWithExceptionId(e.getMessage(), exceptionId);
		LOG.error("Grpc internal server error: {}", messageWithExceptionId, e);

		return createStatusException(buildInternalStatus(e, messageWithExceptionId), buildMetadata(exceptionId));
	}

	private Status buildInternalStatus(RuntimeException e, String message) {
		return Status.INTERNAL.withDescription(message).withCause(e.getCause());
	}

	@GrpcExceptionHandler
	public StatusException handleNotFoundException(de.ozgcloud.apilib.common.errorhandling.NotFoundException e) {
		var exceptionId = createExceptionId();
		var messageWithExceptionId = ExceptionUtil.formatMessageWithExceptionId(e.getMessage(), exceptionId);
		LOG.warn("Grpc internal server error: {}", messageWithExceptionId, e);

		var status = Status.NOT_FOUND.withDescription(messageWithExceptionId).withCause(e.getCause());
		return createStatusException(status, buildMetadata(exceptionId));
	}

	private Metadata buildMetadata(String exceptionId) {
		var metadata = new Metadata();
		addExceptionId(metadata, exceptionId);
		return metadata;
	}

	String createExceptionId() {
		return UUID.randomUUID().toString();
	}

	private void addExceptionId(Metadata metadata, String value) {
		metadata.put(createKeyOf(KEY_EXCEPTION_ID), value);
	}

	private Key<String> createKeyOf(String key) {
		return Key.of(key, Metadata.ASCII_STRING_MARSHALLER);
	}

	private StatusException createStatusException(Status status, Metadata metadata) {
		return new StatusException(status, metadata);
	}

	@GrpcExceptionHandler
	public StatusException handleAccessDeniedException(AccessDeniedException e) {
		var exceptionId = createExceptionId();
		var messageWithExceptionId = ExceptionUtil.formatMessageWithExceptionId(e.getMessage(), exceptionId);
		LOG.error("Grpc permission denied error: {}", messageWithExceptionId, e);

		return createStatusException(buildPermissionDeniedStatus(e, messageWithExceptionId), buildMetadata(exceptionId));
	}

	private Status buildPermissionDeniedStatus(AccessDeniedException e, String message) {
		return Status.PERMISSION_DENIED.withDescription(message).withCause(e.getCause());
	}

	@GrpcExceptionHandler
	public StatusException handleSearchServiceUnavailableException(SearchServiceUnavailableException e) {
		LOG.error(ERROR_MESSAGE, e);
		return createStatusException(buildUnavailableStatus(e), buildMetadata(e.getExceptionId()));
	}

	private Status buildUnavailableStatus(SearchServiceUnavailableException e) {
		return Status.UNAVAILABLE.withDescription(e.getMessage()).withCause(e.getCause());
	}
}