/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.operator;

import java.util.Map;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;

import de.ozgcloud.vorgang.statistic.InvalidFieldPathException;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import lombok.AccessLevel;
import lombok.NoArgsConstructor;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;

@NoArgsConstructor(access = AccessLevel.PRIVATE)
@Log4j2
class FieldPathProcessor {

	static final String PATH_ENTITY_VORGANG = "Vorgang";
	static final String PATH_ENTITY_CLIENT_ATTRIBUTE = "ClientAttribute";

	static final String CRITERIA_PATH_VORGANG = "";
	static final String CRITERIA_PATH_CLIENT_ATTRIBUTE = Vorgang.FIELD_CLIENT_ATTRIBUTES;

	private static final Map<String, String> ENTITY_TO_PATH = Map.of(
			PATH_ENTITY_VORGANG, CRITERIA_PATH_VORGANG,
			PATH_ENTITY_CLIENT_ATTRIBUTE, CRITERIA_PATH_CLIENT_ATTRIBUTE);

	private static final Map<String, String> ENTITY_TO_PATHSUFFIX = Map.of(
			PATH_ENTITY_CLIENT_ATTRIBUTE, "value");

	private static final String CRITERIA_PATH_TEMPLATE = "%s.%s";

	public static String processToCriteriaPath(String fieldPath) {
		// TODO do path validation earlier
		validate(fieldPath);

		if (StringUtils.isBlank(getCriteriaEntity(fieldPath))) {
			return getPath(fieldPath);
		}

		return CRITERIA_PATH_TEMPLATE.formatted(getCriteriaEntity(fieldPath), getPath(fieldPath));
	}

	private static void validate(String path) {
		if (!path.contains(".")) {
			throw new InvalidFieldPathException();
		}
	}

	static String getCriteriaEntity(@NonNull String path) {
		return Optional.ofNullable(ENTITY_TO_PATH.get(getEntity(path)))
				.orElseThrow(() -> logAndBuildUnsupportedEntityException(getEntity(path)));
	}

	private static RuntimeException logAndBuildUnsupportedEntityException(String entity) {
		LOG.error("Given Entity '{}' is unsupported.", entity);
		return new UnsupportedEntityException();
	}

	public static String getPath(String path) {
		var criteriaPath = StringUtils.substringAfter(path, ".");

		return getPathSuffix(path).map(suffix -> CRITERIA_PATH_TEMPLATE.formatted(criteriaPath, suffix))
				.orElse(criteriaPath);
	}

	static Optional<String> getPathSuffix(String path) {
		return Optional.ofNullable(ENTITY_TO_PATHSUFFIX.get(getEntity(path)));
	}

	private static String getEntity(String path) {
		return StringUtils.substringBefore(path, ".");
	}
}
