/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.search;

import static de.ozgcloud.vorgang.common.search.IndexedVorgang.*;
import static java.util.Objects.*;
import static org.apache.commons.collections4.MapUtils.*;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.elasticsearch.client.elc.NativeQueryBuilder;
import org.springframework.data.elasticsearch.core.ElasticsearchOperations;
import org.springframework.data.elasticsearch.core.RefreshPolicy;
import org.springframework.data.elasticsearch.core.SearchHit;
import org.springframework.data.elasticsearch.core.SearchHits;
import org.springframework.data.elasticsearch.core.document.Document;
import org.springframework.data.elasticsearch.core.query.ScriptType;
import org.springframework.data.elasticsearch.core.query.UpdateQuery;
import org.springframework.data.support.PageableExecutionUtils;
import org.springframework.stereotype.Repository;

import co.elastic.clients.elasticsearch._types.FieldValue;
import co.elastic.clients.elasticsearch._types.query_dsl.BoolQuery;
import co.elastic.clients.elasticsearch._types.query_dsl.Query;
import co.elastic.clients.elasticsearch._types.query_dsl.QueryBuilders;
import de.ozgcloud.vorgang.vorgang.FilterCriteria;
import de.ozgcloud.vorgang.vorgang.FindVorgangRequest;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangHeader;

@ConditionalOnProperty(prefix = "ozgcloud.elasticsearch", name = "address")
@Repository
class SearchVorgangCustomRepositoryImpl implements SearchVorgangCustomRepository {

	private static final Pattern SPECIAL_CHARACTERS_PATTERN = Pattern.compile("[+\\-=]|(&&)|(\\|\\|)|[!(){}\\[\\]^\"~*?:\\\\/]");

	private static final String SCRIPT_REMOVE_FIELD_TEMPLATE = "ctx._source.remove('%s')";

	private static final String JOIN_AND = " AND ";
	private static final String KEYWORD = ".keyword";

	private static final String STATUS = FIELD_STATUS + KEYWORD;
	private static final String ASSIGNED_TO = FIELD_ASSIGNED_TO + KEYWORD;

	private static final float DEFAULT_BOOST = 1f;
	private static final float HALF_BOOST = 0.5f;
	private static final float DOUBLE_BOOST = 2f;

	private static final List<String> FIELD_LIST = List.of(
			FIELD_NAME + "^" + HALF_BOOST,
			FIELD_VORGANG_NUMMER + "^" + DOUBLE_BOOST,
			FIELD_AKTENZEICHEN + "^" + DOUBLE_BOOST,
			FIELD_ANTRAGSTELLER_NAME + "^" + DEFAULT_BOOST,
			FIELD_ANTRAGSTELLTER_VORNAME + "^" + DEFAULT_BOOST);

	@Autowired
	private ElasticsearchOperations elasticsearchOperations;

	@Autowired
	private IndexedVorgangMapper mapper;

	@Override
	public Page<VorgangHeader> searchBy(FindVorgangRequest request) {
		var pageable = buildPageable(request);
		var searchQuery = buildQueryBuilder(request).withPageable(pageable).build();

		var result = elasticsearchOperations.search(searchQuery, IndexedVorgang.class);

		return PageableExecutionUtils.getPage(mapResult(result), pageable, result::getTotalHits);
	}

	PageRequest buildPageable(FindVorgangRequest request) {
		return PageRequest.of(request.getOffset() / request.getLimit(), request.getLimit());
	}

	NativeQueryBuilder buildQueryBuilder(FindVorgangRequest request) {
		var queryBuilder = new NativeQueryBuilder();

		queryBuilder.withQuery(buildSearchByQueryBuilder(request));
		queryBuilder.withFilter(buildFilterByQueryBuilder(request.getFilterBy()));

		return queryBuilder;
	}

	Query buildSearchByQueryBuilder(FindVorgangRequest request) {
		var searchBy = Arrays.stream(request.getSearchBy().strip().split(StringUtils.SPACE)).map(this::adjustSearchTerm)
				.collect(Collectors.joining(JOIN_AND));
		return QueryBuilders.queryString(b -> b.query(searchBy).fields(FIELD_LIST));
	}

	String adjustSearchTerm(String searchTerm) {
		var resultBuilder = new StringBuilder("(*");
		var lastEnd = 0;
		var matcher = SPECIAL_CHARACTERS_PATTERN.matcher(searchTerm);
		while (matcher.find()) {
			var start = matcher.start();
			var end = matcher.end();
			resultBuilder.append(searchTerm, lastEnd, start).append("\\").append(searchTerm, start, end);
			lastEnd = end;
		}
		if (lastEnd < searchTerm.length()) {
			resultBuilder.append(searchTerm, lastEnd, searchTerm.length());
		}
		return replaceNotAllowedCharacters(resultBuilder.append("*)").toString());
	}

	String replaceNotAllowedCharacters(String searchTerm) {
		return searchTerm.replaceAll("[><]", "?");
	}

	private Query buildFilterByQueryBuilder(FilterCriteria filterCriteria) {
		var queryBuilder = QueryBuilders.bool();

		if (filterCriteria.isFilterByOrganisationseinheitenId()) {
			queryBuilder.must(buildTermQuery(FIELD_ORGANISATIONSEINHEITEN_ID, filterCriteria.getOrganisationseinheitIds().stream()));
		}
		if (filterCriteria.isFilterByAssignedTo()) {
			addAssignedToFilter(queryBuilder, filterCriteria);
		}
		if (CollectionUtils.isNotEmpty(filterCriteria.getStatus())) {
			queryBuilder.must(buildStatusQueryBuilder(filterCriteria));
		}
		return queryBuilder.build()._toQuery();
	}

	private void addAssignedToFilter(BoolQuery.Builder query, FilterCriteria filterCriteria) {
		if (StringUtils.isNotEmpty(filterCriteria.getAssignedTo())) {
			query.must(buildTermQuery(ASSIGNED_TO, filterCriteria.getAssignedTo()));
		} else {
			query.mustNot(QueryBuilders.exists(e -> e.field(ASSIGNED_TO)));
		}
	}

	private Query buildStatusQueryBuilder(FilterCriteria filterCriteria) {
		return buildTermQuery(STATUS, filterCriteria.getStatus().stream().map(Vorgang.Status::name));
	}

	private Query buildTermQuery(String field, String value) {
		return QueryBuilders.term(ts -> ts.field(field).value(FieldValue.of(value)));
	}

	private Query buildTermQuery(String field, Stream<String> valueStream) {
		return QueryBuilders.terms(ts -> ts.field(field).terms(t -> t.value(valueStream.map(FieldValue::of).toList())));
	}

	private List<VorgangHeader> mapResult(SearchHits<IndexedVorgang> searchHits) {
		return searchHits.get().map(SearchHit::getContent).map(mapper::toVorgangHeader).toList();
	}

	public void update(String vorgangId, Map<String, Object> patch) {
		var fieldsToRemove = getFieldsToRemove(patch);
		if (CollectionUtils.isNotEmpty(fieldsToRemove)) {
			executeUpdate(buildUpdateQueryWithScript(vorgangId, buildRemoveFieldScript(fieldsToRemove)));
		}
		var fieldsToUpdate = getFieldsToUpdate(patch);
		if (isNotEmpty(fieldsToUpdate)) {
			executeUpdate(buildUpdateQueryWithDocument(vorgangId, fieldsToUpdate));
		}
	}

	Set<String> getFieldsToRemove(Map<String, Object> patch) {
		return patch.entrySet().stream().filter(e -> isNull(e.getValue())).map(Map.Entry::getKey).collect(Collectors.toSet());
	}

	String buildRemoveFieldScript(Collection<String> fields) {
		return fields.stream().map(SCRIPT_REMOVE_FIELD_TEMPLATE::formatted).collect(Collectors.joining(";"));
	}

	UpdateQuery buildUpdateQueryWithScript(String vorgangId, String script) {
		return createUpdateBuilder(vorgangId).withScriptType(ScriptType.INLINE).withScript(script).build();
	}

	Map<String, Object> getFieldsToUpdate(Map<String, Object> patch) {
		return patch.entrySet().stream().filter(e -> nonNull(e.getValue())).collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue));
	}

	UpdateQuery buildUpdateQueryWithDocument(String vorgangId, Map<String, Object> fieldMap) {
		return createUpdateBuilder(vorgangId).withDocument(Document.from(fieldMap)).build();
	}

	UpdateQuery.Builder createUpdateBuilder(String vorgangId) {
		return UpdateQuery.builder(vorgangId).withRefreshPolicy(RefreshPolicy.IMMEDIATE);
	}

	void executeUpdate(UpdateQuery updateQuery) {
		elasticsearchOperations.update(updateQuery, elasticsearchOperations.indexOps(IndexedVorgang.class).getIndexCoordinates());
	}
}
