/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.security;

import java.util.List;
import java.util.Set;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Stream;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.access.AccessDeniedException;
import org.springframework.stereotype.Service;

import de.ozgcloud.command.Command;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItem;
import de.ozgcloud.vorgang.callcontext.CallContextUser;
import de.ozgcloud.vorgang.callcontext.CurrentUserService;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangService;

@Service
public class PolicyService {

	private static final Predicate<Boolean> NOT_PERMITTED = permission -> !permission;

	@Autowired
	private CurrentUserService userService;
	@Autowired
	private VorgangService vorgangService;
	@Autowired
	private PolicyRepository repository;

	public void checkPermission(String vorgangId) {
		checkPermission(Set.of(vorgangId));
	}

	public void checkPermission(Set<String> vorgangIds) {
		evaluatePermissions(vorgangIds.stream().map(this::hasOrganisationEinheitenId), Vorgang.class, String.join(",", vorgangIds));
	}

	public boolean hasOrganisationEinheitenId(String vorgangId) {
		return hasUserPermission(user -> vorgangService.exists(vorgangId, user.getOrganisatorischeEinheitenIds()));
	}

	public void checkPermissionByCommand(String commandId) {
		evaluatePermissions(List.of(hasOrganisationEinheitenIdByCommand(commandId)).stream(), Command.class, commandId);
	}

	public boolean hasOrganisationEinheitenIdByCommand(String commandId) {
		return hasUserPermission(user -> repository.existsByCommandId(commandId, user.getOrganisatorischeEinheitenIds()));
	}

	public void checkPermissionByFile(String fileId) {
		evaluatePermissions(List.of(hasOrganisationEinheitenIdByFileId(fileId)).stream(), "File Content ", fileId);
	}

	public boolean hasOrganisationEinheitenIdByFileId(String fileId) {
		return hasUserPermission(user -> repository.existsByFileId(fileId, user.getOrganisatorischeEinheitenIds()));
	}

	public void checkPermissionByVorgangAttachedItem(String vorgangAttachedItemId) {
		evaluatePermissions(Stream.of(hasOrganisationEinheitenIdByVorgangAttachedItem(vorgangAttachedItemId)), VorgangAttachedItem.class,
				vorgangAttachedItemId);
	}

	public boolean hasOrganisationEinheitenIdByVorgangAttachedItem(String vorgangAttachedItemId) {
		return hasUserPermission(user -> repository.existsByVorgangAttachedItem(vorgangAttachedItemId, user.getOrganisatorischeEinheitenIds()));
	}

	boolean hasUserPermission(Function<CallContextUser, Boolean> additionalPermissionCheck) {
		return userService.findUser()
				.filter(CallContextUser::isOrganisationEinheitenIdCheckNecessary)
				.map(additionalPermissionCheck)
				.orElse(true);
	}

	void evaluatePermissions(Stream<Boolean> permissions, Class<?> resource, String id) {
		evaluatePermissions(permissions, resource.getName(), id);
	}

	void evaluatePermissions(Stream<Boolean> permissions, String resourceString, String id) {
		var permissionDenied = permissions.anyMatch(NOT_PERMITTED);

		if (permissionDenied) {
			throw new AccessDeniedException(String.format("Access denied for %s with id %s", resourceString, id));
		}
	}
}