/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import java.io.BufferedInputStream;
import java.io.InputStream;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.stream.Collectors;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;

import com.google.protobuf.ByteString;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.callcontext.CallContextUser;
import de.ozgcloud.vorgang.callcontext.CurrentUserService;
import de.ozgcloud.vorgang.common.errorhandling.NotFoundException;
import de.ozgcloud.vorgang.common.security.PolicyService;
import de.ozgcloud.vorgang.grpc.binaryFile.BinaryFileServiceGrpc.BinaryFileServiceImplBase;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcBinaryFilesRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcFindFilesResponse;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcGetBinaryFileDataRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcGetBinaryFileDataResponse;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUpdateBinaryFileRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUpdateBinaryFileResponse;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileResponse;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.StreamObserver;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;
import net.devh.boot.grpc.server.service.GrpcService;

@Log4j2
@GrpcService
@RequiredArgsConstructor
public class GrpcBinaryFileService extends BinaryFileServiceImplBase {

	private final FileService service;
	private final GrpcOzgFileMapper fileMapper;
	private final PolicyService policyService;
	private final CurrentUserService currentUserService;

	@Override
	public void findBinaryFilesMetaData(GrpcBinaryFilesRequest request, StreamObserver<GrpcFindFilesResponse> responseObserver) {
		var found = service.findOzgCloudFiles(mapFileIds(request.getFileIdList())).toList();

		policyService.checkPermission(found.stream().map(OzgCloudFile::getVorgangId).collect(Collectors.toSet()));

		responseObserver.onNext(buildResponseForOzgFiles(found));
		responseObserver.onCompleted();
	}

	GrpcFindFilesResponse buildResponseForOzgFiles(List<OzgCloudFile> files) {
		var builder = GrpcFindFilesResponse.newBuilder();

		files.stream().map(fileMapper::map).forEach(builder::addFile);
		return builder.build();
	}

	private List<FileId> mapFileIds(List<String> fileIdList) {
		return fileIdList.stream().map(FileId::from).toList();
	}

	@Override
	public StreamObserver<GrpcUploadBinaryFileRequest> uploadBinaryFileAsStream(StreamObserver<GrpcUploadBinaryFileResponse> responseObserver) {
		return UploadServerStreamObserver.<OzgCloudFile, GrpcUploadBinaryFileRequest>builder()
				.requestVerifier(fileUploadRequest -> policyService.checkPermission(fileUploadRequest.getMetadata().getVorgangId()))
				.hasMetadata(GrpcUploadBinaryFileRequest::hasMetadata)
				.metadataMapper(this::buildOzgCloudFile)
				.metadataUploader(this::uploadOzgCloudFile)
				.fileContentUploader(this::uploadFileContent)
				.getFileContent(fileUploadRequest -> fileUploadRequest.getFileContent().toByteArray())
				.responseConsumer(fileId -> completeRequest(buildResponse(fileId), responseObserver))
				.cleanupFileContent(service::deleteContent)
				.build();
	}

	FileId uploadOzgCloudFile(OzgCloudFile ozgCloudFile, Optional<String> contentId) {
		return service.saveOzgCloudFile(updateContentId(ozgCloudFile, contentId));
	}

	OzgCloudFile buildOzgCloudFile(GrpcUploadBinaryFileRequest fileUploadRequest) {
		return OzgCloudFile.builder()
				.vorgangId(fileUploadRequest.getMetadata().getVorgangId())
				.name(fileUploadRequest.getMetadata().getFileName())
				.contentType(fileUploadRequest.getMetadata().getContentType())
				.fieldName(fileUploadRequest.getMetadata().getField())
				.createdBy(getUserId(fileUploadRequest))
				.client(getClient(fileUploadRequest))
				.build();
	}

	String getUserId(GrpcUploadBinaryFileRequest fileUploadRequest) {
		return currentUserService.findUser().flatMap(CallContextUser::getUserId)
				.orElseGet(() -> fileUploadRequest.getMetadata().getContext().getUser().getId());
	}

	String getClient(GrpcUploadBinaryFileRequest fileUploadRequest) {
		return findContextClient().or(() -> getRequestClient(fileUploadRequest)).orElseThrow(() -> new IllegalStateException("Client is missing"));
	}

	Optional<String> getRequestClient(GrpcUploadBinaryFileRequest fileUploadRequest) {
		return Optional.of(fileUploadRequest.getMetadata().getContext().getClient());
	}

	GrpcUploadBinaryFileResponse buildResponse(FileId fileId) {
		return GrpcUploadBinaryFileResponse.newBuilder().setFileId(fileId.toString()).build();
	}

	@Override
	public void getBinaryFileContent(GrpcGetBinaryFileDataRequest request, StreamObserver<GrpcGetBinaryFileDataResponse> responseObserver) {
		policyService.checkPermissionByFile(request.getFileId());

		var fileContent = service.getUploadedFileStream(FileId.from(request.getFileId()));

		downloadFile(responseObserver, new BufferedInputStream(fileContent, FileService.CHUNK_SIZE));
	}

	void downloadFile(StreamObserver<GrpcGetBinaryFileDataResponse> streamObserver, InputStream fileStream) {
		AtomicBoolean doSendFile = new AtomicBoolean(true);
		var callObserver = (CallStreamObserver<GrpcGetBinaryFileDataResponse>) streamObserver;
		callObserver.setOnReadyHandler(() -> sendChunks(callObserver, fileStream, doSendFile));
		sendChunks(callObserver, fileStream, doSendFile);
	}

	synchronized void sendChunks(CallStreamObserver<GrpcGetBinaryFileDataResponse> callObserver, InputStream fileStream, AtomicBoolean doSendFile) {
		if (!doSendFile.get()) {
			return;
		}
		while (callObserver.isReady()) {
			LOG.debug("observer is ready. sending...");
			int size = sendNextChunk(callObserver, fileStream);
			if (size < FileService.CHUNK_SIZE) {
				handleFileEndReached(callObserver, fileStream, doSendFile);
				break;
			}
		}
		LOG.debug("Stop sending");
	}

	int sendNextChunk(CallStreamObserver<GrpcGetBinaryFileDataResponse> callObserver, InputStream fileStream) {
		var bytes = new byte[FileService.CHUNK_SIZE];
		try {
			int size = fileStream.read(bytes);
			LOG.debug("Read {} bytes from file.", size);
			if (size > 0) {
				callObserver.onNext(buildChunkResponse(bytes, size));
			}
			return size;

		} catch (Exception exception) {
			IOUtils.closeQuietly(fileStream, e -> LOG.error("InputStream cannot be closed.", e));
			throw new TechnicalException("Fehler beim Senden des Datei-Chunks", exception);
		}
	}

	void handleFileEndReached(CallStreamObserver<GrpcGetBinaryFileDataResponse> callObserver, InputStream fileStream, AtomicBoolean doSendFile) {
		LOG.debug("File end reached");
		doSendFile.getAndSet(false);
		IOUtils.closeQuietly(fileStream, e -> LOG.error("InputStream cannot be closed.", e));
		callObserver.onCompleted();
	}

	GrpcGetBinaryFileDataResponse buildChunkResponse(byte[] bytes, int size) {
		return GrpcGetBinaryFileDataResponse.newBuilder().setFileContent(ByteString.copyFrom(bytes, 0, size)).build();
	}

	@Override
	public StreamObserver<GrpcUpdateBinaryFileRequest> updateBinaryFile(StreamObserver<GrpcUpdateBinaryFileResponse> responseObserver) {
		return UploadServerStreamObserver.<OzgCloudFile, GrpcUpdateBinaryFileRequest>builder()
				.requestVerifier(this::verifyUpdateRequest)
				.hasMetadata(GrpcUpdateBinaryFileRequest::hasMetadata)
				.metadataMapper(this::buildOzgCloudFile)
				.metadataUploader(this::patchOzgCloudFile)
				.fileContentUploader(this::uploadFileContent)
				.getFileContent(fileUpdateRequest -> fileUpdateRequest.getContent().toByteArray())
				.responseConsumer(fileId -> completeRequest(GrpcUpdateBinaryFileResponse.getDefaultInstance(), responseObserver))
				.cleanupFileContent(service::deleteContent)
				.build();
	}

	void verifyUpdateRequest(GrpcUpdateBinaryFileRequest fileUpdateRequest) {
		validateMetadata(fileUpdateRequest);
		policyService.checkPermissionByFile(fileUpdateRequest.getMetadata().getFileId());
	}

	void validateMetadata(GrpcUpdateBinaryFileRequest fileUpdateRequest) {
		if (!fileUpdateRequest.hasMetadata()) {
			throw new IllegalArgumentException("Metadata is missing");
		}
		var metadata = fileUpdateRequest.getMetadata();
		if (StringUtils.isBlank(metadata.getFileId())) {
			throw new IllegalArgumentException("File ID is missing");
		} else if (StringUtils.isBlank(metadata.getFileName())) {
			throw new IllegalArgumentException("File name is missing");
		} else if (StringUtils.isBlank(metadata.getContentType())) {
			throw new IllegalArgumentException("Content type is missing");
		}
	}

	OzgCloudFile buildOzgCloudFile(GrpcUpdateBinaryFileRequest fileUpdateRequest) {
		var fileId = FileId.from(fileUpdateRequest.getMetadata().getFileId());
		return service.findOzgCloudFiles(List.of(fileId)).map(ozgCloudFile -> updateOzgCloudFile(ozgCloudFile, fileUpdateRequest)).findFirst()
				.orElseThrow(() -> new NotFoundException(OzgCloudFile.class, fileId));
	}

	OzgCloudFile updateOzgCloudFile(OzgCloudFile ozgCloudFile, GrpcUpdateBinaryFileRequest fileUpdateRequest) {
		return ozgCloudFile.toBuilder()
				.name(fileUpdateRequest.getMetadata().getFileName())
				.contentType(fileUpdateRequest.getMetadata().getContentType())
				.build();
	}

	FileId patchOzgCloudFile(OzgCloudFile ozgCloudFile, Optional<String> newContentId) {
		var fileId = FileId.from(ozgCloudFile.getId());
		service.patch(fileId, ozgCloudFile.getVersion(), buildPatch(updateContentId(ozgCloudFile, newContentId)));
		newContentId.ifPresent(newId -> service.deleteContent(ozgCloudFile.getContentId()));
		return fileId;
	}

	OzgCloudFile updateContentId(OzgCloudFile ozgCloudFile, Optional<String> contentId) {
		return contentId.map(id -> ozgCloudFile.toBuilder().contentId(id).build()).orElse(ozgCloudFile);
	}

	Map<String, Object> buildPatch(OzgCloudFile ozgCloudFile) {
		return Map.of(
				OzgCloudFile.FIELD_FILE_NAME, ozgCloudFile.getName(),
				OzgCloudFile.FIELD_CONTENT_TYPE, ozgCloudFile.getContentType(),
				OzgCloudFile.FIELD_CONTENT_ID, ozgCloudFile.getContentId(),
				OzgCloudFile.FIELD_CREATED_BY, getUserId(),
				OzgCloudFile.FIELD_CLIENT, getClient()
		);
	}

	String getUserId() {
		return currentUserService.getUser().getUserId().orElseThrow(() -> new IllegalStateException("User ID is missing"));
	}

	String getClient() {
		return findContextClient().orElseThrow(() -> new IllegalStateException("Client is missing"));
	}

	Optional<String> findContextClient() {
		return currentUserService.findUser().map(CallContextUser::getClientName);
	}

	CompletableFuture<String> uploadFileContent(OzgCloudFile ozgCloudFile, InputStream pipedInput) {
		return service.uploadFileContentStreamAsync(buildUploadFilesReferences(ozgCloudFile), pipedInput);
	}

	UploadedFilesReference buildUploadFilesReferences(OzgCloudFile ozgCloudFile) {
		return UploadedFilesReference.builder()
				.vorgangId(ozgCloudFile.getVorgangId())
				.client(ozgCloudFile.getClient())
				.fileName(ozgCloudFile.getName())
				.name(ozgCloudFile.getFieldName()).build();
	}

	<T> void completeRequest(T response, StreamObserver<T> responseObserver) {
		responseObserver.onNext(response);
		responseObserver.onCompleted();
	}

}