/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.status;

import static de.ozgcloud.vorgang.vorgang.Vorgang.*;

import java.util.Map;
import java.util.NoSuchElementException;
import java.util.function.Consumer;
import java.util.function.Predicate;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.context.event.EventListener;
import org.springframework.stereotype.Component;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandCreatedEvent;
import de.ozgcloud.command.CommandFailedEvent;
import de.ozgcloud.command.CommandRevokeFailedEvent;
import de.ozgcloud.command.RevokeCommandEvent;
import de.ozgcloud.vorgang.command.CommandService;
import de.ozgcloud.vorgang.command.Order;
import de.ozgcloud.vorgang.common.errorhandling.NotFoundException;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangService;
import lombok.extern.log4j.Log4j2;

@Component
@Log4j2
class StatusEventListener {
	@Autowired
	private StatusService statusService;
	@Autowired
	private VorgangService vorgangService;
	@Autowired
	private CommandService commandService;

	@Autowired
	private ApplicationEventPublisher publisher;

	private static final String IS_ABSCHLIESSEN = "{T(de.ozgcloud.vorgang.status.StatusEventListener).IS_ABSCHLIESSEN_EVENT.test(event.getSource())}";
	private static final String IS_ANNEHMEN = "{T(de.ozgcloud.vorgang.status.StatusEventListener).IS_ANNEHMEN_EVENT.test(event.getSource())}";
	private static final String IS_VERWERFEN = "{T(de.ozgcloud.vorgang.status.StatusEventListener).IS_VERWERFEN_EVENT.test(event.getSource())}";
	private static final String IS_BEARBEITEN = "{T(de.ozgcloud.vorgang.status.StatusEventListener).IS_BEARBEITEN_EVENT.test(event.getSource())}";
	private static final String IS_BESCHEIDEN = "{T(de.ozgcloud.vorgang.status.StatusEventListener).IS_BESCHEIDEN_EVENT.test(event.getSource())}";
	private static final String IS_WIEDEREROEFFNEN = "{T(de.ozgcloud.vorgang.status.StatusEventListener).IS_WIEDEREROEFFNEN_EVENT.test(event.getSource())}";
	private static final String IS_ZURUECKHOLEN = "{T(de.ozgcloud.vorgang.status.StatusEventListener).IS_ZURUECKHOLEN_EVENT.test(event.getSource())}";
	private static final String IS_ZURUECKSTELLEN = "{T(de.ozgcloud.vorgang.status.StatusEventListener).IS_ZURUECKSTELLEN_EVENT.test(event.getSource())}";
	private static final String IS_ZUMLOESCHENMARKIEREN = "{T(de.ozgcloud.vorgang.status.StatusEventListener).IS_ZUMLOESCHENMARKIEREN_EVENT.test(event.getSource())}";
	private static final String IS_STATUS_CHANGE = "{T(de.ozgcloud.vorgang.status.StatusEventListener).IS_STATUS_CHANGE_EVENT.test(event.getSource())}";

	public static final Predicate<Command> IS_ABSCHLIESSEN_EVENT = command -> command.getOrder().equals(Order.VORGANG_ABSCHLIESSEN.name());
	public static final Predicate<Command> IS_ANNEHMEN_EVENT = command -> command.getOrder().equals(Order.VORGANG_ANNEHMEN.name());
	public static final Predicate<Command> IS_VERWERFEN_EVENT = command -> command.getOrder().equals(Order.VORGANG_VERWERFEN.name());
	public static final Predicate<Command> IS_BEARBEITEN_EVENT = command -> command.getOrder().equals(Order.VORGANG_BEARBEITEN.name());
	public static final Predicate<Command> IS_BESCHEIDEN_EVENT = command -> command.getOrder().equals(Order.VORGANG_BESCHEIDEN.name());
	public static final Predicate<Command> IS_WIEDEREROEFFNEN_EVENT = command -> command.getOrder().equals(Order.VORGANG_WIEDEREROEFFNEN.name());
	public static final Predicate<Command> IS_ZURUECKHOLEN_EVENT = command -> command.getOrder().equals(Order.VORGANG_ZURUECKHOLEN.name());
	public static final Predicate<Command> IS_ZURUECKSTELLEN_EVENT = command -> command.getOrder().equals(Order.VORGANG_ZURUECKSTELLEN.name());
	public static final Predicate<Command> IS_ZUMLOESCHENMARKIEREN_EVENT = command -> command.getOrder()
			.equals(Order.VORGANG_ZUM_LOESCHEN_MARKIEREN.name());
	public static final Predicate<Command> IS_STATUS_CHANGE_EVENT = command -> command.getOrder().startsWith("VORGANG_");

	@EventListener(condition = IS_ABSCHLIESSEN)
	public void abschiessen(CommandCreatedEvent event) {
		processEvent(event, statusService::abschliessen);
	}

	@EventListener(condition = IS_ANNEHMEN)
	public void annehmen(CommandCreatedEvent event) {
		processEvent(event, statusService::annehmen);
	}

	@EventListener(condition = IS_VERWERFEN)
	public void verwerfen(CommandCreatedEvent event) {
		processEvent(event, statusService::verwerfen);
	}

	@EventListener(condition = IS_BEARBEITEN)
	public void bearbeiten(CommandCreatedEvent event) {
		processEvent(event, statusService::bearbeiten);
	}

	@EventListener(condition = IS_BESCHEIDEN)
	public void bescheiden(CommandCreatedEvent event) {
		processEvent(event, statusService::bescheiden);
	}

	@EventListener(condition = IS_WIEDEREROEFFNEN)
	public void wiedereroeffnen(CommandCreatedEvent event) {
		processEvent(event, statusService::wiedereroeffnen);
	}

	@EventListener(condition = IS_ZURUECKHOLEN)
	public void zurueckholen(CommandCreatedEvent event) {
		processEvent(event, statusService::setStatusNeu);
	}

	@EventListener(condition = IS_ZURUECKSTELLEN)
	public void zurueckstellen(CommandCreatedEvent event) {
		processEvent(event, statusService::annehmen);
	}

	@EventListener(condition = IS_ZUMLOESCHENMARKIEREN)
	public void zuLoeschenMarkieren(CommandCreatedEvent event) {
		processEvent(event, statusService::zumLoeschenMarkieren);
	}

	private void processEvent(CommandCreatedEvent event, Consumer<Command> eventConsumer) {
		var command = event.getSource();
		try {
			command = setPreviousState(command);
			eventConsumer.accept(command);
		} catch (NotFoundException e) {
			handleError("Error processing event because related vorgang (%s) not found.".formatted(command.getVorgangId()), e, command.getId());
		} catch (Exception e) {
			handleError("Unexpected Error on handling Status event.", e, command.getId());
		}
	}

	Command setPreviousState(Command command) {
		var vorgang = vorgangService.getById(command.getVorgangId());
		commandService.setPreviousState(command.getId(), getPreviousStateMap(vorgang));
		return commandService.findById(command.getId())
				.orElseThrow(() -> new NoSuchElementException("No command with Id %s found".formatted(command.getId())));
	}

	Map<String, Object> getPreviousStateMap(Vorgang vorgang) {
		return Map.of(MONGODB_FIELDNAME_STATUS, vorgang.getStatus());
	}

	@EventListener(condition = IS_STATUS_CHANGE)
	public void statusRevokedEvent(RevokeCommandEvent event) {
		var command = event.getSource();
		try {
			statusService.revokeStatusChange(command);
		} catch (Exception e) {
			LOG.error("Unexpected error on command revoke.", e);
			publisher.publishEvent(new CommandRevokeFailedEvent(command.getId(), "Unexpected error on command revoke"));
		}
	}

	private void handleError(String message, Exception e, String commandId) {
		LOG.error(message, e);
		publisher.publishEvent(new CommandFailedEvent(commandId, message));
	}
}
