/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.attached_item;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import jakarta.validation.ValidationException;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.NullSource;
import org.junit.jupiter.params.provider.ValueSource;
import org.mapstruct.factory.Mappers;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.context.ApplicationEventPublisher;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.RevokeCommandEvent;
import de.ozgcloud.command.VorgangAttachedItemCreatedEvent;
import de.ozgcloud.vorgang.command.CommandCreatedEventTestFactory;
import de.ozgcloud.vorgang.command.CommandService;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangDeletedEventTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

class VorgangAttachedItemEventListenerTest {

	@Spy
	@InjectMocks
	private VorgangAttachedItemEventListener listener;

	@Mock
	private VorgangAttachedItemService service;
	@Mock
	private ApplicationEventPublisher publisher;
	@Mock
	private CommandService commandService;
	@Spy
	private VorgangAttachedItemMapper mapper = Mappers.getMapper(VorgangAttachedItemMapper.class);

	@DisplayName("Create item")
	@Nested
	class TestCreateItem {

		private final Map<String, Object> bodyMap = VorgangAttachedItemTestFactory.asMap();

		private final Command command = CommandTestFactory.createBuilder().bodyObject(bodyMap).relationId(VorgangTestFactory.ID).build();

		@Captor
		private ArgumentCaptor<VorgangAttachedItem> itemCaptor;
		@Captor
		private ArgumentCaptor<VorgangAttachedItemCreatedEvent> vorgangAttachedItemCreatedEventCaptor;

		@BeforeEach
		void mock() {
			when(service.create(any(), any())).thenReturn(VorgangAttachedItemTestFactory.create());
		}

		@DisplayName("should call service")
		@Nested
		class TestCallService {

			@Test
			void withItem() {
				listener.createItem(CommandCreatedEventTestFactory.create(command));

				verify(service).create(eq(command), itemCaptor.capture());
				assertThat(itemCaptor.getValue()).usingRecursiveComparison()
						.ignoringFields(VorgangAttachedItem.FIELDNAME_ID, VorgangAttachedItem.FIELDNAME_VERSION)
						.isEqualTo(VorgangAttachedItemTestFactory.create());
			}

			@Test
			void withIdAndVersionAsNull() {
				listener.createItem(CommandCreatedEventTestFactory.create(command));

				verify(service).create(eq(command), itemCaptor.capture());
				var item = itemCaptor.getValue();

				assertThat(item.getId()).isNull();
				assertThat(item.getVersion()).isZero();
			}

			@Test
			void withCommandId() {
				listener.createItem(CommandCreatedEventTestFactory.create(command));

				verify(service).create(any(), any());
			}
		}
	}

	@DisplayName("Validate create command")
	@Nested
	class TestValidateCreateCommand {

		@Test
		void shouldThrowExceptionOnIdMismatching() {
			var command = CommandTestFactory.createBuilder().vorgangId("other").build();

			assertThatThrownBy(() -> listener.validateCreateCommand(command)).isInstanceOf(ValidationException.class);
		}
	}

	@DisplayName("Update item")
	@Nested
	class TestUpdateItem {
		private final Map<String, Object> bodyMap = VorgangAttachedItemTestFactory.asMap();

		private Command command = buildCommand(bodyMap);

		@Captor
		private ArgumentCaptor<VorgangAttachedItem> itemCaptor;
		@Captor
		private ArgumentCaptor<Map<String, Object>> itemMapCaptor;
		@Captor
		private ArgumentCaptor<VorgangAttachedItemUpdatedEvent> eventCaptor;

		@BeforeEach
		void mock() {
			doReturn(VorgangAttachedItemTestFactory.create()).when(listener).getVorgangAttachedItem(any());
		}

		@Test
		void shouldGetVorgangAttachedItem() {
			doUpdateItem();

			verify(listener).getVorgangAttachedItem(VorgangAttachedItemTestFactory.ID);
		}

		@Test
		void shouldCallSetCommandPreviousState() {
			doUpdateItem();

			verify(listener).setCommandPreviousState(CommandTestFactory.ID, VorgangAttachedItemTestFactory.STRING_MAP);
		}

		@DisplayName("service Call")
		@Nested
		class TestCallService {

			@Test
			void shouldHaveItem() {
				doUpdateItem();

				verify(service).update(eq(CommandTestFactory.ID), itemCaptor.capture());
				assertThat(itemCaptor.getValue()).usingRecursiveComparison().isEqualTo(VorgangAttachedItemTestFactory.create());
			}

			@Test
			void shouldHaveCommand() {
				doUpdateItem();

				verify(service).update(any(), any());
			}

			@Test
			void shouldHaveSetRelationIdAsId() {
				var bMap = new HashMap<>(bodyMap);
				bMap.put(VorgangAttachedItem.FIELDNAME_ID, "otherId");

				listener.updateItem(CommandCreatedEventTestFactory.create(buildCommand(bMap)));

				verify(service).update(eq(CommandTestFactory.ID), itemCaptor.capture());
				assertThat(itemCaptor.getValue().getId()).isEqualTo(VorgangAttachedItemTestFactory.ID);
			}

			@DisplayName("should call forceUpdate when")
			@ParameterizedTest(name = "relationVersion is {0}")
			@NullSource
			@ValueSource(longs = { -10, 0 })
			void shouldCallForceUpdate(Long relationVersion) {
				command = CommandTestFactory.createBuilder().relationId(VorgangAttachedItemTestFactory.ID).relationVersion(relationVersion)
						.bodyObject(bodyMap)
						.build();

				listener.updateItem(CommandCreatedEventTestFactory.create(command));

				verify(service).forceUpdate(CommandTestFactory.ID, VorgangAttachedItemTestFactory.ID, VorgangAttachedItemTestFactory.STRING_MAP);
			}
		}

		private void doUpdateItem() {
			listener.updateItem(CommandCreatedEventTestFactory.create(command));
		}

		private Command buildCommand(Map<String, Object> bodyMap) {
			return CommandTestFactory.createBuilder().relationId(VorgangAttachedItemTestFactory.ID)
					.relationVersion(VorgangAttachedItemTestFactory.VERSION).bodyObject(bodyMap).build();
		}
	}

	@DisplayName("Get vorgangAttachedItem")
	@Nested
	class TestGetVorgangAttachedItem {

		@BeforeEach
		void mock() {
			when(service.getById(any())).thenReturn(VorgangAttachedItemTestFactory.create());
		}

		@Test
		void shouldCallService() {
			listener.getVorgangAttachedItem(VorgangAttachedItemTestFactory.ID);

			verify(service).getById(VorgangAttachedItemTestFactory.ID);
		}

		@Test
		void shouldReturnValue() {
			var item = listener.getVorgangAttachedItem(VorgangAttachedItemTestFactory.ID);

			assertThat(item).usingRecursiveComparison().isEqualTo(VorgangAttachedItemTestFactory.create());
		}
	}

	@DisplayName("Set command previous state")
	@Nested
	class TestSetCommandPreviousState {

		@Test
		void shouldCallCommandService() {
			listener.setCommandPreviousState(CommandTestFactory.ID, VorgangAttachedItemTestFactory.STRING_MAP);

			verify(commandService).setPreviousState(CommandTestFactory.ID, VorgangAttachedItemTestFactory.STRING_MAP);
		}
	}

	@DisplayName("Patch item")
	@Nested
	class TestPatchItem {

		@Captor
		private ArgumentCaptor<VorgangAttachedItemUpdatedEvent> eventCaptor;

		private final Map<String, Object> itemMap = Map.of("number", "73");
		private final Set<String> itemKeys = Set.of("number");
		private final Map<String, Object> bodyMap = Map.of("id", VorgangAttachedItemTestFactory.ID, VorgangAttachedItem.FIELDNAME_ITEM,
				itemMap);

		private final Command command = CommandTestFactory.createBuilder().relationId(VorgangAttachedItemTestFactory.ID)
				.relationVersion(VorgangAttachedItemTestFactory.VERSION).bodyObject(bodyMap).build();

		private final Map<String, Object> patchMap = Collections.emptyMap();

		@BeforeEach
		void mock() {
			doReturn(patchMap).when(listener).getExistingItemMapEntries(any(), any());
		}

		@Test
		void shouldBuildPatchMap() {
			doPatchItem();

			verify(listener).getExistingItemMapEntries(CommandTestFactory.RELATION_ID, itemKeys);
		}

		@Test
		void shouldCallSetCommandPreviousState() {
			doPatchItem();

			verify(listener).setCommandPreviousState(CommandTestFactory.ID, patchMap);
		}

		@DisplayName("service Call")
		@Nested
		class TestCallService {
			@SuppressWarnings("unchecked")
			@Test
			void shouldHaveItem() {
				doPatchItem();

				verify(service).patch(any(), (Map<String, Object>) eq(bodyMap.get(VorgangAttachedItem.FIELDNAME_ITEM)));
			}

			@Test
			void shouldHaveCommand() {
				doPatchItem();

				verify(service).patch(eq(command), any());
			}
		}

		private void doPatchItem() {
			listener.patchItem(CommandCreatedEventTestFactory.create(command));
		}
	}

	@DisplayName("Build patch map")
	@Nested
	class TestBuildPatchMap {

		private final Set<String> commandPatchKeySet = Set.of("KeyToPatch", "KeyWithNullValueToPatch");
		private final Map<String, Object> itemMap = buildItemMap();

		private Map<String, Object> buildItemMap() {
			var itemMap = new HashMap<String, Object>();
			itemMap.put("KeyToPatch", "ValueToPatch");
			itemMap.put("KeyToIgnore", "ValueToIgnore");
			itemMap.put("KeyWithNullValueToPatch", null);
			return itemMap;
		}

		private final VorgangAttachedItem vorgangAttachedItem = VorgangAttachedItemTestFactory.createBuilder().item(itemMap).build();

		@BeforeEach
		void mock() {
			when(service.getById(any())).thenReturn(vorgangAttachedItem);
		}

		@Test
		void shouldCallGetVorgangAttachedItem() {
			listener.getExistingItemMapEntries(VorgangAttachedItemTestFactory.ID, commandPatchKeySet);

			verify(service).getById(VorgangAttachedItemTestFactory.ID);
		}

		@Test
		void shouldReturnMapWithPatchEntry() {
			var patchMap = listener.getExistingItemMapEntries(VorgangAttachedItemTestFactory.ID, commandPatchKeySet);

			assertThat(patchMap).hasSize(2).containsEntry("KeyToPatch", "ValueToPatch");
		}

		@Test
		void shouldReturnMapWithNullPatchEntry() {
			var patchMap = listener.getExistingItemMapEntries(VorgangAttachedItemTestFactory.ID, commandPatchKeySet);

			assertThat(patchMap).hasSize(2).containsKey("KeyWithNullValueToPatch");
			assertThat(patchMap.get("KeyWithNullValueToPatch")).isNull();
		}
	}

	@Nested
	@DisplayName("Delete item")
	class TestDeleteItem {

		private final Command command = CommandTestFactory.createBuilder().relationId(VorgangAttachedItemTestFactory.ID)
				.relationVersion(VorgangAttachedItemTestFactory.VERSION).build();

		@Nested
		@DisplayName("by itemId and version")
		class TestDeleteByIdAndVersion {

			@Test
			void shouldCallService() {
				listener.deleteItem(CommandCreatedEventTestFactory.create(command));

				verify(service).markAsDeleteByIdAndVersion(VorgangAttachedItemTestFactory.ID, VorgangAttachedItemTestFactory.VERSION);
			}

			@Test
			void shouldPublishCommandExecutedEvent() {
				listener.deleteItem(CommandCreatedEventTestFactory.create(command));

				verify(publisher).publishEvent(any(VorgangAttachedItemDeletedEvent.class));
			}

			@Test
			void shouldCallHandleException() {
				doThrow(new RuntimeException()).when(service).markAsDeleteByIdAndVersion(anyString(), anyLong());

				listener.deleteItem(CommandCreatedEventTestFactory.create(command));

				verify(listener).handleException(any(), eq(CommandTestFactory.ID));
			}
		}

		@Nested
		@DisplayName("by vorgangId")
		class TestDeleteByVorgangId {

			@Test
			void shouldCallService() {
				listener.onVorgangDeleted(VorgangDeletedEventTestFactory.create());

				verify(service).deleteByVorgangId(VorgangTestFactory.ID);
			}
		}
	}

	@DisplayName("On revoke create item")
	@Nested
	class TestOnRevokeCreateItem {

		private final Command command = CommandTestFactory.create();
		private final RevokeCommandEvent revokeEvent = new RevokeCommandEvent(command);

		@Test
		void shouldCallService() {
			listener.onRevokeCreateItem(revokeEvent);

			verify(service).revokeCreate(command);
		}
	}

	@DisplayName("On revoke delete item")
	@Nested
	class TestOnRevokeDeleteItem {

		private final Command command = CommandTestFactory.create();
		private final RevokeCommandEvent revokeEvent = new RevokeCommandEvent(command);

		@Test
		void shouldCallService() {
			listener.onRevokeDeleteItem(revokeEvent);

			verify(service).revokeDelete(command);
		}
	}

	@DisplayName("On revoke update item")
	@Nested
	class TestOnRevokeUpdateItem {

		private final Command command = CommandTestFactory.create();
		private final RevokeCommandEvent revokeEvent = new RevokeCommandEvent(command);

		@BeforeEach
		void mock() {
			when(commandService.getById(any())).thenReturn(command);
		}

		@Test
		void shouldGetCommand() {
			listener.onRevokeUpdateItem(revokeEvent);

			verify(commandService).getById(CommandTestFactory.ID);
		}

		@Test
		void shouldCallService() {
			listener.onRevokeUpdateItem(revokeEvent);

			verify(service).revokeUpdate(command, CommandTestFactory.PREVIOUS_STATE);
		}
	}

	@DisplayName("On revoke patch item")
	@Nested
	class TestOnRevokePatchItem {

		private final Command command = CommandTestFactory.create();
		private final RevokeCommandEvent revokeEvent = new RevokeCommandEvent(command);

		@BeforeEach
		void mock() {
			when(commandService.getById(any())).thenReturn(CommandTestFactory.create());
		}

		@Test
		void shouldGetCommand() {
			listener.onRevokePatchItem(revokeEvent);

			verify(commandService).getById(CommandTestFactory.ID);
		}

		@Test
		void shouldCallService() {
			listener.onRevokePatchItem(revokeEvent);

			verify(service).revokePatch(command, CommandTestFactory.PREVIOUS_STATE);
		}
	}

}