/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.clientattribute;

import static de.ozgcloud.vorgang.clientattribute.ClientAttributeTestFactory.*;
import static org.assertj.core.api.Assertions.*;

import java.util.Map;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.EnumSource.Mode;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.security.access.AccessDeniedException;
import org.springframework.security.test.context.support.WithMockUser;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.callcontext.CallContextTestFactory;
import de.ozgcloud.vorgang.clientattribute.ClientAttribute.AccessPermission;
import de.ozgcloud.vorgang.common.errorhandling.NotFoundException;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

@DataITCase
class ClientAttributeRepositoryITCase {

	@Autowired
	private ClientAttributeRepository repository;

	@Autowired
	private MongoOperations mongoOperations;

	@BeforeEach
	void clearCollection() {
		mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
	}

	@Nested
	class TestPersistAttribute {

		private Vorgang vorgang;

		@BeforeEach
		void persistVorgang() {
			vorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).clientAttributes(null).build());
		}

		@Test
		void withIntValue() {
			repository.set(vorgang.getId(), ClientAttributeTestFactory.create());

			Map<String, ClientAttribute> attributes = loadTestClientAttributes();

			assertThat(attributes.get(ATTRIBUTE_NAME))
					.usingRecursiveComparison().ignoringFields(ClientAttribute.FIELD_CLIENTNAME, ClientAttribute.FIELD_ATTRIBUTENAME)
					.isEqualTo(ClientAttributeTestFactory.create());
		}

		@Test
		void withStringValue() {
			ClientAttribute attribute = ClientAttributeTestFactory.createBuilder().intValue(Optional.empty())
					.stringValue(Optional.of(STR_VALUE)).build();

			repository.set(vorgang.getId(), attribute);

			Map<String, ClientAttribute> attributes = loadTestClientAttributes();
			assertThat(attributes.get(ATTRIBUTE_NAME))
					.usingRecursiveComparison().ignoringFields(ClientAttribute.FIELD_CLIENTNAME, ClientAttribute.FIELD_ATTRIBUTENAME)
					.isEqualTo(attribute);
		}

		@Test
		void withBooleanValue() {
			ClientAttribute attribute = ClientAttributeTestFactory.createBuilder().intValue(Optional.empty())
					.boolValue(Optional.of(BOOL_VALUE)).build();

			repository.set(vorgang.getId(), attribute);

			Map<String, ClientAttribute> attributes = loadTestClientAttributes();
			assertThat(attributes.get(ATTRIBUTE_NAME))
					.usingRecursiveComparison().ignoringFields(ClientAttribute.FIELD_CLIENTNAME, ClientAttribute.FIELD_ATTRIBUTENAME)
					.isEqualTo(attribute);
		}

		@Test
		void withDoubleValue() {
			ClientAttribute attribute = ClientAttributeTestFactory.createBuilder().intValue(Optional.empty())
					.doubleValue(Optional.of(DOUBLE_VALUE)).build();

			repository.set(vorgang.getId(), attribute);

			Map<String, ClientAttribute> attributes = loadTestClientAttributes();
			assertThat(attributes.get(ATTRIBUTE_NAME))
					.usingRecursiveComparison().ignoringFields(ClientAttribute.FIELD_CLIENTNAME, ClientAttribute.FIELD_ATTRIBUTENAME)
					.isEqualTo(attribute);
		}

	}

	@Nested
	class ToUpdateAttribute {

		@Nested
		class CheckExists {
			@Test
			void shouldBeTrue() {
				var vorgang = persistVorgangWithAttribute(ClientAttributeTestFactory.create());

				var exists = repository.exists(ClientAttributeIdentificatorTestFactory.createBuilder().vorgangId(vorgang.getId()).build());

				assertThat(exists).isTrue();
			}

			@Test
			void shouldBeFalseOnWrongClientName() {
				var vorgang = persistVorgangWithAttribute(ClientAttributeTestFactory.create());

				var exists = repository
						.exists(ClientAttributeIdentificatorTestFactory.createBuilder().vorgangId(vorgang.getId()).clientName("other").build());

				assertThat(exists).isFalse();
			}

			@Test
			void shouldBeFalseOnOtherAttributeName() {
				var vorgang = persistVorgangWithAttribute(ClientAttributeTestFactory.create());

				var exists = repository
						.exists(ClientAttributeIdentificatorTestFactory.createBuilder().vorgangId(vorgang.getId()).attributeName("other").build());

				assertThat(exists).isFalse();
			}
		}
	}

	@WithMockUser(username = CallContextTestFactory.CLIENT)
	@Nested
	class TestDelete {

		@Test
		void shouldRemoveClientAttribute() {
			var vorgang = persistVorgangWithAttribute(ClientAttributeTestFactory.createBuilder().access(AccessPermission.READ_WRITE).build());

			repository.delete(ClientAttributeIdentificatorTestFactory.createBuilder().vorgangId(vorgang.getId()).build());

			var attributes = loadTestClientAttributes();
			assertThat(attributes).isEmpty();
		}

		@WithMockUser(username = "OTHER")
		@ParameterizedTest
		@EnumSource(mode = Mode.EXCLUDE, names = "READ_WRITE")
		void shouldDenyIfNotWritableForCallingClient(AccessPermission permission) {
			var vorgang = persistVorgangWithAttribute(ClientAttributeTestFactory.createBuilder().access(permission).build());
			var attributeId = ClientAttributeIdentificatorTestFactory.createBuilder().vorgangId(vorgang.getId()).build();

			assertThatThrownBy(() -> repository.delete(attributeId)).isInstanceOf(AccessDeniedException.class);

			var attributes = loadTestClientAttributes();
			assertThat(attributes).isNotEmpty();
		}

		@ParameterizedTest
		@EnumSource
		void shouldAllowAllForOwningClient(AccessPermission permission) {
			var vorgang = persistVorgangWithAttribute(ClientAttributeTestFactory.createBuilder().access(permission).build());

			repository.delete(ClientAttributeIdentificatorTestFactory.createBuilder().vorgangId(vorgang.getId()).build());

			var attributes = loadTestClientAttributes();
			assertThat(attributes).isEmpty();
		}

		@Test
		void shouldThrowNotFoundExceptionForOtherClient() {
			var vorgang = persistVorgangWithAttribute(ClientAttributeTestFactory.createBuilder().clientName("OTHER").build());
			var attributeId = ClientAttributeIdentificatorTestFactory.createBuilder().vorgangId(vorgang.getId()).build();

			assertThatThrownBy(() -> repository.delete(attributeId)).isInstanceOf(NotFoundException.class);
		}

		@WithMockUser(username = "OTHER-OTHER")
		@Test
		void shouldThrowNotFoundExceptionWhenCallingWithOtherClient() {
			var vorgang = persistVorgangWithAttribute(ClientAttributeTestFactory.createBuilder().clientName("OTHER").build());
			var attributeId = ClientAttributeIdentificatorTestFactory.createBuilder().vorgangId(vorgang.getId()).build();

			assertThatThrownBy(() -> repository.delete(attributeId)).isInstanceOf(NotFoundException.class);
		}
	}

	@WithMockUser(username = CallContextTestFactory.CLIENT)
	@Nested
	class TestUpdateAttribute {

		@Test
		void shouldUpdateAccessForOwningClient() {
			var vorgang = persistVorgangWithAttribute(ClientAttributeTestFactory.createBuilder().access(AccessPermission.READ_ONLY).build());

			repository.update(vorgang.getId(), ClientAttributeTestFactory.createBuilder().access(AccessPermission.READ_WRITE).build());
			var attribute = loadTestClientAttributes().get(ATTRIBUTE_NAME);

			assertThat(attribute.getAccess()).isEqualTo(AccessPermission.READ_WRITE);
		}

		@WithMockUser(username = "other")
		@Test
		void shouldNOTUpdateAccessForOtherClient() {
			var vorgang = persistVorgangWithAttribute(
					ClientAttributeTestFactory.createBuilder().access(AccessPermission.READ_WRITE).build());

			repository.update(vorgang.getId(),
					ClientAttributeTestFactory.createBuilder().access(AccessPermission.READ_ONLY).intValue(Optional.of(5L)).build());
			var attribute = loadTestClientAttributes().get(ATTRIBUTE_NAME);

			assertThat(attribute.getAccess()).isEqualTo(AccessPermission.READ_WRITE);
		}

		@WithMockUser(username = "other")
		@Test
		void shouldUpdateValueForOtherClient() {
			var vorgang = persistVorgangWithAttribute(
					ClientAttributeTestFactory.createBuilder().access(AccessPermission.READ_WRITE).build());

			repository.update(vorgang.getId(),
					ClientAttributeTestFactory.createBuilder().intValue(Optional.empty()).stringValue(Optional.of("changed")).build());
			var attribute = loadTestClientAttributes().get(ATTRIBUTE_NAME);

			assertThat(attribute.getStringValue()).contains("changed");
		}

		@WithMockUser(username = "other")
		@Test
		void shouldThrowExceptionOnReadOnly() {
			var vorgang = persistVorgangWithAttribute(
					ClientAttributeTestFactory.createBuilder().access(AccessPermission.READ_ONLY).build());

			var attribute = ClientAttributeTestFactory.createBuilder().intValue(Optional.empty()).stringValue(Optional.of("changed")).build();
			var vorgangId = vorgang.getId();
			assertThatThrownBy(() -> repository.update(vorgangId, attribute)).isInstanceOf(AccessDeniedException.class);

			var loaded = loadTestClientAttributes().get(ATTRIBUTE_NAME);
			assertThat(loaded.getStringValue()).isEmpty();
		}
	}

	private Vorgang persistVorgangWithAttribute(ClientAttribute attribute) {
		var attributesMap = ClientAttributesMapTestFactory.createWithAttribute(attribute);
		return mongoOperations.save(VorgangTestFactory.createBuilder().id(null).clientAttributes(attributesMap).build());
	}

	@SuppressWarnings("unchecked")
	private Map<String, ClientAttribute> loadTestClientAttributes() {
		var vorgangList = mongoOperations.findAll(Map.class, "vorgang");

		assertThat(vorgangList).hasSize(1);
		return (Map<String, ClientAttribute>) ((Map<String, Object>) vorgangList.get(0).get(Vorgang.FIELD_CLIENT_ATTRIBUTES))
				.get(CallContextTestFactory.CLIENT);
	}
}
