/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.command;

import static org.assertj.core.api.Assertions.*;
import static org.awaitility.Awaitility.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.concurrent.TimeUnit;

import org.assertj.core.api.InstanceOfAssertFactories;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.security.test.context.support.WithMockUser;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandStatus;
import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.common.errorhandling.NotFoundException;
import de.ozgcloud.vorgang.grpc.command.GrpcCommand;
import de.ozgcloud.vorgang.grpc.command.GrpcCommandBodyField;
import de.ozgcloud.vorgang.grpc.command.GrpcCommandResponse;
import de.ozgcloud.vorgang.grpc.command.GrpcCreateCommandRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcGetCommandRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcGetPendingCommandsRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcGetPendingCommandsResponse;
import de.ozgcloud.vorgang.grpc.command.GrpcRevokeCommandRequest;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.Vorgang.Status;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import io.grpc.stub.StreamObserver;

/**
 * TODO Diesen Test Umbauen
 *
 * einige Testfälle in den UnitTest (GrpcCommandServiceTest) umziehen
 *
 * einige Testfälle in den Service-ITCase (GrpcCommandServiceITCase - neu zu
 * erstellen oder diesen wieder umbauen) umziehen
 *
 * einige wenige Testfälle in eine E2E-ITCase (CommandITCase) ziehen, dabei aber
 * reaktionen von Umsystemen (NAchrichtenMAnager, NotificationManager)
 * berúcksichtigen / ausschalten
 */

@WithMockUser
@DataITCase
class CommandITCase {

	@Autowired
	private GrpcCommandService service;
	@Autowired
	private CommandService commandService;

	@Autowired
	private MongoOperations mongoOperations;

	@Mock
	private StreamObserver<GrpcCommandResponse> responseObserver;

	@Nested
	@DisplayName("Create Command")
	class TestCreateCommand {

		private Vorgang persistedVorgang;

		@Mock
		private StreamObserver<GrpcCommandResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcCommandResponse> responseCaptor;

		private GrpcCreateCommandRequest request = GrpcCreateCommandRequestTestFactory.createBuilder().setOrderString(Order.VORGANG_ANNEHMEN.name())
				.build();

		@BeforeEach
		void persistVorgang() {
			mongoOperations.dropCollection(Command.class);
			mongoOperations.dropCollection(Vorgang.class);

			persistedVorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build());
			request = GrpcCreateCommandRequestTestFactory.createBuilder().clearRelationId().setVorgangId(persistedVorgang.getId())
					.setOrderString(Order.VORGANG_ANNEHMEN.name()).setRelationId(persistedVorgang.getId()).build();
		}

		@Test
		void validateGrpcResponse() {
			callServiceCreateCommand();

			var response = responseCaptor.getValue();

			assertThat(response.getResponseCode()).isEqualTo(GrpcCommandResponse.GrpcResponseCode.OK);
			assertThat(response.getCommand()).isNotNull();
		}

		@Test
		void shouldCreateCommand() {
			callServiceCreateCommand();
			var response = responseCaptor.getValue();

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				var command = commandService.findById(response.getCommand().getId());

				assertThat(command).isPresent().get().satisfies(cmd -> {
					assertThat(cmd.getId()).isNotNull();
					assertThat(cmd.getCreatedAt()).isNotNull();
					assertThat(cmd.getStatus()).isEqualTo(CommandStatus.FINISHED);
					assertThat(cmd.getRelationId()).isEqualTo(persistedVorgang.getId());
				});
			});

		}

		@Test
		void validateUpdatedVorgang() {
			callServiceCreateCommand();

			await().atMost(20, TimeUnit.SECONDS).untilAsserted(() -> {
				var vorgang = mongoOperations.findById(persistedVorgang.getId(), Vorgang.class);

				assertThat(vorgang).isNotNull();
				assertThat(vorgang.getVersion()).isEqualTo(VorgangTestFactory.VERSION + 1);
			});

		}

		@Test
		void validateCreatedCommand() {
			callServiceCreateCommand();

			await().atMost(20, TimeUnit.SECONDS).untilAsserted(() -> {
				var commands = mongoOperations.findAll(Command.class);

				assertThat(commands).hasSize(1).first().satisfies(command -> {
					assertThat(command.getId()).hasSize(36);
					assertThat(command.getCreatedAt()).isNotNull();
					assertThat(command.getStatus()).isEqualTo(CommandStatus.FINISHED);
					assertThat(command.getRelationId()).isEqualTo(persistedVorgang.getId());
					assertThat(command.getRelationVersion()).isEqualTo(VorgangTestFactory.VERSION);
				});
			});
		}

		@DisplayName("with Order SEND_POSTFACH_MAIL")
		@Nested
		class TestOrderSendPostfachMail {
			private GrpcCreateCommandRequest request;

			@BeforeEach
			void buildRequest() {
				request = GrpcCreateCommandRequestTestFactory.createBuilder()
						.setVorgangId(persistedVorgang.getId()).setRelationId(persistedVorgang.getId())
						.setOrderString("SEND_POSTFACH_MAIL")
						.clearBody().addBody(GrpcCommandBodyField.newBuilder().setName("mailBody").build())
						.build();
			}

			@Test
			void shouldPersistCommand() {
				callServiceCreateCommand(request);

				var commands = mongoOperations.findAll(Command.class);
				assertThat(commands).hasSize(1);
			}

			@Test
			void shouldPersistBodyWithNullValue() {
				callServiceCreateCommand(request);

				var commands = mongoOperations.findAll(Command.class);
				assertThat(commands.get(0).getBody().get("mailBody")).isNull();
			}

			@DisplayName("load command")
			@Nested
			class TestLoadCommand {

				@Mock
				private StreamObserver<GrpcGetPendingCommandsResponse> responseObserver;
				@Captor
				private ArgumentCaptor<GrpcGetPendingCommandsResponse> responseCaptor;

				@BeforeEach
				void prepareDatabase() {
					mongoOperations.save(
							CommandTestFactory.createBuilder()
									.vorgangId(persistedVorgang.getId())
									.relationId(persistedVorgang.getId())
									.order(Order.SEND_POSTFACH_MAIL.toString())
									.status(CommandStatus.PENDING).build());
				}

				@Test
				void shouldReturnPendingCommand() {
					var request = GrpcGetPendingCommandsRequest.newBuilder().setVorgangId(persistedVorgang.getId()).build();

					callServiceGetPendingCommands(request);

					assertThat(responseCaptor.getValue().getCommandList()).hasSize(1);
				}

				private void callServiceGetPendingCommands(GrpcGetPendingCommandsRequest request) {
					service.getPendingCommands(request, responseObserver);

					verify(responseObserver).onNext(responseCaptor.capture());

				}
			}
		}

		private void callServiceCreateCommand() {
			callServiceCreateCommand(request);
		}

		private void callServiceCreateCommand(GrpcCreateCommandRequest request) {
			service.createCommand(request, responseObserver);

			verify(responseObserver).onNext(responseCaptor.capture());
		}
	}

	@Nested
	class TestGetCommand {

		@Mock
		private StreamObserver<GrpcCommand> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcCommand> responseCaptor;

		@BeforeEach
		void persistCommand() {
			mongoOperations.dropCollection(Command.class);
			mongoOperations.dropCollection(Vorgang.class);

			mongoOperations.save(CommandTestFactory.create());
		}

		@Test
		void shouldReturnCommand() {
			var request = GrpcGetCommandRequest.newBuilder().setId(CommandTestFactory.ID).build();

			callServiceGetCommand(request);

			var command = responseCaptor.getValue();

			assertThat(command.getId()).isEqualTo(CommandTestFactory.ID);
			assertThat(command.getCreatedAt()).isEqualTo(CommandTestFactory.CREATED_AT_STR);
			assertThat(command.getCreatedBy()).isEqualTo(CommandTestFactory.CREATED_BY);
			assertThat(command.getStatus()).isEqualTo(CommandTestFactory.STATUS.toString());
			assertThat(command.getRelationId()).isEqualTo(CommandTestFactory.RELATION_ID);
		}

		@Test
		void shouldThrowOnCommandNotFoundById() {
			var request = GrpcGetCommandRequest.newBuilder().setId("invalid_command_id").build();

			assertThatThrownBy(() -> callServiceGetCommand(request)).isInstanceOf(NotFoundException.class);
		}

		private void callServiceGetCommand(GrpcGetCommandRequest request) {
			service.getCommand(request, responseObserver);

			verify(responseObserver).onNext(responseCaptor.capture());
		}
	}

	@Nested
	class TestRevokeCommand {

		@Mock
		private StreamObserver<GrpcCommandResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcCommandResponse> responseCaptor;
		private final GrpcRevokeCommandRequest request = GrpcRevokeCommandRequest.newBuilder().setId(CommandTestFactory.ID).build();

		@Nested
		class TestRevokeAngenommen {

			@BeforeEach
			void persistVorgangWithVorgangAnnehmenCommand() {
				mongoOperations.dropCollection(Vorgang.class);
				mongoOperations.save(VorgangTestFactory.createBuilder().status(Status.ANGENOMMEN).build());

				mongoOperations.save(CommandTestFactory.createBuilder().status(CommandStatus.FINISHED)
						.relationId(VorgangTestFactory.ID).relationVersion(VorgangTestFactory.VERSION - 1)
						.previousState(Map.of(Vorgang.MONGODB_FIELDNAME_STATUS, Status.NEU)).build());
			}

			@Test
			void shouldSetRevokeStatus() {
				callServiceRevokeCommand();

				await().atMost(20, TimeUnit.SECONDS).untilAsserted(() -> {
					var command = findCommand(responseCaptor.getValue().getCommand().getId());

					assertThat(command).isNotNull().extracting(Command::getStatus).isEqualTo(CommandStatus.REVOKED);
				});
			}

			@Test
			void shouldUpdateVorgang() {
				callServiceRevokeCommand();

				await().atMost(20, TimeUnit.SECONDS).untilAsserted(() -> {
					var vorgang = mongoOperations.findById(VorgangTestFactory.ID, Vorgang.class);
					assertThat(vorgang).isNotNull().extracting(Vorgang::getStatus).isEqualTo(Status.NEU);
				});
			}

			@Test
			void shouldUpdateCommand() {
				callServiceRevokeCommand();

				await().atMost(20, TimeUnit.SECONDS).untilAsserted(() -> {
					var command = findCommand();
					assertThat(command).isNotNull().extracting(Command::getStatus).isEqualTo(CommandStatus.REVOKED);
				});
			}

			@Test
			void shouldNOTUpdateRelationVersion() {
				callServiceRevokeCommand();

				var command = findCommand();
				assertThat(command).isNotNull().extracting(Command::getRelationVersion).isEqualTo(VorgangTestFactory.VERSION - 1);
			}

			private Command findCommand() {
				return findCommand(CommandTestFactory.ID);
			}

			private Command findCommand(String id) {
				return mongoOperations.findById(id, Command.class);
			}
		}

		@Nested
		class TestRevokeVerworfen {

			@BeforeEach
			void persistVorgangWithVorgangVerwerfenCommand() {
				mongoOperations.save(VorgangTestFactory.createBuilder().status(Status.VERWORFEN).build());

				mongoOperations.save(CommandTestFactory.createBuilder().status(CommandStatus.FINISHED).relationId(VorgangTestFactory.ID)
						.relationVersion(VorgangTestFactory.VERSION - 1).order(Order.VORGANG_VERWERFEN.toString())
						.previousState(Map.of(Vorgang.MONGODB_FIELDNAME_STATUS, Status.IN_BEARBEITUNG))
						.build());
			}

			@Test
			void shouldSetRevokeStatus() {
				callServiceRevokeCommand();

				await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
					var command = commandService.findById(responseCaptor.getValue().getCommand().getId());

					assertThat(command).isPresent().get().extracting(Command::getStatus).isEqualTo(CommandStatus.REVOKED);
				});
			}

			@Test
			void shouldUpdateVorgang() {
				callServiceRevokeCommand();

				await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
					var vorgang = mongoOperations.findById(VorgangTestFactory.ID, Vorgang.class);
					assertThat(vorgang).isNotNull().extracting(Vorgang::getStatus).isEqualTo(Status.IN_BEARBEITUNG);
				});
			}

			@Test
			void shouldUpdateCommand() {
				callServiceRevokeCommand();

				await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
					var command = mongoOperations.findById(CommandTestFactory.ID, Command.class);
					assertThat(command).isNotNull().extracting(Command::getStatus).isEqualTo(CommandStatus.REVOKED);
				});
			}
		}

		void verifyVorgang() {
			var vorgang = mongoOperations.findById(VorgangTestFactory.ID, Vorgang.class);

			assertThat(vorgang).isNotNull();
			assertThat(vorgang.getStatus()).isEqualTo(Status.NEU);
		}

		void verifyCommand() {
			var command = mongoOperations.findById(CommandTestFactory.ID, Command.class);

			assertThat(command).isNotNull();
			assertThat(command.getStatus()).isEqualTo(CommandStatus.REVOKED);
			assertThat(command.getRelationId()).isEqualTo(VorgangTestFactory.ID);
			assertThat(command.getRelationVersion()).isEqualTo(VorgangTestFactory.VERSION);

			assertThat(command).isInstanceOf(PersistedCommand.class)
					.extracting("previousState")
					.asInstanceOf(InstanceOfAssertFactories.map(String.class, Object.class))
					.containsKey(Vorgang.MONGODB_FIELDNAME_STATUS).isEqualTo(Status.NEU);
		}

		private void callServiceRevokeCommand() {
			service.revokeCommand(request, responseObserver);

			verify(responseObserver).onNext(responseCaptor.capture());
		}
	}

}