/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.command;

import static org.assertj.core.api.Assertions.*;
import static org.awaitility.Awaitility.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.TimeUnit;

import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mapstruct.factory.Mappers;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.test.annotation.DirtiesContext;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandStatus;
import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItem;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItemTestFactory;
import de.ozgcloud.vorgang.callcontext.WithMockCustomUser;
import de.ozgcloud.vorgang.clientattribute.ClientAttributeReadPermitted;
import de.ozgcloud.vorgang.common.grpc.GrpcObjectMapper;
import de.ozgcloud.vorgang.grpc.command.CommandServiceGrpc.CommandServiceBlockingStub;
import de.ozgcloud.vorgang.grpc.command.GrpcAddSubCommandsRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcCommand;
import de.ozgcloud.vorgang.grpc.command.GrpcCreateCommand;
import de.ozgcloud.vorgang.grpc.command.GrpcCreateCommandRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcFindCommandsRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcRevokeCommandRequest;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangHeadTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import io.grpc.StatusRuntimeException;
import lombok.extern.log4j.Log4j2;
import net.devh.boot.grpc.client.inject.GrpcClient;

@SpringBootTest(properties = {
		"grpc.server.inProcessName=test",
		"grpc.client.inProcess.address=in-process:test"
})
@DirtiesContext
@DataITCase
@WithMockCustomUser
@Log4j2
class GrpcCommandServiceITCase {

	@GrpcClient("inProcess")
	private CommandServiceBlockingStub serviceBlockingStub;

	@Autowired
	private MongoOperations mongoOperations;

	@MockBean
	private ClientAttributeReadPermitted readIsPermitted;

	private final GrpcObjectMapper grpcObjectMapper = Mappers.getMapper(GrpcObjectMapper.class);

	@Nested
	class TestCreateCommand {

		private GrpcCreateCommandRequest request;

		@BeforeEach
		void prepareDatabase() {
			mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
			mongoOperations.dropCollection(Command.COLLECTION_NAME);
			mongoOperations.dropCollection(VorgangAttachedItem.COLLECTION_NAME);

			var vorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build());

			request = GrpcCreateCommandRequestTestFactory.createBuilder()
					.setVorgangId(vorgang.getId()).setRelationId(vorgang.getId())
					.setOrderString("CREATE_ATTACHED_ITEM")
					.setBodyObj(grpcObjectMapper.fromMap(VorgangAttachedItemTestFactory.asMap())).build();

			when(readIsPermitted.test(any())).thenReturn(true);
		}

		@Test
		void shouldPersistCommand() {
			serviceBlockingStub.createCommand(request);

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> {
				var commands = mongoOperations.findAll(Command.class);

				assertThat(commands).hasSize(1).first().satisfies(command -> {
					assertThat(command.getBodyObject()).isNotNull();
					assertThat(command.getBodyObject()).containsEntry("itemName", VorgangAttachedItemTestFactory.ITEM_NAME);
				});
			});
		}

		@Test
		void shouldExecuteCommand() {
			serviceBlockingStub.createCommand(request);

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> {
				var commands = mongoOperations.findAll(Command.class);

				assertThat(commands).hasSize(1).first().extracting(Command::getStatus).isEqualTo(CommandStatus.FINISHED);
			});
		}

		@Test
		void shouldNotPersistCommand() {
			var lockedVorgang = mongoOperations.save(buildLockedVorgang());
			var request = GrpcCreateCommandRequestTestFactory.createBuilder().setVorgangId(lockedVorgang.getId()).build();

			assertThrows(StatusRuntimeException.class, () -> serviceBlockingStub.createCommand(request));

			assertThat(mongoOperations.findAll(Command.class)).isEmpty();
		}

		Vorgang buildLockedVorgang() {
			return VorgangTestFactory.createBuilder().id(null).version(0).header(VorgangHeadTestFactory.create()).build();
		}
	}

	@Nested
	class TestRevokeCommand {

		@Test
		void shouldHaveSetStatus() {
			var vorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build());
			var command = mongoOperations.save(CommandTestFactory.createBuilder().id(null).status(CommandStatus.FINISHED).vorgangId(vorgang.getId())
					.previousState(Map.of(Vorgang.MONGODB_FIELDNAME_STATUS, "NEW")).build());

			var result = serviceBlockingStub.revokeCommand(GrpcRevokeCommandRequest.newBuilder().setId(command.getId()).build());

			assertThat(result.getCommand().getStatus()).isEqualTo(CommandStatus.REVOKE_PENDING.name());
		}

		@Test
		void shouldThrowExceptionIfCommandNotFound() {
			var revokeRequest = GrpcRevokeCommandRequest.newBuilder().setId("not-existing-id").build();

			assertThrows(StatusRuntimeException.class, () -> serviceBlockingStub.revokeCommand(revokeRequest));
		}

	}

	@Nested
	class TestLoadOutdatedOrders {

		@DisplayName("should load outdated orders")
		@ParameterizedTest(name = "for order {0}")
		@ValueSource(strings = { "CREATE_WIEDERVORLAGE", "EDIT_WIEDERVORLAGE", "WIEDERVORLAGE_ERLEDIGEN", "WIEDERVORLAGE_WIEDEREROEFFNEN",
				"CREATE_KOMMENTAR", "EDIT_KOMMENTAR" })
		void shouldLoadCommands(String order) {
			var vorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build());
			var command = mongoOperations.save(CommandTestFactory.createBuilder().id(null).vorgangId(vorgang.getId()).order(order).build());

			var response =
					serviceBlockingStub.findCommands(
							GrpcFindCommandsRequest.newBuilder().setVorgangId(vorgang.getId()).setOrderString(order).build());

			assertThat(response.getCommandList()).hasSize(1);
		}
	}

	@Nested
	class TestAddSubCommands {

		private static final String PARENT_TEST_ORDER = "PARENT_TEST_ORDER";

		private Vorgang vorgang;
		private GrpcCommand parentCommand;

		@BeforeEach
		void init() {
			mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
			mongoOperations.dropCollection(Command.COLLECTION_NAME);
			mongoOperations.dropCollection(VorgangAttachedItem.COLLECTION_NAME);

			vorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).version(0).build());
			when(readIsPermitted.test(any())).thenReturn(true);

			parentCommand = createParentCommand();
		}

		@Test
		void shouldCreateAndProcessCommands() {
			var request = buildAddSubCommandsRequest(parentCommand.getId(), buildCreateCommand(), buildCreateCommand());

			var grpcCommandsResponse = serviceBlockingStub.addSubCommands(request);

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				var command = loadCommand(parentCommand);
				assertThat(command.getStatus()).isEqualTo(CommandStatus.FINISHED);
			});
			grpcCommandsResponse.getCommandList().stream().map(this::loadCommand).forEach(command ->
					assertThat(command.getStatus()).isEqualTo(CommandStatus.FINISHED)
			);
		}

		@Test
		void shouldSetErrorStatusForParentCommand() {
			var request = buildAddSubCommandsRequest(parentCommand.getId(), buildCreateCommand(), buildFailedCreateCommand(), buildCreateCommand());

			serviceBlockingStub.addSubCommands(request);

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				var command = loadCommand(parentCommand);
				assertThat(command.getStatus()).isEqualTo(CommandStatus.ERROR);
				assertThat(command.getErrorMessage()).isNotBlank();
			});
		}

		@Test
		void shouldRevokeSubCommands() {
			var request = buildAddSubCommandsRequest(parentCommand.getId(), buildCreateCommand(), buildFailedCreateCommand(), buildCreateCommand());

			var grpcCommandsResponse = serviceBlockingStub.addSubCommands(request);

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				var revokedCommand = grpcCommandsResponse.getCommandList().stream().map(this::loadCommand)
						.filter(command -> command.getStatus() != CommandStatus.ERROR).toList();
				assertThat(revokedCommand).hasSize(2);
			});
		}

		@Test
		void shouldFailOnLockedVorgang() {
			mongoOperations.save(vorgang.toBuilder().header(VorgangHeadTestFactory.create()).build());
			var request = buildAddSubCommandsRequest(parentCommand.getId(), buildCreateCommand());

			Assertions.assertThrows(StatusRuntimeException.class, () -> serviceBlockingStub.addSubCommands(request));
		}

		private GrpcCommand createParentCommand() {
			var createParentCommandRequest = GrpcCreateCommandRequestTestFactory.createBuilder()
					.setVorgangId(vorgang.getId())
					.setRelationId(vorgang.getId())
					.setOrderString(PARENT_TEST_ORDER)
					.setBodyObj(grpcObjectMapper.fromMap(Map.of("key", "value"))).build();
			return serviceBlockingStub.createCommand(createParentCommandRequest).getCommand();
		}

		private GrpcAddSubCommandsRequest buildAddSubCommandsRequest(String parentCommandId, GrpcCreateCommand... subCommands) {
			var requestBuilder = GrpcAddSubCommandsRequest.newBuilder()
					.setParentId(parentCommandId)
					.setExecutionMode(SubCommandExecutionMode.PARALLEL.name())
					.setCompletedIfSubsCompleted(true);
			Arrays.stream(subCommands).forEach(requestBuilder::addSubCommands);
			return requestBuilder.build();
		}

		private GrpcCreateCommand buildFailedCreateCommand() {
			var bodyObject = new HashMap<>(VorgangAttachedItemTestFactory.asMap());
			bodyObject.put("vorgangId", vorgang.getId());
			return GrpcCreateCommand.newBuilder()
					.setOrder(Order.VORGANG_ABSCHLIESSEN.name())
					.setRelationId(vorgang.getId())
					.setRelationVersion(vorgang.getVersion() - 1)
					.setBodyObj(grpcObjectMapper.fromMap(bodyObject))
					.build();
		}

		private GrpcCreateCommand buildCreateCommand() {
			var bodyObject = new HashMap<>(VorgangAttachedItemTestFactory.asMap());
			bodyObject.put("vorgangId", vorgang.getId());
			return GrpcCreateCommand.newBuilder()
					.setOrder(Order.CREATE_ATTACHED_ITEM.name())
					.setRelationId(vorgang.getId())
					.setRelationVersion(vorgang.getVersion())
					.setBodyObj(grpcObjectMapper.fromMap(bodyObject))
					.build();
		}

		private Command loadCommand(GrpcCommand grpcCommand) {
			return mongoOperations.findById(grpcCommand.getId(), Command.class);
		}

	}
}
