/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.migration;

import static org.assertj.core.api.Assertions.*;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.util.Random;
import java.util.UUID;

import org.bson.Document;
import org.bson.types.ObjectId;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.gridfs.GridFsTemplate;
import org.springframework.data.mongodb.gridfs.GridFsUpload;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.files.OzgCloudFileTestUtil;

@DataITCase
class M013_CreateOzgCloudFileCollectionTest { // NOSONAR

	private static final String VORGANG_ID = UUID.randomUUID().toString();
	private static final String CLIENT = LoremIpsum.getInstance().getWords(1);
	private static final String FIELD_NAME = LoremIpsum.getInstance().getWords(1);
	private static final String FILE_NAME = LoremIpsum.getInstance().getWords(1);
	private static final String CONTENT_TYPE = LoremIpsum.getInstance().getWords(1);
	private static final String CREATED_BY = UUID.randomUUID().toString();
	private static final byte[] CONTENT = LoremIpsum.getInstance().getWords(11).getBytes();
	private static final long SIZE = CONTENT.length;

	@InjectMocks
	private M013_CreateOzgCloudFileCollection migration;

	@Autowired
	private MongoTemplate template;
	@Autowired
	private OzgCloudFileTestUtil ozgCloudFileTestUtil;
	@Autowired
	private GridFsTemplate gridFsTemplate;

	private String fileId;
	@BeforeEach
	void init() {
		ozgCloudFileTestUtil.dropFileCollections();
		fileId = uploadFile();
	}

	@Test
	void shouldMigrateMetadata() {
		migration.doMigration(template);

		assertThat(loadOzgCloudFile()).containsAllEntriesOf(createOzgCloudFileFromMetadata(fileId));
	}

	@Test
	void shouldLoadContentByContentId() {
		migration.doMigration(template);

		var ozgCloudFileDocument = loadOzgCloudFile();
		assertThat(getFileContentByContentId(ozgCloudFileDocument)).hasBinaryContent(CONTENT);
	}

	@Test
	void shouldSkipExistingFiles() {
		var ozgCloudFile = createNewOzgCloudFileFrom(fileId);
		template.save(ozgCloudFile, M013_CreateOzgCloudFileCollection.NEW_COLLECTION_NAME);

		migration.doMigration(template);

		var ozgCloudFiles = template.findAll(Document.class, M013_CreateOzgCloudFileCollection.NEW_COLLECTION_NAME);
		assertThat(ozgCloudFiles).hasSize(1).first().usingRecursiveComparison().isEqualTo(ozgCloudFile);
	}

	private String uploadFile() {
		var upload = GridFsUpload.fromStream(new ByteArrayInputStream(CONTENT)).metadata(createMetadata()).filename("test").build();
		return gridFsTemplate.store(upload).toHexString();
	}

	private Document createNewOzgCloudFileFrom(String fileId) {
		var ozgCloudFile = createMetadata();
		ozgCloudFile.append(M013_CreateOzgCloudFileCollection.FIELD_ID, new ObjectId(fileId));
		ozgCloudFile.append(M013_CreateOzgCloudFileCollection.FIELD_CONTENT_ID, fileId);
		ozgCloudFile.append(M013_CreateOzgCloudFileCollection.FIELD_VORGANG_ID, UUID.randomUUID().toString());
		ozgCloudFile.append(M013_CreateOzgCloudFileCollection.FIELD_CLIENT, LoremIpsum.getInstance().getWords(1));
		ozgCloudFile.append(M013_CreateOzgCloudFileCollection.FIELD_FIELD_NAME, LoremIpsum.getInstance().getWords(1));
		ozgCloudFile.append(M013_CreateOzgCloudFileCollection.FIELD_CONTENT_TYPE, LoremIpsum.getInstance().getWords(1));
		ozgCloudFile.append(M013_CreateOzgCloudFileCollection.FIELD_FILE_NAME, LoremIpsum.getInstance().getWords(1));
		ozgCloudFile.append(M013_CreateOzgCloudFileCollection.FIELD_CREATED_BY, UUID.randomUUID().toString());
		ozgCloudFile.append(M013_CreateOzgCloudFileCollection.FIELD_SIZE, new Random().nextLong());
		return ozgCloudFile;
	}

	private Document createOzgCloudFileFromMetadata(String fileId) {
		var ozgCloudFile = createMetadata();
		ozgCloudFile.append(M013_CreateOzgCloudFileCollection.FIELD_ID, new ObjectId(fileId));
		ozgCloudFile.append(M013_CreateOzgCloudFileCollection.FIELD_CONTENT_ID, fileId);
		return ozgCloudFile;
	}

	private Document createMetadata() {
		var metadata = new Document();
		metadata.append(M013_CreateOzgCloudFileCollection.FIELD_VORGANG_ID, VORGANG_ID);
		metadata.append(M013_CreateOzgCloudFileCollection.FIELD_CLIENT, CLIENT);
		metadata.append(M013_CreateOzgCloudFileCollection.FIELD_FIELD_NAME, FIELD_NAME);
		metadata.append(M013_CreateOzgCloudFileCollection.FIELD_CONTENT_TYPE, CONTENT_TYPE);
		metadata.append(M013_CreateOzgCloudFileCollection.FIELD_FILE_NAME, FILE_NAME);
		metadata.append(M013_CreateOzgCloudFileCollection.FIELD_CREATED_BY, CREATED_BY);
		metadata.append(M013_CreateOzgCloudFileCollection.FIELD_SIZE, SIZE);
		return metadata;
	}

	private Document loadOzgCloudFile() {
		return template.findById(fileId, Document.class, M013_CreateOzgCloudFileCollection.NEW_COLLECTION_NAME);
	}

	private InputStream getFileContentByContentId(Document ozgCloudFile) {
		return ozgCloudFileTestUtil.getFileContent(ozgCloudFile.getString(M013_CreateOzgCloudFileCollection.FIELD_CONTENT_ID));
	}

}