/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import static org.assertj.core.api.Assertions.*;

import java.util.List;
import java.util.Optional;
import java.util.UUID;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoOperations;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.vorgang.EingangTestFactory;
import de.ozgcloud.vorgang.vorgang.IncomingFile;
import de.ozgcloud.vorgang.vorgang.IncomingFileGroupTestFactory;
import de.ozgcloud.vorgang.vorgang.IncomingFileTestFactory;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

@DataITCase
class EingangFilesRepositoryITCase {

	@Autowired
	private EingangFilesRepository repository;

	@Autowired
	private MongoOperations mongoOperations;

	private final static FileId ID2 = FileId.createNew();

	@Nested
	class TestFindAttachments {

		@BeforeEach
		void init() {
			mongoOperations.dropCollection(Vorgang.class);
			mongoOperations.save(createVorgang());
		}

		@Nested
		class TestByVorgangId {

			@Test
			void shouldFindWhenOneEingang() {
				var vorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).version(0L).build());

				var result = findAttachments(vorgang.getId());

				assertThat(result).hasSize(1);
			}

			@Test
			void shouldFindWhenManyEingangs() {
				var vorang = mongoOperations.save(
						VorgangTestFactory.createBuilder().id(null).version(0L).clearEingangs()
								.eingang(EingangTestFactory.createBuilder().id(UUID.randomUUID().toString()).clearAttachments()
										.attachment(IncomingFileGroupTestFactory.createBuilder().clearFiles().files(createTestFiles()).build())
										.build())
								.eingang(EingangTestFactory.createBuilder().id(UUID.randomUUID().toString()).clearAttachments()
										.attachment(IncomingFileGroupTestFactory.create())
										.attachment(IncomingFileGroupTestFactory.create())
										.build())
								.build());

				var result = findAttachments(vorang.getId());

				assertThat(result).hasSize(4);
			}

			@Test
			void shouldReturnEmpty() {
				var vorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).version(0L).clearEingangs()
						.eingang(EingangTestFactory.createBuilder().clearAttachments().build()).build());

				var result = findAttachments(vorgang.getId());

				assertThat(result).isEmpty();
			}

			private List<IncomingFile> findAttachments(String vorgangId) {
				return repository.findAttachments(vorgangId, Optional.empty());
			}
		}

		@Nested
		class TestByVorgangIdAndEingangId {

			@Test
			void shouldReturnEmpty() {
				var vorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).version(0L).build());

				var result = repository.findAttachments(vorgang.getId(), Optional.of("not-existing"));

				assertThat(result).isEmpty();
			}

			@Test
			void shouldFindAttachments() {
				var eingangId = UUID.randomUUID().toString();
				var vorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).version(0L).clearEingangs()
						.eingang(EingangTestFactory.createBuilder().id(eingangId)
								.attachment(IncomingFileGroupTestFactory.createBuilder().clearFiles().files(createTestFiles()).build())
								.build())
						.build());

				var result = repository.findAttachments(vorgang.getId(), Optional.of(eingangId));

				assertThat(result).hasSize(3);
			}
		}
	}

	@Nested
	class TestFindRepresentations {

		private final Vorgang vorgangWithRepresentations = createVorgang();

		@BeforeEach
		void init() {
			mongoOperations.dropCollection(Vorgang.class);
			mongoOperations.save(vorgangWithRepresentations);
		}

		@Nested
		class TestByVorgangId {

			@Test
			void shouldFindWhenOneEingang() {
				var vorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).version(0L).build());

				var result = findRepresentations(vorgang.getId());

				assertThat(result).hasSize(1);
			}

			@Test
			void shouldFindWhenManyEingangs() {
				var vorang = mongoOperations.save(
						VorgangTestFactory.createBuilder().id(null).version(0L).clearEingangs()
								.eingang(EingangTestFactory.createBuilder().id(UUID.randomUUID().toString()).clearRepresentations()
										.representation(IncomingFileTestFactory.create())
										.representation(IncomingFileTestFactory.create())
										.build())
								.eingang(EingangTestFactory.createBuilder().id(UUID.randomUUID().toString()).clearRepresentations()
										.representation(IncomingFileTestFactory.create())
										.build())
								.build());

				var result = findRepresentations(vorang.getId());

				assertThat(result).hasSize(3);
			}

			@Test
			void shouldReturnEmpty() {
				var vorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).version(0L).clearEingangs()
						.eingang(EingangTestFactory.createBuilder().clearRepresentations().build()).build());

				var result = findRepresentations(vorgang.getId());

				assertThat(result).isEmpty();
			}

			private List<IncomingFile> findRepresentations(String vorgangId) {
				return repository.findRepresentations(vorgangId, Optional.empty());
			}
		}

		@Nested
		class TestByVorgangIdAndEingangId {

			@Test
			void shouldReturnEmpty() {
				var vorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).version(0L).build());

				var result = repository.findAttachments(vorgang.getId(), Optional.of("not-existing"));

				assertThat(result).isEmpty();
			}

			@Test
			void shouldFindRepresentations() {
				var eingangId = UUID.randomUUID().toString();
				var vorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).version(0L).clearEingangs()
						.eingang(EingangTestFactory.createBuilder().id(eingangId).representation(IncomingFileTestFactory.create()).build()).build());

				var result = repository.findRepresentations(vorgang.getId(), Optional.of(eingangId));

				assertThat(result).hasSize(2);
			}
		}
	}

	private Vorgang createVorgang() {
		return VorgangTestFactory.createBuilder().clearEingangs()
				.eingang(EingangTestFactory.createBuilder().clearAttachments().clearRepresentations()
						.attachments(List.of(IncomingFileGroupTestFactory.createBuilder().clearFiles().files(createTestFiles()).build()))
						.representations(createTestFiles()).build())
				.build();
	}

	private List<IncomingFile> createTestFiles() {
		return List.of(IncomingFileTestFactory.create(),
				IncomingFileTestFactory.createBuilder().id(ID2).build());
	}
}
