/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.vorgang.common.security.PolicyService;
import de.ozgcloud.vorgang.vorgang.EingangTestFactory;
import de.ozgcloud.vorgang.vorgang.IncomingFile;
import de.ozgcloud.vorgang.vorgang.IncomingFileTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import lombok.SneakyThrows;

class FileServiceTest {

	@Spy
	@InjectMocks
	private FileService service;

	@Mock
	private PolicyService policyService;
	@Mock
	private EingangFilesRepository repository;
	@Mock
	private OzgCloudFileRepository ozgCloudFileRepository;
	@Mock
	private FileIdMapper fileIdMapper;

	private final UploadedFilesReference ref = UploadedFilesReferenceTestFactory.create();
	private final OzgCloudFile file = OzgCloudFileTestFactory.create();
	private final InputStream contentStream = new ByteArrayInputStream(IncomingFileTestFactory.CONTENT);

	@DisplayName("Upload file")
	@Nested
	class TestUploadFile {

		private static final String SAVED_FILE_ID = UUID.randomUUID().toString();
		private final OzgCloudFile savedFile = OzgCloudFileTestFactory.createBuilder().id(SAVED_FILE_ID).build();

		@BeforeEach
		void init() {
			when(ozgCloudFileRepository.save(any())).thenReturn(savedFile);
			when(fileIdMapper.toFileId(any())).thenReturn(FileId.from(SAVED_FILE_ID));
		}

		@Test
		void shouldCallSave() {
			service.saveOzgCloudFile(file);

			verify(ozgCloudFileRepository).save(file);
		}

		@Test
		void shouldCallFileIdMapper() {
			service.saveOzgCloudFile(file);

			verify(fileIdMapper).toFileId(SAVED_FILE_ID);
		}

		@Test
		void shouldReturnFileId() {
			var result = service.saveOzgCloudFile(file);

			assertThat(result).hasToString(SAVED_FILE_ID);
		}

	}

	@DisplayName("upload filestream async")
	@Nested
	class TestUploadFileContentStreamAsync {

		@Test
		void shouldCallUploadFileContent() {
			service.uploadFileContentStreamAsync(ref, contentStream);

			verify(service).uploadFileContent(ref, contentStream);
		}

		@SneakyThrows
		@Test
		void shouldReturnFileId() {
			doReturn(OzgCloudFileTestFactory.CONTENT_ID).when(service).uploadFileContent(any(), any());

			var id = service.uploadFileContentStreamAsync(ref, contentStream);

			assertThat(id.get()).isEqualTo(OzgCloudFileTestFactory.CONTENT_ID);
		}
	}

	@DisplayName("Upload file")
	@Nested
	class TestUploadFileContent {

		@BeforeEach
		void init() {
			when(ozgCloudFileRepository.uploadContent(any(), any())).thenReturn(OzgCloudFileTestFactory.CONTENT_ID);
		}

		@Test
		void shouldCallAddContentSream() {
			service.uploadFileContent(ref, contentStream);

			verify(ozgCloudFileRepository).uploadContent(ref, contentStream);
		}

		@Test
		void shouldReturnId() {
			var id = service.uploadFileContent(ref, contentStream);

			assertThat(id).isEqualTo(OzgCloudFileTestFactory.CONTENT_ID);
		}
	}

	@Nested
	class TestFindFilesMetaData {

		private final Collection<FileId> ids = Collections.singleton(OzgCloudFileTestFactory.ID);

		@Test
		void shouldCallRepository() {
			findFilesMetaData();

			verify(ozgCloudFileRepository).findAll(ids);
		}

		@Test
		void shouldReturnFilesMetaData() {
			var file = OzgCloudFileTestFactory.create();
			when(ozgCloudFileRepository.findAll(any())).thenReturn(Stream.of(file));

			var result = findFilesMetaData();

			assertThat(result).containsExactly(file);
		}

		private List<OzgCloudFile> findFilesMetaData() {
			return service.findOzgCloudFiles(ids).toList();
		}
	}

	@Nested
	class TestDeleteAllByVorgang {
		@Test
		void shouldCallRepository() {
			service.deleteAllByVorgang(VorgangTestFactory.ID);

			verify(ozgCloudFileRepository).deleteAllByVorgang(VorgangTestFactory.ID);
		}
	}

	@Nested
	class TestGetAttachments {

		@Test
		void shouldCallCheckPermission() {
			getAttachments();

			verify(policyService).checkPermission(VorgangTestFactory.ID);
		}

		@Test
		void shouldCallFindAttachments() {
			getAttachments();

			verify(repository).findAttachments(VorgangTestFactory.ID, Optional.of(EingangTestFactory.ID));
		}

		@Test
		void shouldReturnAttachmentsByVorgangIdAndEingangId() {
			var attachment = IncomingFileTestFactory.create();
			when(repository.findAttachments(any(), any())).thenReturn(List.of(attachment));

			var result = getAttachments();

			assertThat(result).containsExactly(attachment);
		}

		private List<IncomingFile> getAttachments() {
			return service.getAttachments(VorgangTestFactory.ID, Optional.of(EingangTestFactory.ID));
		}
	}

	@Nested
	class TestGetRepresentations {

		@Test
		void shouldCallCheckPermission() {
			getRepresentations();

			verify(policyService).checkPermission(VorgangTestFactory.ID);
		}

		@Test
		void shouldCallFindRepresentations() {
			getRepresentations();

			verify(repository).findRepresentations(VorgangTestFactory.ID, Optional.of(EingangTestFactory.ID));
		}

		@Test
		void shouldReturnRepresentationsByVorgangIdAndEingangId() {
			var representation = IncomingFileTestFactory.create();
			when(repository.findRepresentations(any(), any())).thenReturn(List.of(representation));

			var result = getRepresentations();

			assertThat(result).containsExactly(representation);
		}

		private List<IncomingFile> getRepresentations() {
			return service.getRepresentations(VorgangTestFactory.ID, Optional.of(EingangTestFactory.ID));
		}
	}
}