/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.vorgang.grpc.file.GrpcGetAttachmentsResponse;
import de.ozgcloud.vorgang.grpc.file.GrpcGetRepresentationsResponse;
import de.ozgcloud.vorgang.grpc.file.GrpcOzgFile;
import de.ozgcloud.vorgang.vorgang.IncomingFile;
import de.ozgcloud.vorgang.vorgang.IncomingFileTestFactory;
import io.grpc.stub.StreamObserver;

class GrpcFileServiceTest {

	private static final GrpcOzgFile GRPC_OZG_FILE = GrpcOzgFileTestFactory.create();
	private static final IncomingFile INCOMING_FILE = IncomingFileTestFactory.create();

	@Spy
	@InjectMocks
	private GrpcFileService grpcFileService;

	@Mock
	private FileService fileService;

	@Mock
	private GrpcOzgFileMapper fileMapper;

	@Nested
	class TestGetAttachments {

		@Mock
		private GrpcGetAttachmentsResponse getAttachmentsResponse;
		@Mock
		private StreamObserver<GrpcGetAttachmentsResponse> responseObserver;

		@BeforeEach
		void init() {
			doReturn(getAttachmentsResponse).when(grpcFileService).buildAttachmentsResponse(anyList());
		}

		@Test
		void shouldCallGetAttachments() {
			getAttachments();

			verify(fileService).getAttachments(GrpcGetAttachmentsRequestTestFactory.VORGANG_ID,
					Optional.ofNullable(GrpcGetAttachmentsRequestTestFactory.EINGANG_ID));
		}

		@Test
		void shouldCallBuildAttachmentsResponse() {
			var files = List.of(INCOMING_FILE);
			when(fileService.getAttachments(any(), any())).thenReturn(files);

			getAttachments();

			verify(grpcFileService).buildAttachmentsResponse(files);
		}

		@Test
		void shouldAddFilesToResponse() {
			getAttachments();

			verify(responseObserver).onNext(getAttachmentsResponse);
		}

		@Test
		void shouldCallCompleted() {
			getAttachments();

			verify(responseObserver).onCompleted();
		}

		private void getAttachments() {
			grpcFileService.getAttachments(GrpcGetAttachmentsRequestTestFactory.create(), responseObserver);
		}
	}

	@Nested
	class TestBuildAttachmentsResponse {

		@BeforeEach
		void init() {
			when(fileMapper.map(any(IncomingFile.class))).thenReturn(GRPC_OZG_FILE);
		}

		@Test
		void shouldCallFileMapper() {
			when(fileMapper.map(INCOMING_FILE)).thenReturn(GRPC_OZG_FILE);

			buildAttachmentsResponse();

			verify(fileMapper).map(INCOMING_FILE);
		}

		@Test
		void shouldSetGrpcFile() {
			var result = buildAttachmentsResponse();

			assertThat(result.getFileList()).containsExactly(GRPC_OZG_FILE);
		}

		private GrpcGetAttachmentsResponse buildAttachmentsResponse() {
			return grpcFileService.buildAttachmentsResponse(List.of(INCOMING_FILE));
		}
	}

	@Nested
	class TestGetRepresentations {

		@Mock
		private GrpcGetRepresentationsResponse getRepresentationsResponse;
		@Mock
		private StreamObserver<GrpcGetRepresentationsResponse> responseObserver;

		@BeforeEach
		void init() {
			doReturn(getRepresentationsResponse).when(grpcFileService).buildRepresentationsResponse(any());
		}

		@Test
		void shouldCallGetRepresentations() {
			getRepresentations();

			verify(fileService).getRepresentations(GrpcGetRepresentationsRequestTestFactory.VORGANG_ID,
					Optional.of(GrpcGetRepresentationsRequestTestFactory.EINGANG_ID));
		}

		@Test
		void shouldCallBuildRepresentationsResponse() {
			var file = IncomingFileTestFactory.create();
			when(fileService.getRepresentations(any(), any())).thenReturn(List.of(file));

			getRepresentations();

			verify(grpcFileService).buildRepresentationsResponse(List.of(file));
		}

		@Test
		void shouldAddFilesToResponse() {
			getRepresentations();

			verify(responseObserver).onNext(getRepresentationsResponse);
		}

		@Test
		void shouldCallCompleted() {
			getRepresentations();

			verify(responseObserver).onCompleted();
		}

		private void getRepresentations() {
			grpcFileService.getRepresentations(GrpcGetRepresentationsRequestTestFactory.create(), responseObserver);
		}
	}

	@Nested
	class TestBuildRepresentationsResponse {

		@BeforeEach
		void init() {
			when(fileMapper.map(any(IncomingFile.class))).thenReturn(GRPC_OZG_FILE);
		}

		@Test
		void shouldCallFileMapper() {
			when(fileMapper.map(INCOMING_FILE)).thenReturn(GRPC_OZG_FILE);

			buildRepresentationsResponse();

			verify(fileMapper).map(INCOMING_FILE);
		}

		@Test
		void shouldSetGrpcFile() {
			var result = buildRepresentationsResponse();

			assertThat(result.getFileList()).containsExactly(GRPC_OZG_FILE);
		}

		private GrpcGetRepresentationsResponse buildRepresentationsResponse() {
			return grpcFileService.buildRepresentationsResponse(List.of(INCOMING_FILE));
		}
	}
}