/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.fail;
import static org.junit.jupiter.api.Assertions.*;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.TimeUnit;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.test.annotation.DirtiesContext;

import com.google.protobuf.ByteString;
import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.callcontext.TestCallContextAttachingInterceptor;
import de.ozgcloud.vorgang.grpc.binaryFile.BinaryFileServiceGrpc.BinaryFileServiceBlockingStub;
import de.ozgcloud.vorgang.grpc.binaryFile.BinaryFileServiceGrpc.BinaryFileServiceStub;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcBinaryFileMetadata;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcFindFilesResponse;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcGetBinaryFileDataRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUpdateBinaryFileRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUpdateBinaryFileResponse;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileMetaData;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileResponse;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import io.grpc.StatusRuntimeException;
import io.grpc.stub.StreamObserver;
import lombok.AccessLevel;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import lombok.SneakyThrows;
import net.devh.boot.grpc.client.inject.GrpcClient;

@SpringBootTest(properties = {
		"grpc.server.inProcessName=binary-file-test",
		"grpc.server.port=-1",
		"grpc.client.inProcess.address=in-process:binary-file-test"
})
@DirtiesContext
@DataITCase
public class OzgCloudFileITCase {

	@GrpcClient("inProcess")
	private BinaryFileServiceBlockingStub binaryFileServiceBlockingStub;
	@GrpcClient("inProcess")
	private BinaryFileServiceStub binaryFileServiceStub;

	@Autowired
	private MongoOperations mongoOperations;
	@Autowired
	private OzgCloudFileTestUtil ozgCloudFileTestUtil;

	@BeforeEach
	void init() {
		mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
		ozgCloudFileTestUtil.dropFileCollections();
	}

	@DisplayName("Find binary files meta data")
	@Nested
	class TestFindOzgCloudFilesMetaData {

		private String fileId;

		@BeforeEach
		void initDatabase() {
			fileId = ozgCloudFileTestUtil.saveOzgCloudFile(OzgCloudFileTestFactory.CONTENT).getId();
		}

		@Nested
		class OnMatchingOrganisationEinheitIdAtVorgang {

			@BeforeEach
			void initDatabase() {
				mongoOperations.save(VorgangTestFactory.create());
			}

			@Test
			void shouldReturnStoredFile() {
				var response = callFindBinaryFilesMetaData();

				assertThat(response.getFileList()).hasSize(1).first().satisfies(file -> {
					assertThat(file.getId()).isEqualTo(fileId);
					assertThat(file.getName()).isEqualTo(OzgCloudFileTestFactory.NAME);
					assertThat(file.getContentType()).isEqualTo(OzgCloudFileTestFactory.CONTENT_TYPE);
				});
			}
		}

		@Nested
		class OnNonMatchingOrganisationEinheitIdOnVorgang {

			@BeforeEach
			void initDatabase() {
				mongoOperations.save(VorgangTestFactory.createWithOrganisationEinheitId("73"));
			}

			@Test
			void shouldThrowStatusRuntimeException() {
				assertThrows(StatusRuntimeException.class, () -> callFindBinaryFilesMetaData());
			}
		}

		private GrpcFindFilesResponse callFindBinaryFilesMetaData() {
			return prepareBinaryFileServiceBlockingStub().findBinaryFilesMetaData(
					GrpcBinaryFilesRequestTestFactory.createBuilder().clearFileId().addFileId(fileId).build());
		}

	}

	@Nested
	@DisplayName("Download files of different sizes")
	class TestDownloadFile {

		@Nested
		class OnMatchingOrganisationEinheitIdAtVorgang {

			@BeforeEach
			void initDatabase() {
				mongoOperations.save(VorgangTestFactory.create());
			}

			@Test
			void shouldReturnStoredSmallFile() {
				var fileId = ozgCloudFileTestUtil.saveOzgCloudFile().getId();

				var file = downloadBinaryFile(fileId);

				assertThat(file).hasSameSizeAs(OzgCloudFileTestFactory.CONTENT).isEqualTo(OzgCloudFileTestFactory.CONTENT);
			}

			@Test
			void shouldReturnStoredFileWithSizeEqualChunkSize() {
				var content = OzgCloudFileTestFactory.createContentInByte(FileService.CHUNK_SIZE);
				var fileId = ozgCloudFileTestUtil.saveOzgCloudFile(content).getId();

				var file = downloadBinaryFile(fileId);

				assertThat(file).hasSameSizeAs(content).isEqualTo(content);
			}

			@Test
			void shouldReturnStoredLargeFile() {
				var input = OzgCloudFileTestFactory.createContentInMiB(5);
				var fileId = ozgCloudFileTestUtil.saveOzgCloudFile(input).getId();

				var downloadedFile = downloadBinaryFile(fileId);

				assertThat(downloadedFile).hasSameSizeAs(input).isEqualTo(input);
			}
		}

		@Nested
		class OnNonMatchingOrganisationEinheitIdAtVorgang {

			@BeforeEach
			void initDatabase() {
				mongoOperations.save(VorgangTestFactory.createWithOrganisationEinheitId("73"));
			}

			@Test
			void shouldThrowStatusRuntimeException() {
				var fileId = ozgCloudFileTestUtil.saveOzgCloudFile().getId();

				assertThrows(StatusRuntimeException.class, () -> downloadBinaryFile(fileId));
			}
		}

		private byte[] downloadBinaryFile(String fileId) {
			var request = GrpcGetBinaryFileDataRequest.newBuilder().setFileId(fileId).build();
			var responseIterator = prepareBinaryFileServiceBlockingStub().getBinaryFileContent(request);
			var content = new ByteArrayOutputStream();
			while (responseIterator.hasNext()) {
				ByteString chunkContent = responseIterator.next().getFileContent();
				try {
					content.write(chunkContent.toByteArray());
				} catch (IOException e) {
					fail(e);
				}
			}

			return content.toByteArray();
		}
	}

	private BinaryFileServiceBlockingStub prepareBinaryFileServiceBlockingStub() {
		return binaryFileServiceBlockingStub.withInterceptors(new TestCallContextAttachingInterceptor());
	}

	@Nested
	@DisplayName("Upload file")
	class TestFileUpload {

		private static final GrpcUploadBinaryFileMetaData REQUEST_METADATA = GrpcUploadBinaryFileMetaDataTestFactory.create();

		private UploadTestStreamObserver responseObserver = new UploadTestStreamObserver();

		@Nested
		class OnMatchingOrganisationEinheitIdAtVorgang {

			@BeforeEach
			void init() {
				mongoOperations.save(VorgangTestFactory.create());
			}

			@SneakyThrows
			@Test
			void shouldSaveVorgangId() {
				var fileId = uploadFileWithContent();

				var file = mongoOperations.findById(fileId, OzgCloudFile.class);
				assertThat(file.getVorgangId()).isEqualTo(VorgangTestFactory.ID);
			}

			@SneakyThrows
			@Test
			void shouldSaveFileContent() {
				var fileId = uploadFileWithContent();

				var file = mongoOperations.findById(fileId, OzgCloudFile.class);
				try (var fileStream = ozgCloudFileTestUtil.getFileContent(file.getContentId())) {
					byte[] fileContentFromDb = fileStream.readAllBytes();
					assertThat(fileContentFromDb).isEqualTo(OzgCloudFileTestFactory.CONTENT);
				}

			}
		}

		@Nested
		class OnMatchingNonOrganisationEinheitIdAtVorgang {

			@BeforeEach
			void init() {
				mongoOperations.save(VorgangTestFactory.createWithOrganisationEinheitId("73"));
			}

			@Test
			void shouldThrowStatusRuntimeException() {
				assertThatThrownBy(TestFileUpload.this::uploadFileWithContent).cause().isInstanceOf(StatusRuntimeException.class);
			}
		}

		@SneakyThrows
		private String uploadFileWithContent() {
			GrpcUploadBinaryFileRequest metadataRequest = GrpcUploadBinaryFileRequest.newBuilder().setMetadata(REQUEST_METADATA).build();
			GrpcUploadBinaryFileRequest contentRequest = GrpcUploadBinaryFileRequest.newBuilder()
					.setFileContent(ByteString.copyFrom(OzgCloudFileTestFactory.CONTENT)).build();

			StreamObserver<GrpcUploadBinaryFileRequest> observer = prepareBinaryFileServiceStub().uploadBinaryFileAsStream(responseObserver);

			observer.onNext(metadataRequest);
			observer.onNext(contentRequest);
			observer.onCompleted();

			return responseObserver.getFileIdFuture().get(10, TimeUnit.SECONDS);
		}

	}

	@Nested
	class TestUpdateOzgCloudFile {

		private static final String NEW_NAME = "new_name_" + LoremIpsum.getInstance().getWords(1);
		private static final String NEW_CONTENT_TYPE = "new_content_type_" + LoremIpsum.getInstance().getWords(1);
		private static final byte[] NEW_CONTENT = LoremIpsum.getInstance().getWords(10).getBytes();

		private UpdateTestStreamObserver responseStreamObserver = new UpdateTestStreamObserver();

		private OzgCloudFile ozgCloudFile;

		@BeforeEach
		void init() {
			mongoOperations.save(VorgangTestFactory.create());
			ozgCloudFile = ozgCloudFileTestUtil.saveOzgCloudFile(OzgCloudFileTestFactory.CONTENT);
		}

		@Test
		void shouldUpdateMetadata() {
			updateFileMetadata(ozgCloudFile.getId());

			var updatedFile = mongoOperations.findById(ozgCloudFile.getId(), OzgCloudFile.class);
			assertThat(updatedFile.getName()).isEqualTo(NEW_NAME);
			assertThat(updatedFile.getContentType()).isEqualTo(NEW_CONTENT_TYPE);
			assertThat(ozgCloudFileTestUtil.getFileContent(ozgCloudFile.getContentId())).hasBinaryContent(OzgCloudFileTestFactory.CONTENT);
		}

		@Test
		void shouldUpdateFileWithContent() {
			updateFileWithContent(ozgCloudFile.getId());

			var updatedFile = mongoOperations.findById(ozgCloudFile.getId(), OzgCloudFile.class);
			assertThat(updatedFile.getName()).isEqualTo(NEW_NAME);
			assertThat(updatedFile.getContentType()).isEqualTo(NEW_CONTENT_TYPE);
			assertThat(ozgCloudFileTestUtil.getFileContent(updatedFile.getContentId())).hasBinaryContent(NEW_CONTENT);
		}

		@Test
		void shouldThrowExceptionWhenFileNotFound() {
			assertThatThrownBy(() -> updateFileMetadata("not_existing_file_id")).cause().isInstanceOf(StatusRuntimeException.class);
		}

		@SneakyThrows
		private void updateFileMetadata(String fileId) {
			var observer = prepareBinaryFileServiceStub().updateBinaryFile(responseStreamObserver);

			observer.onNext(buildMetadataRequest(fileId));
			observer.onCompleted();

			responseStreamObserver.getFileIdFuture().get(10, TimeUnit.SECONDS);
		}

		@SneakyThrows
		private void updateFileWithContent(String fileId) {
			var contentRequest = GrpcUpdateBinaryFileRequest.newBuilder().setContent(ByteString.copyFrom(NEW_CONTENT)).build();

			var observer = prepareBinaryFileServiceStub().updateBinaryFile(responseStreamObserver);

			observer.onNext(buildMetadataRequest(fileId));
			observer.onNext(contentRequest);
			observer.onCompleted();

			responseStreamObserver.getFileIdFuture().get(10, TimeUnit.SECONDS);
		}

		private GrpcUpdateBinaryFileRequest buildMetadataRequest(String fileId) {
			var metadata = GrpcBinaryFileMetadata.newBuilder().setFileId(fileId).setFileName(NEW_NAME).setContentType(NEW_CONTENT_TYPE).build();
			return GrpcUpdateBinaryFileRequest.newBuilder().setMetadata(metadata).build();
		}
	}

	private BinaryFileServiceStub prepareBinaryFileServiceStub() {
		return binaryFileServiceStub.withInterceptors(new TestCallContextAttachingInterceptor());
	}

	@Getter
	@RequiredArgsConstructor(access = AccessLevel.PROTECTED)
	static class UploadTestStreamObserver implements StreamObserver<GrpcUploadBinaryFileResponse> {

		private final CompletableFuture<String> fileIdFuture = new CompletableFuture<>();
		private String fileId;

		@Override
		public void onNext(GrpcUploadBinaryFileResponse value) {
			fileId = value.getFileId();
		}

		@Override
		public void onError(Throwable t) {
			fileIdFuture.completeExceptionally(t);
		}

		@Override
		public void onCompleted() {
			fileIdFuture.complete(fileId);
		}
	}

	@Getter
	@RequiredArgsConstructor(access = AccessLevel.PROTECTED)
	static class UpdateTestStreamObserver implements StreamObserver<GrpcUpdateBinaryFileResponse> {

		private final CompletableFuture<Void> fileIdFuture = new CompletableFuture<>();

		@Override
		public void onNext(final GrpcUpdateBinaryFileResponse value) {
		}

		@Override
		public void onError(Throwable t) {
			fileIdFuture.completeExceptionally(t);
		}

		@Override
		public void onCompleted() {
			fileIdFuture.complete(null);
		}
	}

}