/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.status;

import static org.awaitility.Awaitility.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;
import java.util.concurrent.TimeUnit;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.context.ApplicationEventPublisher;

import de.ozgcloud.command.RevokeCommandEvent;
import de.ozgcloud.common.test.ITCase;
import de.ozgcloud.vorgang.command.CommandCreatedEventTestFactory;
import de.ozgcloud.vorgang.command.CommandService;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.command.Order;
import de.ozgcloud.vorgang.command.PersistedCommand;
import de.ozgcloud.vorgang.vorgang.VorgangService;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

@ITCase
class StatusEventListenerITCase {
	@Autowired
	private ApplicationEventPublisher publisher;

	@MockBean
	private StatusService statusService;

	@MockBean
	private VorgangService vorgangService;

	@MockBean
	private CommandService commandService;

	@BeforeEach
	void init() {
		when(vorgangService.getById(anyString())).thenReturn(VorgangTestFactory.create());

		when(commandService.findById(anyString())).thenReturn(Optional.of(CommandTestFactory.create()));
	}

	@DisplayName("Test creating status events")
	@Nested
	class TestStatusEvents {

		@Test
		void shouldCreateStatusChangeEventForVorgangAnnehmen() {
			publishCommandCreatedEvent(Order.VORGANG_ANNEHMEN);

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(statusService).annehmen(any()));
		}

		@Test
		void shouldCreateStatusChangeEventForVorgangAbschliessen() {
			publishCommandCreatedEvent(Order.VORGANG_ABSCHLIESSEN);

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(statusService).abschliessen(any()));
		}

		@Test
		void shouldCreateStatusChangeEventForVorgangBearbeiten() {
			publishCommandCreatedEvent(Order.VORGANG_BEARBEITEN);

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(statusService).bearbeiten(any()));
		}

		@Test
		void shouldCreateStatusChangeEventForVorgangBescheiden() {
			publishCommandCreatedEvent(Order.VORGANG_BESCHEIDEN);

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(statusService).bescheiden(any()));
		}

		@Test
		void shouldCreateStatusChangeEventForVorgangVerwerfen() {
			publishCommandCreatedEvent(Order.VORGANG_VERWERFEN);

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(statusService).verwerfen(any()));
		}

		@Test
		void shouldCreateStatusChangeEventForVorgangWiedereroeffnen() {
			publishCommandCreatedEvent(Order.VORGANG_WIEDEREROEFFNEN);

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(statusService).wiedereroeffnen(any()));
		}

		@Test
		void shouldCreateStatusChangeEventForVorgangZurueckholen() {
			publishCommandCreatedEvent(Order.VORGANG_ZURUECKHOLEN);

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(statusService).setStatusNeu(any()));
		}

		@Test
		void shouldCreateStatusChangeEventForVorgangZurueckstellen() {
			publishCommandCreatedEvent(Order.VORGANG_ZURUECKSTELLEN);

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(statusService).annehmen(any()));
		}

		@Nested
		class ZumLoeschenMarkieren {
			@Test
			void shouldCallService() {
				publishCommandCreatedEvent(Order.VORGANG_ZUM_LOESCHEN_MARKIEREN);

				await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(statusService).zumLoeschenMarkieren(any()));
			}
		}

		private void publishCommandCreatedEvent(Order order) {
			publisher.publishEvent(CommandCreatedEventTestFactory.create(
					CommandTestFactory.createBuilder().order(order.name()).build()));
		}
	}

	@Nested
	class TestRevokeStatusEvent {

		@BeforeEach
		void init() {
			when(vorgangService.getById(anyString())).thenReturn(VorgangTestFactory.create());
		}

		@Test
		void shouldSkipRevokeStatusEvents() {
			var command = CommandTestFactory.createBuilder().order(Order.REDIRECT_VORGANG.name()).build();

			publisher.publishEvent(new RevokeCommandEvent(command));

			verify(statusService, never()).revokeStatusChange(any(PersistedCommand.class));
		}

		@DisplayName("Test revoking status events")
		@ParameterizedTest(name = "should revoke status change event for {0}")
		@ValueSource(strings = { "VORGANG_ANNEHMEN", "VORGANG_ABSCHLIESSEN", "VORGANG_BEARBEITEN",
				"VORGANG_BESCHEIDEN", "VORGANG_VERWERFEN", "VORGANG_WIEDEREROEFFNEN", "VORGANG_ZURUECKHOLEN",
				"VORGANG_ZURUECKSTELLEN", "VORGANG_ZUM_LOESCHEN_MARKIEREN" })
		void shouldRevokeStatusChangeEvent(String orderName) {
			var command = CommandTestFactory.createBuilder().order(orderName).build();

			publisher.publishEvent(new RevokeCommandEvent(command));

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(statusService).revokeStatusChange(any(PersistedCommand.class)));
		}
	}
}
