/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.status;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.ConcurrentModificationException;
import java.util.Map;
import java.util.Optional;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.context.ApplicationEventPublisher;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandFailedEvent;
import de.ozgcloud.command.CommandRevokedEvent;
import de.ozgcloud.command.status.StatusChangedEvent;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.command.PersistedCommand;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.Vorgang.Status;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

class StatusServiceTest {

	@DisplayName("Test handling vorgang status changes")
	@Nested
	class TestVorgangStatusChanges {
		@Spy
		@InjectMocks
		private StatusService service;

		@Mock
		private StatusRepository repository;

		@Mock
		private ApplicationEventPublisher publisher;

		private PersistedCommand command = CommandTestFactory.create();

		@DisplayName("Test Vorgang ABSCHLIESSEN events")
		@Nested
		class OnVorgangAbschliessen {

			@Test
			void shouldHandleVorgangAbschliessen() {
				service.abschliessen(command);

				verify(repository).patch(anyString(), anyLong(), any());
			}

			@Test
			void shouldUpdateStatus() {
				service.abschliessen(command);

				verify(repository).patch(anyString(), anyLong(), eq(Map.of(Vorgang.MONGODB_FIELDNAME_STATUS, Vorgang.Status.ABGESCHLOSSEN.name())));
			}
		}

		@DisplayName("Test Vorgang ANNEHMEN events")
		@Nested
		class OnVorgangAnnehmen {
			@Test
			void shouldHandleVorgangAnnehmen() {
				service.annehmen(command);

				verify(repository).patch(anyString(), anyLong(), any());
			}

			@Test
			void shouldUpdateStatus() {
				service.annehmen(command);

				verify(repository).patch(anyString(), anyLong(), eq(Map.of(Vorgang.MONGODB_FIELDNAME_STATUS, Vorgang.Status.ANGENOMMEN.name())));
			}
		}

		@DisplayName("Test Vorgang IN_BEARBEITUNG events")
		@Nested
		class OnVorgangBearbieten {
			@Test
			void shouldHandleVorgangBearbeiten() {
				service.bearbeiten(command);

				verify(repository).patch(anyString(), anyLong(), any());
			}

			@Test
			void shouldUpdateStatus() {
				service.bearbeiten(command);

				verify(repository).patch(anyString(), anyLong(), eq(Map.of(Vorgang.MONGODB_FIELDNAME_STATUS, Vorgang.Status.IN_BEARBEITUNG.name())));
			}
		}

		@DisplayName("Test Vorgang BESCHIEDEN events")
		@Nested
		class OnVorgangBescheiden {
			@Test
			void shouldHandleVorgangBeschieden() {
				service.bescheiden(command);

				verify(repository).patch(anyString(), anyLong(), any());
			}

			@Test
			void shouldUpdateStatus() {
				service.bescheiden(command);

				verify(repository).patch(anyString(), anyLong(), eq(Map.of(Vorgang.MONGODB_FIELDNAME_STATUS, Vorgang.Status.BESCHIEDEN.name())));
			}
		}

		@DisplayName("Test Vorgang VERWORFEN events")
		@Nested
		class OnVorgangVerwerfen {
			@Test
			void shouldHandleVorgangVerworfen() {
				service.verwerfen(command);

				verify(repository).patch(anyString(), anyLong(), any());
			}

			@Test
			void shouldUpdateStatus() {
				service.verwerfen(command);

				verify(repository).patch(anyString(), anyLong(), eq(Map.of(Vorgang.MONGODB_FIELDNAME_STATUS, Vorgang.Status.VERWORFEN.name())));
			}
		}

		@DisplayName("Test Vorgang WIEDEREROEFFNEN events")
		@Nested
		class OnVorgangWiedereroeffnen {
			@Test
			void shouldHandleVorgangWiedereroeffnen() {
				service.wiedereroeffnen(command);

				verify(repository).patch(anyString(), anyLong(), any());
			}

			@Test
			void shouldUpdateStatus() {
				service.bearbeiten(command);

				verify(repository).patch(anyString(), anyLong(), eq(Map.of(Vorgang.MONGODB_FIELDNAME_STATUS, Vorgang.Status.IN_BEARBEITUNG.name())));
			}
		}

		@DisplayName("Test Vorgang ZURUECKHOLEN events")
		@Nested
		class OnVorgangZurueckholen {
			@Test
			void shouldHandleVorgangZurueckholen() {
				service.setStatusNeu(command);

				verify(repository).patch(anyString(), anyLong(), any());
			}

			@Test
			void shouldUpdateStatus() {
				service.setStatusNeu(command);

				verify(repository).patch(anyString(), anyLong(), eq(Map.of(Vorgang.MONGODB_FIELDNAME_STATUS, Vorgang.Status.NEU.name())));
			}
		}

		@DisplayName("Test Vorgang ZURUECKSTELLEN events")
		@Nested
		class OnVorgangZurueckstellen {
			@Test
			void shouldHandleVorgangZurueckStellen() {
				service.annehmen(command);

				verify(repository).patch(anyString(), anyLong(), any());
			}

			@Test
			void shouldUpdateStatus() {
				service.annehmen(command);

				verify(repository).patch(anyString(), anyLong(), eq(Map.of(Vorgang.MONGODB_FIELDNAME_STATUS, Vorgang.Status.ANGENOMMEN.name())));
			}
		}

		@DisplayName("Test Vorgang ZUM LOESCHEN MARKIEREN events")
		@Nested
		class OnVorgangZumLoeschenMarkieren {
			@Test
			void shouldHandleVorgangZumLoeschenMarkieren() {
				service.zumLoeschenMarkieren(command);

				verify(service).executeStatusChangeCommand(command, Status.ZU_LOESCHEN);
			}
		}

		@Nested
		@DisplayName("Changing status by command")
		class TestExecuteStatsuChangeByCommand {

			private Command command = CommandTestFactory.create();

			@Test
			@DisplayName("should trigger status change")
			void shouldTriggerUpdate() {
				service.executeStatusChangeCommand(command, Status.ANGENOMMEN);

				verify(service).doUpdateStatus(CommandTestFactory.RELATION_ID, CommandTestFactory.RELATION_VERSION, Status.ANGENOMMEN.toString());
			}

			@Test
			@DisplayName("should publish status changed event")
			void shouldPublishEvent() {
				service.executeStatusChangeCommand(command, Status.ANGENOMMEN);

				verify(publisher).publishEvent(any(StatusChangedEvent.class));
			}

			@Test
			@DisplayName("should publish failed event on error")
			void shouldPublishFailedEvent() {
				doThrow(ConcurrentModificationException.class).when(service).doUpdateStatus(anyString(), anyLong(), anyString());

				service.executeStatusChangeCommand(command, Status.ANGENOMMEN);

				verify(publisher).publishEvent(any(CommandFailedEvent.class));
			}
		}

	}

	@DisplayName("Test status changes when a Vorgang is forwarded")
	@Nested
	class TestVorgangWeiterleiten {
		@Spy
		@InjectMocks
		private StatusService service;

		@Mock
		private StatusRepository repository;

		@Mock
		private ApplicationEventPublisher publisher;

		@Mock
		private PersistedCommand command;

		@Test
		void shouldNotUpdateVersionOnWeitergeleitet() {
			service.setStatusToWeitergeleitet(VorgangTestFactory.ID, CommandTestFactory.ID);

			verify(repository).updateStatusWithoutVersionCheck(anyString(), eq(Vorgang.Status.WEITERGELEITET));
		}

		@Test
		void shouldNotUpdateVersionOnBearbeitet() {
			service.setStatusToInBearbeitung(VorgangTestFactory.ID, CommandTestFactory.ID);

			verify(repository).updateStatusWithoutVersionCheck(anyString(), eq(Vorgang.Status.IN_BEARBEITUNG));
		}

		@Test
		void shouldPatchOnInBearbeitung() {
			service.setStatusToInBearbeitung(VorgangTestFactory.ID, VorgangTestFactory.VERSION);

			verify(repository).patch(anyString(), anyLong(), eq(Map.of(Vorgang.MONGODB_FIELDNAME_STATUS, Vorgang.Status.IN_BEARBEITUNG.name())));
		}
	}

	@DisplayName("Test handling revoke command status changes")
	@Nested
	class TestRevokeStatusChanges {
		@Spy
		@InjectMocks
		private StatusService service;

		@Mock
		private StatusRepository repository;

		@Mock
		private ApplicationEventPublisher publisher;

		private static final String PREVIOUS_STATUS = Vorgang.Status.BESCHIEDEN.name();
		private PersistedCommand command = CommandTestFactory.createBuilder()
				.previousState(Map.of(Vorgang.MONGODB_FIELDNAME_STATUS, PREVIOUS_STATUS)).build();

		@Test
		void shouldUpdateVersion() {
			service.revokeStatusChange(command);

			verify(repository).patch(anyString(), eq(CommandTestFactory.RELATION_VERSION + 1), any());
		}

		@Test
		void shouldUpdateStatus() {
			service.revokeStatusChange(command);

			verify(repository).patch(anyString(), anyLong(), eq(Map.of(Vorgang.MONGODB_FIELDNAME_STATUS, Vorgang.Status.BESCHIEDEN.name())));
		}

		@Test
		void shouldPublishEvent() {
			service.revokeStatusChange(command);

			verify(publisher).publishEvent(any(CommandRevokedEvent.class));
		}

		@Test
		void shouldPublishFailEventOnConcurrentModification() {
			doThrow(ConcurrentModificationException.class).when(service).doUpdateStatus(any(), anyLong(), any());

			service.revokeStatusChange(command);

			verify(publisher).publishEvent(any(CommandFailedEvent.class));
		}

		@Test
		void shouldThrowTechnicalExceptionOnWrongPrevState() {
			doReturn(Optional.empty()).when(service).getPreviousStatus(any());

			assertThatThrownBy(() -> service.revokeStatusChange(command)).isInstanceOf(TechnicalException.class);
		}

		@Nested
		class TestCheckPreviousStatus {
			@Test
			void shouldReturnStatus() {
				var prevStatus = service.getPreviousStatus(command);

				assertThat(prevStatus).contains(PREVIOUS_STATUS);
			}

			@Test
			void shouldReturnEmptyOnMissingState() {
				var prevStatus = service.getPreviousStatus(CommandTestFactory.createBuilder().previousState(null).build());

				assertThat(prevStatus).isEmpty();
			}

			@Test
			void shouldReturnEmptyOnEmptyState() {
				var prevStatus = service.getPreviousStatus(CommandTestFactory.createBuilder().previousState(Map.of()).build());

				assertThat(prevStatus).isEmpty();
			}

			@Test
			void shouldReturnEmptyOnWrongStateField() {
				var prevStatus = service.getPreviousStatus(CommandTestFactory.create());

				assertThat(prevStatus).isEmpty();
			}
		}
	}
}
