/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import static de.ozgcloud.vorgang.common.grpc.GrpcSubFormTestFactory.*;
import static de.ozgcloud.vorgang.vorgang.LabelProcessor.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.vorgang.common.grpc.GrpcFormDataMapper;
import de.ozgcloud.vorgang.common.grpc.GrpcSubFormTestFactory;
import de.ozgcloud.vorgang.files.FileIdMapper;

class EingangMapperTest {

	@InjectMocks
	private EingangMapper mapper = Mappers.getMapper(EingangMapper.class);
	@Spy
	private IncomingFileMapper ozgFileMapper = Mappers.getMapper(IncomingFileMapper.class);
	@Spy
	private FileIdMapper fileIdMapper = Mappers.getMapper(FileIdMapper.class);
	@Spy
	private GrpcFormDataMapper grpcFormDataMapper = Mappers.getMapper(GrpcFormDataMapper.class);

	@Mock
	private AntragstellerMapper antragstellerMapper;
	@Mock
	private EingangHeaderMapper eingangHeaderMapper;

	@Nested
	class TestGrpcEingangToEingang {

		private final Antragsteller antragsteller = AntragstellerTestFactory.create();
		private final EingangHeader eingangHeader = EingangHeaderTestFactory.create();

		@BeforeEach
		void mockMapperReturnValues() {
			lenient().when(antragstellerMapper.fromGrpc(any())).thenReturn(antragsteller);
			lenient().when(eingangHeaderMapper.fromGrpc(any())).thenReturn(eingangHeader);

			ReflectionTestUtils.setField(ozgFileMapper, "fileIdMapper", fileIdMapper);
		}

		@Test
		void shouldMapIdToId() {
			GrpcEingang grpcEingang = GrpcEingangTestFactory.create();

			var eingang = mapper.fromGrpc(grpcEingang);

			assertThat(eingang.getId()).isEqualTo(GrpcEingangTestFactory.ID);
		}

		@Test
		void shouldHaveAntragsteller() {
			GrpcEingang grpcEingang = GrpcEingangTestFactory.create();

			var eingang = mapper.fromGrpc(grpcEingang);

			assertThat(eingang.getAntragsteller()).isEqualTo(antragsteller);
		}

		@Test
		void shouldHaveEingangHeader() {
			GrpcEingang grpcEingang = GrpcEingangTestFactory.create();

			var eingang = mapper.fromGrpc(grpcEingang);

			assertThat(eingang.getHeader()).isEqualTo(eingangHeader);
		}

		@Test
		void shouldProceedWithNullAndEmptyValues() {
			GrpcEingang grpcEingang = GrpcEingang.newBuilder().build();

			var eingang = mapper.fromGrpc(grpcEingang);

			assertThat(eingang.getId()).isEqualTo(StringUtils.EMPTY);
			assertThat(eingang.getAntragsteller()).isNull();
			assertThat(eingang.getHeader()).isNull();
		}

		@Test
		void shouldMapNumberOfAttachments() {

			GrpcEingang grpcEingang = GrpcEingangTestFactory.create();

			var eingang = mapper.fromGrpc(grpcEingang);

			assertThat(eingang.getNumberOfAttachments()).isEqualTo(4);
			assertThat(eingang.getNumberOfRepresentations()).isEqualTo(5);
		}

		@DisplayName("Mapped FormData")
		@Nested
		class TestMapSubFormToFormData {

			private GrpcEingang grpcEingang = GrpcEingangTestFactory.create();

			@Test
			void shouldNotBeEmpty() {
				var eingang = mapper.fromGrpc(grpcEingang);

				assertThat(eingang.getFormData()).isNotEmpty();
			}

			@Test
			void shouldHaveMappedSubForm() {
				var eingang = mapper.fromGrpc(grpcEingang);

				assertThat(eingang.getFormData()).containsKey(TITLE);
			}

			@DisplayName("mapped FormData should contain fields")
			@Test
			void fieldsInSubForm() {
				var eingang = mapper.fromGrpc(grpcEingang);

				Map<String, Object> formData = getSubForm(eingang);
				assertThat(formData).containsEntry(FIELD_NAME, FIELD);
			}

			@DisplayName("mapped FormData should contains subForm")
			@Test
			void shouldHaveMappedSubFormWithSubForm() {
				var eingang = mapper.fromGrpc(grpcEingang);

				Map<String, Object> formData = getSubForm(eingang);
				assertThat(formData).containsKey(GrpcSubFormTestFactory.SUBFORM_TITLE);
			}

			@SuppressWarnings("unchecked")
			private Map<String, Object> getSubForm(Eingang eingang) {
				Map<String, Object> formData = (Map<String, Object>) eingang.getFormData().get(TITLE);
				formData = (Map<String, Object>) formData.get(VALUE_KEY);
				return formData;
			}

			@SuppressWarnings("unchecked")
			@DisplayName("mapped SubForm should contain List of SubForms")
			@Test
			void shouldHaveMappedListOfSubForms() {
				GrpcEingang grpcEingang = GrpcEingangTestFactory.createBuilder().setFormData(GrpcFormData.newBuilder()
						.addForm(GrpcSubFormTestFactory.create())
						.addForm(GrpcSubFormTestFactory.create())
						.build()).build();

				var eingang = mapper.fromGrpc(grpcEingang);

				var formData = (Map<String, Object>) eingang.getFormData().get(TITLE);
				var formList = (List<Map<String, Object>>) formData.get(VALUE_KEY);

				assertThat(formList).contains(EingangTestFactory.FORM_DATA, EingangTestFactory.FORM_DATA);
			}

			@SuppressWarnings("unchecked")
			@DisplayName("mapped SubForm should contain field")
			@Test
			void subFormShouldContainField() {
				var eingang = mapper.fromGrpc(grpcEingang);

				Map<String, Object> subForm = (Map<String, Object>) getSubForm(eingang).get(SUBFORM_TITLE);
				subForm = (Map<String, Object>) subForm.get(VALUE_KEY);

				assertThat(subForm).containsEntry(SUBFORM_FIELD_NAME, SUBFORM_FIELD);
			}
		}

		@Nested
		@DisplayName("Test Attachment Mapping")
		class TestAttachmentsMapping {

			private Eingang eingang;

			@BeforeEach
			void init() {
				eingang = mapper.fromGrpc(GrpcEingangTestFactory.create());
			}

			@Test
			void testTotalAttachmentsCount() {

				assertThat(eingang.getNumberOfAttachments()).isEqualTo(4);
			}

			@Test
			void testAttachmentGroupCount() {

				assertThat(eingang.getAttachments()).isNotNull().isNotEmpty().hasSize(1);
			}

			@Test
			@DisplayName("Validate Number of Files in first AttachmentGroup")
			void validateAttachmentGroup1FilesCount() {

				assertThat(GrpcEingangTestFactory.create().getAttachments(0).getFilesCount()).isEqualTo(1);
				assertThat(eingang.getAttachments().get(0).getFiles()).hasSize(1);
			}

			@Test
			void validateGroup1File1() {

				IncomingFile attachment = eingang.getAttachments().get(0).getFiles().get(0);

				assertThat(attachment.getId()).isEqualTo(IncomingFileTestFactory.ID);
				assertThat(attachment.getName()).isEqualTo(IncomingFileTestFactory.NAME);
				assertThat(attachment.getSize()).isEqualTo(IncomingFileTestFactory.SIZE);
			}
		}

		@Nested
		@DisplayName("Test Representations Mapping")
		class TestRepresentationsMapping {

			private Eingang eingang;

			@BeforeEach
			void init() {

				eingang = mapper.fromGrpc(GrpcEingangTestFactory.create());
			}

			@Test
			void testRepresentationGroupCount() {

				assertThat(eingang.getRepresentations()).isNotNull().isNotEmpty().hasSize(1);
			}

			@Test
			void validateTotalNumberOfRepresentations() {

				assertThat(eingang.getNumberOfRepresentations()).isEqualTo(5);
			}

			@Test
			void testRepresentation1() {

				IncomingFile representation = eingang.getRepresentations().get(0);

				assertThat(representation.getId()).isEqualTo(IncomingFileTestFactory.ID);
				assertThat(representation.getName()).isEqualTo(IncomingFileTestFactory.NAME);
				assertThat(representation.getSize()).isEqualTo(IncomingFileTestFactory.SIZE);
			}
		}
	}

	@Nested
	class TestEingangToGrpcEingang {
		private final GrpcAntragsteller antragsteller = GrpcAntragstellerTestFactory.create();
		private final GrpcEingangHeader eingangHeader = GrpcEingangHeaderTestFactory.create();

		@BeforeEach
		void mockMapperReturnValues() {
			lenient().when(antragstellerMapper.toGrpc(any())).thenReturn(antragsteller);
			lenient().when(eingangHeaderMapper.toGrpc(any())).thenReturn(eingangHeader);
		}

		@Test
		void shouldMapIdToId() {
			Eingang eingang = EingangTestFactory.create();

			var grpcEingang = mapper.toGrpc(eingang);

			assertThat(grpcEingang.getId()).isEqualTo(EingangTestFactory.ID);
		}

		@Test
		void shouldHaveAntragsteller() {
			Eingang eingang = EingangTestFactory.create();

			var grpcEingang = mapper.toGrpc(eingang);

			assertThat(grpcEingang.getAntragsteller()).isEqualTo(antragsteller);
		}

		@Test
		void shouldHaveEingangHeader() {
			Eingang eingang = EingangTestFactory.create();

			var grpcEingang = mapper.toGrpc(eingang);

			assertThat(grpcEingang.getHeader()).isEqualTo(eingangHeader);
		}

		@Test
		void shouldProceedWithNullValues() {
			Eingang eingang = Eingang.builder().build();

			var grpcEingang = mapper.toGrpc(eingang);

			assertThat(grpcEingang.getId()).isEqualTo(StringUtils.EMPTY);
			assertThat(grpcEingang.hasAntragsteller()).isFalse();
			assertThat(grpcEingang.hasHeader()).isFalse();
		}

		@DisplayName("Mapped Eingang")
		@Nested
		class TestMapEingangToGrpcEingang {

			@DisplayName("contains formdata")
			@Nested
			class TestMappingFormData {

				private Eingang eingang = EingangTestFactory.create();

				@Test
				void shouldNotBeEmpty() {
					var grpcEingang = mapper.toGrpc(eingang);

					assertThat(grpcEingang.getFormData()).isNotNull();
				}

				@Test
				void shouldHaveFields() {
					var formData = mapper.toGrpc(eingang).getFormData();

					assertThat(formData.getFieldList()).hasSize(1);
				}

				@Test
				void shouldHaveMappedFormData() {
					var formData = mapper.toGrpc(eingang).getFormData();

					assertThat(formData.getFormList()).hasSize(1);
				}

				@Test
				void shouldHaveFormWithTitle() {
					var formData = mapper.toGrpc(eingang).getFormData();

					assertThat(formData.getForm(0).getTitle()).isEqualTo(EingangTestFactory.SUBFORM_NAME);
				}

				@Test
				void formDataShouldHaveFields() {
					var formData = mapper.toGrpc(eingang).getFormData();

					assertThat(formData.getFieldList()).hasSize(1);
					assertThat(formData.getField(0).getName()).isEqualTo(EingangTestFactory.SINGLE_FIELD_NAME);
					assertThat(formData.getField(0).getValue()).isEqualTo(EingangTestFactory.SINGLE_FIELD_VALUE);
				}

			}

		}
	}
}