/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import static de.ozgcloud.vorgang.common.grpc.GrpcResponseBatchStreamer.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.security.test.context.support.WithMockUser;

import de.ozgcloud.common.test.ITCase;
import de.ozgcloud.vorgang.common.grpc.GrpcFormDataTestFactory;
import de.ozgcloud.vorgang.common.grpc.GrpcFormFieldTestFactory;
import de.ozgcloud.vorgang.common.grpc.GrpcSubFormTestFactory;
import io.grpc.stub.StreamObserver;

@ITCase
class VorgangGrpcServiceITCase {

	static final String VALUE_KEY = "value";
	static final String LABEL_KEY = "label";

	private static final String NAME_FIELD_1 = "field1";
	private static final String VALUE_FIELD_1 = "value1";
	private static final String LABEL_FIELD_1 = "label1";
	private static final Map<String, Object> FIELD_1 = Map.of(LABEL_KEY, LABEL_FIELD_1, VALUE_KEY, VALUE_FIELD_1);

	private static final String NAME_FIELD_2 = "field2";
	private static final String VALUE_FIELD_2 = "value2";
	private static final String LABEL_FIELD_2 = "label2";
	private static final Map<String, Object> FIELD_2 = Map.of(LABEL_KEY, LABEL_FIELD_2, VALUE_KEY, VALUE_FIELD_2);

	private static final String NAME_FIELD_3 = "field3";
	private static final String VALUE_FIELD_3 = "value3";
	private static final String LABEL_FIELD_3 = "label3";
	private static final Map<String, Object> FIELD_3 = Map.of(LABEL_KEY, LABEL_FIELD_3, VALUE_KEY, VALUE_FIELD_3);

	private static final List<GrpcFormField> formFields = List.of(
			createFormField(NAME_FIELD_1, VALUE_FIELD_1, LABEL_FIELD_1),
			createFormField(NAME_FIELD_2, VALUE_FIELD_2, LABEL_FIELD_2),
			createFormField(NAME_FIELD_3, VALUE_FIELD_3, LABEL_FIELD_3));

	private static final String TITLE_SUBFORM_1 = "SUBFORM_1";
	private static final String TITLE_SUBFORM_2 = "SUBFORM_2";
	private static final String TITLE_SUBFORM_3 = "SUBFORM_3";

	@Autowired
	private VorgangGrpcService grpcVorgangService;
	@MockBean
	private VorgangService service;

	@Nested
	class TestCreateVorgang {
		@Mock
		private StreamObserver<GrpcCreateVorgangResponse> responseObserver;
		@Captor
		private ArgumentCaptor<Eingang> eingangCaptor;

		@BeforeEach
		void init() {
			when(service.startCreation(any())).thenReturn(VorgangTestFactory.create());
		}

		@Test
		void shouldKeepFieldsOrder() {
			var grpcFormData = GrpcFormData.newBuilder().addAllField(formFields).build();

			var formData = startCreation(grpcFormData);

			assertThat(removeControlData(formData)).containsExactly(
					entry(NAME_FIELD_1, FIELD_1),
					entry(NAME_FIELD_2, FIELD_2),
					entry(NAME_FIELD_3, FIELD_3));
		}

		@Test
		void shouldKeepSubFormOrder() {
			var grpcFormData = GrpcFormData.newBuilder().addForm(GrpcSubFormTestFactory.createBuilder().setTitle(TITLE_SUBFORM_1))
					.addForm(GrpcSubFormTestFactory.createBuilder().setTitle(TITLE_SUBFORM_2))
					.addForm(GrpcSubFormTestFactory.createBuilder().setTitle(TITLE_SUBFORM_3)).build();

			var formData = startCreation(grpcFormData);

			verifySubFormOrder(formData, TITLE_SUBFORM_1, TITLE_SUBFORM_2, TITLE_SUBFORM_3);
		}

		@SuppressWarnings("unchecked")
		@Test
		void shouldKeepFieldsOrderInSubForm() {
			var grpcFormData = GrpcFormDataTestFactory.createBuilder()
					.addForm(GrpcSubFormTestFactory.createBuilder().clearField().clearSubForm().setTitle(TITLE_SUBFORM_1).addAllField(formFields))
					.build();

			var formData = (Map<String, Object>) startCreation(grpcFormData).get(TITLE_SUBFORM_1);
			formData = (Map<String, Object>) formData.get(VALUE_KEY);

			assertThat(removeControlData(formData)).containsExactly(
					entry(NAME_FIELD_1, FIELD_1),
					entry(NAME_FIELD_2, FIELD_2),
					entry(NAME_FIELD_3, FIELD_3));
		}

		private LinkedHashMap<String, Object> removeControlData(Map<String, Object> formData) {
			var editable = new LinkedHashMap<>(formData);
			editable.remove("_kopControlData");
			return editable;
		}

		@SuppressWarnings("unchecked")
		@Test
		void shouldKeepOrderInSubForm() {
			var grpcFormData = GrpcFormDataTestFactory.createBuilder().addForm(
							GrpcSubFormTestFactory.createBuilder().setTitle(TITLE_SUBFORM_1).clearField().clearSubForm()
									.addSubForm(GrpcSubFormTestFactory.createBuilder().setTitle(TITLE_SUBFORM_3))
									.addSubForm(GrpcSubFormTestFactory.createBuilder().setTitle(TITLE_SUBFORM_2)))
					.build();

			var formData = (Map<String, Object>) startCreation(grpcFormData).get(TITLE_SUBFORM_1);
			formData = (Map<String, Object>) formData.get(VALUE_KEY);

			verifySubFormOrder(formData, TITLE_SUBFORM_3, TITLE_SUBFORM_2);
		}

		private static void verifySubFormOrder(Map<String, Object> formData, String... subFormTitle) {
			var entries = new ArrayList<>(formData.entrySet());
			for (int i = 0; i < subFormTitle.length; i++) {
				var entry = entries.get(i);
				assertThat(entry.getKey()).isEqualTo(subFormTitle[i]);
			}
		}

		private Map<String, Object> startCreation(GrpcFormData grpcFormData) {
			grpcVorgangService.startCreation(GrpcCreateVorgangRequestTestFactory.createWithFormData(grpcFormData), responseObserver);

			verify(service).startCreation(eingangCaptor.capture());
			return eingangCaptor.getValue().getFormData();

		}
	}

	private static GrpcFormField createFormField(String fieldName, String fieldValue, String label) {
		return GrpcFormFieldTestFactory.createBuilder().setName(fieldName).setValue(fieldValue).setLabel(label).build();
	}

	@Nested
	class TestLoadVorgang {

		private static final String REQUEST_ID = "request_id";

		@Mock
		private StreamObserver<GrpcFindVorgangWithEingangResponse> streamObserver;
		@Captor
		private ArgumentCaptor<GrpcFindVorgangWithEingangResponse> findVorgangResponseCaptor;

		@BeforeEach
		void initVorgang() {
			var formData = createFormData();

			var eingang = EingangTestFactory.createBuilder().formData(formData).build();
			var vorgang = VorgangTestFactory.createBuilder().clearEingangs().eingangs(List.of(eingang)).build();

			when(service.getById(any(), any())).thenReturn(vorgang);
		}

		@Test
		@WithMockUser
		void shouldKeepFieldOrder() {
			var grpcFormData = requestFormData();

			assertThat(grpcFormData).isEqualTo(expectedFormData());
		}

		private static Map<String, Object> createFormData() {
			Map<String, Object> formData = new LinkedHashMap<>();
			formData.put(NAME_FIELD_1, FIELD_1);
			formData.put(NAME_FIELD_2, FIELD_2);
			formData.put(NAME_FIELD_3, FIELD_3);
			for (String subFormName : List.of(TITLE_SUBFORM_1, TITLE_SUBFORM_2, TITLE_SUBFORM_3)) {
				var subForm = new LinkedHashMap<>();
				formData.put(subFormName, subForm);
				subForm.put(NAME_FIELD_1, FIELD_1);
				subForm.put(NAME_FIELD_2, FIELD_2);
				subForm.put(NAME_FIELD_3, FIELD_3);
			}
			return formData;
		}

		private static GrpcFormData expectedFormData() {
			var grpcFormData = GrpcFormDataTestFactory.createBuilder().clearField().addAllField(formFields);
			for (String subFormName : List.of(TITLE_SUBFORM_1, TITLE_SUBFORM_2, TITLE_SUBFORM_3)) {
				grpcFormData.addForm(GrpcSubForm.newBuilder()
						.setTitle(subFormName)
						.addAllField(formFields));
			}
			return grpcFormData.build();
		}

		private GrpcFormData requestFormData() {
			grpcVorgangService.findVorgangWithEingang(GrpcFindVorgangWithEingangRequest.newBuilder().setId(REQUEST_ID).build(), streamObserver);
			verify(streamObserver).onNext(findVorgangResponseCaptor.capture());
			return findVorgangResponseCaptor.getValue().getVorgangWithEingang().getEingang().getFormData();
		}
	}

	@Nested
	class TestFindDeletedVorgang {

		private static final int INCOMPLETE_BATCH_SIZE = DEFAULT_BATCH_SIZE - 1;
		private static final int DELETED_VORGANG_COUNT = DEFAULT_BATCH_SIZE + INCOMPLETE_BATCH_SIZE;

		@Mock
		private StreamObserver<GrpcFindDeletedVorgangResponse> streamObserver;
		@Captor
		private ArgumentCaptor<GrpcFindDeletedVorgangResponse> findVorgangResponseCaptor;

		@BeforeEach
		void init() {
			when(service.findDeleted()).thenReturn(generateVorgangStubs());
		}

		@Test
		void shouldSendResponses() {
			grpcVorgangService.findDeletedVorgang(GrpcFindDeletedVorgangRequestTestFactory.create(), streamObserver);

			verify(streamObserver, times(2)).onNext(findVorgangResponseCaptor.capture());
			var responses = findVorgangResponseCaptor.getAllValues();
			assertThat(responses).extracting(response -> response.getVorgangList().size()).containsExactly(DEFAULT_BATCH_SIZE, INCOMPLETE_BATCH_SIZE);
		}

		private Stream<VorgangStub> generateVorgangStubs() {
			return IntStream.range(0, DELETED_VORGANG_COUNT).mapToObj(idx -> VorgangStubTestFactory.create());
		}
	}
}
