/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;

import de.ozgcloud.common.errorhandling.TechnicalException;

class VorgangHeaderServiceTest {

	@InjectMocks
	@Spy
	private VorgangHeaderService service;
	@Mock
	private VorgangHeaderRepository repository;

	@Mock
	private VorgangService vorgangService;

	@Captor
	private ArgumentCaptor<FindVorgangRequest> requestCaptor;

	@DisplayName("Find all paged")
	@Nested
	class TestFindAllPaged {

		private static final long TOTAL = 10;

		@Mock
		private Page<VorgangHeader> pageMock;
		@Mock
		private Pageable pageableMock;

		@BeforeEach
		void mockPage() {
			when(pageMock.getPageable()).thenReturn(pageableMock);
			when(pageMock.getTotalElements()).thenReturn(TOTAL);
			when(pageMock.getContent()).thenReturn(Collections.singletonList(VorgangHeaderTestFactory.create()));
		}

		@Test
		void shouldRemoveUnpermittedClientAttributes() {
			when(repository.findAll(any(FindVorgangRequest.class))).thenReturn(pageMock);

			service.findAll(buildRequest(0, 10));

			verify(vorgangService).removeUnpermitted(notNull());
		}

		@Test
		void returnRepositoryResult() {
			when(repository.findAll(any(FindVorgangRequest.class))).thenReturn(pageMock);

			var result = service.findAll(buildRequest(10, 0));

			assertThat(result.getContent()).hasSize(1);
		}

		@Test
		void shouldKeepTotal() {
			when(repository.findAll(any(FindVorgangRequest.class))).thenReturn(pageMock);

			var result = service.findAll(buildRequest(0, 10));

			assertThat(result.getTotalElements()).isEqualTo(TOTAL);
		}

		@Test
		void shouldKeepPageable() {
			when(repository.findAll(any(FindVorgangRequest.class))).thenReturn(pageMock);

			var result = service.findAll(buildRequest(0, 10));

			assertThat(result.getPageable()).isSameAs(pageableMock);
		}

		private FindVorgangRequest buildRequest(int limit, int offset) {
			return FindVorgangRequest.builder().limit(limit).offset(offset).searchBy(StringUtils.EMPTY)
					.build();
		}
	}

	@Nested
	class TestFindById {

		@Test
		void shouldCallRepository() {
			service.findById(VorgangTestFactory.ID);

			verify(repository).findById(VorgangTestFactory.ID);
		}

		@Test
		void shouldReturnRepositoryResult() {
			var repositoryResult = Optional.of(VorgangHeaderTestFactory.create());
			when(repository.findById(VorgangTestFactory.ID)).thenReturn(repositoryResult);

			var serviceResult = service.findById(VorgangTestFactory.ID);

			assertThat(serviceResult).isEqualTo(repositoryResult);
		}
	}

	@Nested
	class TestGetById {

		@Test
		void shouldCallFindById() {
			doReturn(Optional.of(VorgangHeaderTestFactory.create())).when(service).findById(VorgangTestFactory.ID);

			service.getById(VorgangTestFactory.ID);

			verify(service).findById(VorgangTestFactory.ID);
		}

		@Test
		void shouldReturnVorgangHeader() {
			var expectedVorgangHeader = VorgangHeaderTestFactory.create();
			doReturn(Optional.of(expectedVorgangHeader)).when(service).findById(VorgangTestFactory.ID);

			var vorgangHeader = service.getById(VorgangTestFactory.ID);

			assertThat(vorgangHeader).isEqualTo(expectedVorgangHeader);
		}

		@Test
		@DisplayName("should throw exception if VorgangHeader could not be found")
		void shouldThrowException() {
			doReturn(Optional.empty()).when(service).findById(VorgangTestFactory.ID);

			assertThatThrownBy(() -> service.getById(VorgangTestFactory.ID))
					.isInstanceOf(TechnicalException.class)
					.hasMessageContaining(VorgangTestFactory.ID + " could not be found");
		}
	}
}