/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.time.ZonedDateTime;
import java.time.temporal.ChronoUnit;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.assertj.core.api.*;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.NullAndEmptySource;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.security.access.AccessDeniedException;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.VorgangAssignedEvent;
import de.ozgcloud.command.VorgangCreatedEvent;
import de.ozgcloud.command.VorgangLockedEvent;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.clientattribute.ClientAttributeReadPermitted;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.common.errorhandling.NotFoundException;
import de.ozgcloud.vorgang.servicekonto.ServiceKontoTestFactory;
import de.ozgcloud.vorgang.vorgang.redirect.ForwardingService;

class VorgangServiceTest {

	@Spy
	@InjectMocks
	private VorgangService service;

	@Mock
	private VorgangRepository repository;
	@Mock
	private ForwardingService redirectService;

	@Mock
	private ApplicationEventPublisher publisher;
	@Mock
	private AktenzeichenProvider aktenzeichenProvider;
	@Mock
	private VorgangAuthorizationService vorgangAuthorizationService;
	@Mock
	private ClientAttributeReadPermitted readIsPermitted;

	@Captor
	private ArgumentCaptor<Vorgang> vorgangCaptor;

	@Mock
	private LabelProcessor kopControlDataMapper;

	@Mock
	private VorgangStubMapper stubMapper;

	@DisplayName("Start creation")
	@Nested
	class TestStartCreation {

		private final Eingang eingang = EingangTestFactory.create();
		private static final String VORGANG_NUMMER = "TEST_VN";

		@BeforeEach
		void mockAktenzeichenProvider() {
			when(aktenzeichenProvider.provideAktenzeichen(any())).thenReturn(VorgangTestFactory.AKTENZEICHEN);

			when(kopControlDataMapper.moveLabelsToControlData(any(Eingang.class))).then(v -> v.getArgument(0));
		}

		@Test
		void shouldSetName() {
			var created = callStartCreation();

			assertThat(created.getName()).isEqualTo(EingangHeaderTestFactory.FORM_NAME);
		}

		@Test
		void shouldSetStatus() {
			var created = callStartCreation();

			assertThat(created.getStatus()).isEqualTo(Vorgang.Status.NEU);
		}

		@Test
		void shouldSetVorgangNummer() {
			doReturn(VORGANG_NUMMER).when(service).getVorgangNummer(any());

			var created = callStartCreation();

			assertThat(created.getNummer()).isEqualTo(VORGANG_NUMMER);
		}

		@DisplayName("eingang on vorgang")
		@Nested
		class TestEingang {

			@Test
			void shouldHaveFields() {
				var created = callStartCreation();

				assertThat(created.getEingangs().get(0)).usingRecursiveComparison().isEqualTo(eingang);
			}

			@DisplayName("with existing serviceKonto")
			@Nested
			class TestWithExistingServiceKonto {

				@Test
				void shouldHaveServiceKonto() {
					var vorgang = callStartCreation();

					assertThat(vorgang.getHeader().getServiceKonto()).usingRecursiveComparison().isEqualTo(ServiceKontoTestFactory.create());
				}
			}

			@DisplayName("without serviceKonto")
			@Nested
			class TestWithoutServiceKonto {

				private static final EingangHeader EINGANG_HEADER_WITHOUT_SERVICE_KONTO = EingangHeaderTestFactory.createBuilder().serviceKonto(null)
						.build();
				private static final Eingang EINGANG_WITHOUT_SERVICE_KONTO = EingangTestFactory.createBuilder()
						.header(EINGANG_HEADER_WITHOUT_SERVICE_KONTO).build();

				@Test
				void shouldNotHaveServiceKonto() {
					var vorgang = callStartCreation(EINGANG_WITHOUT_SERVICE_KONTO);

					assertThat(vorgang.getHeader().getServiceKonto()).isNull();
				}

			}

		}

		@Test
		void shouldSetFormEngineName() {
			var created = callStartCreation();

			assertThat(created.getFormEngineName()).isEqualTo(VorgangTestFactory.FORM_ENGINE_NAME);
		}

		@Test
		void shouldCallAktenzeichenProvider() {
			callStartCreation();

			verify(aktenzeichenProvider).provideAktenzeichen(eingang);
		}

		@Test
		void shouldSetAktenzeichen() {
			var created = callStartCreation();

			assertThat(created.getAktenzeichen()).isEqualTo(VorgangTestFactory.AKTENZEICHEN);
		}

		@Test
		void shouldSetInCreationTrue() {
			var created = callStartCreation();

			assertThat(created.isInCreation()).isTrue();
		}

		private Vorgang callStartCreation() {
			return callStartCreation(eingang);
		}

		private Vorgang callStartCreation(Eingang eingang) {
			service.startCreation(eingang);

			verify(repository).save(vorgangCaptor.capture());

			return vorgangCaptor.getValue();
		}
	}

	@Nested
	class getVorgangNummer {

		@Test
		void shouldReturnGivenNummer() {
			var result = service.getVorgangNummer(EingangHeaderTestFactory.create());

			assertThat(result).isEqualTo(EingangHeaderTestFactory.VORGANG_NUMMER);
		}

		@DisplayName("should return requestId")
		@ParameterizedTest(name = "when vorgangNummer {0}")
		@NullAndEmptySource
		void shouldReturnRequestId(String vorgangNummer) {
			var header = EingangHeaderTestFactory.createBuilder().vorgangNummer(vorgangNummer).build();

			var result = service.getVorgangNummer(header);

			assertThat(result).isEqualTo(EingangHeaderTestFactory.REQUEST_ID);
		}
	}

	@DisplayName("Finish creation")
	@Nested
	class TestFinishCreation {

		@Captor
		private ArgumentCaptor<VorgangCreatedEvent> eventCaptor;

		private final Eingang eingang = EingangTestFactory.createBuilder().clearAttachments().clearRepresentations().build();
		private final Vorgang vorgang = VorgangTestFactory.createBuilder().clearEingangs().eingang(eingang).inCreation(true).build();

		private final List<IncomingFileGroup> attachments = List.of(IncomingFileGroupTestFactory.create());
		private final List<IncomingFile> represenatations = List.of(IncomingFileTestFactory.create());

		@BeforeEach
		void mockRepository() {
			when(repository.findById(any())).thenReturn(Optional.of(vorgang));
			when(repository.save(any())).thenReturn(VorgangTestFactory.create());
		}

		@Test
		void shouldCallRepositoryToFindVorgang() {
			callFinishCreation();

			verify(repository).findById(VorgangTestFactory.ID);
		}

		@Test
		void shouldSetAttachments() {
			callFinishCreation();

			verify(repository).save(vorgangCaptor.capture());
			assertThat(vorgangCaptor.getValue().getEingangs().get(0).getAttachments()).isEqualTo(attachments);
		}

		@Test
		void shouldSetRepresenations() {
			callFinishCreation();

			verify(repository).save(vorgangCaptor.capture());
			assertThat(vorgangCaptor.getValue().getEingangs().get(0).getRepresentations()).isEqualTo(represenatations);
		}

		@Test
		void shouldSetInCreationFalse() {
			callFinishCreation();

			verify(repository).save(vorgangCaptor.capture());
			assertThat(vorgangCaptor.getValue().isInCreation()).isFalse();
		}

		@Test
		void shouldPublishVorgangCreatedEvent() {
			callFinishCreation();

			verify(publisher).publishEvent(eventCaptor.capture());
			assertThat(eventCaptor.getValue()).isInstanceOf(VorgangCreatedEvent.class)
					.extracting(VorgangCreatedEvent::getSource).isEqualTo(VorgangTestFactory.ID);
		}

		private void callFinishCreation() {
			service.finishCreation(VorgangTestFactory.ID, attachments, represenatations);
		}
	}

	@Nested
	class TestVorgangNotFound {
		@BeforeEach
		void mockRepository() {
			when(repository.findById(anyString())).thenReturn(Optional.of(VorgangTestFactory.create()));
		}

		@Test
		void shouldThrowException() {
			when(repository.findById(anyString())).thenReturn(Optional.empty());

			assertThrows(NotFoundException.class, () -> service.getById(VorgangTestFactory.ID));
		}
	}

	@Nested
	class TestGetById {

		@BeforeEach
		void mockAttributeUtilService() {
			when(readIsPermitted.test(any())).thenReturn(true);
		}

		@BeforeEach
		void mockRepository() {
			when(repository.findById(anyString())).thenReturn(Optional.of(VorgangTestFactory.create()));

			when(kopControlDataMapper.addLabels(any(Eingang.class))).then(v -> v.getArgument(0));
		}

		@Nested
		class AndOrganisationseinheitenIds {

			@BeforeEach
			void mockAttributeUtilService() {
				when(readIsPermitted.test(any())).thenReturn(true);
			}

			@Test
			void shouldCallFindById() {
				service.getById(VorgangTestFactory.ID);

				verify(repository).findById(VorgangTestFactory.ID);
			}

			@Test
			void shouldReturnVorgang() {
				var vorgang = service.getById(VorgangTestFactory.ID);

				assertThat(vorgang).isNotNull().usingRecursiveComparison().isEqualTo(VorgangTestFactory.create());
			}

			@Test
			void shouldThrowAccessDeniedException() {
				when(vorgangAuthorizationService.authorizeByOrganisationseinheitenId(any(), any())).thenReturn(false);

				assertThrows(AccessDeniedException.class, () -> service.getById(VorgangTestFactory.ID, FilterCriteriaTestFactory.create())); // NOSONAR
			}
		}

		@Test
		void shouldCallFindById() {
			service.getById(VorgangTestFactory.ID);

			verify(repository).findById(VorgangTestFactory.ID);
		}

		@Test
		void shouldReturnVorgang() {
			var vorgang = service.getById(VorgangTestFactory.ID);

			assertThat(vorgang).isNotNull().usingRecursiveComparison().isEqualTo(VorgangTestFactory.create());
		}

		@Test
		void shouldRemoveClientAttributesWithoutReadPermission() {
			when(readIsPermitted.test(any())).thenReturn(false);

			var loaded = service.getById(VorgangTestFactory.ID);

			assertThat(loaded.getClientAttributes()).isEmpty();
		}

		@Test
		void shouldContainClientAttributesWithReadPermission() {
			var loaded = service.getById(VorgangTestFactory.ID);

			assertThat(loaded.getClientAttributes()).isNotEmpty();
		}
	}

	@Nested
	class TestGetOriginalVorgangById {

		private final static Vorgang VORGANG = VorgangTestFactory.create();

		@BeforeEach
		void init() {
			doReturn(VORGANG).when(service).loadById(anyString());
		}

		@Test
		void shouldCallLoadById() {
			service.getOriginalVorgangById(VorgangTestFactory.ID);

			verify(service).loadById(VorgangTestFactory.ID);
		}

		@Test
		void shouldCallRemoveUnpermittedClientAttributes() {
			service.getOriginalVorgangById(VorgangTestFactory.ID);

			verify(service).removeUnpermittedClientAttributes(VORGANG);
		}

		@Test
		void shouldReturnVorgang() {
			var cleanedVorgang = VorgangTestFactory.create();
			doReturn(cleanedVorgang).when(service).removeUnpermittedClientAttributes(any());

			var result = service.getOriginalVorgangById(VorgangTestFactory.ID);

			assertThat(result).isSameAs(cleanedVorgang);
		}
	}

	@Nested
	class TestAssignToUser {

		private final Command command = CommandTestFactory.createBuilder().body(Map.of("assignedTo", "Klaus")).build();

		@Test
		void shouldCallRepository() {
			service.assignToUser(command);

			verify(repository).patch(CommandTestFactory.RELATION_ID, VorgangTestFactory.VERSION, Map.of("assignedTo", "Klaus"));
		}

		@Test
		void shouldPublishEvent() {
			service.assignToUser(command);

			verify(publisher).publishEvent(any(VorgangAssignedEvent.class));
		}
	}

	@DisplayName("Exists")
	@Nested
	class TestExists {

		private final List<String> organisationseinheitenIds = List.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID);

		@Test
		void shouldCallRepository() {
			service.exists(VorgangTestFactory.ID, organisationseinheitenIds);

			verify(repository).exists(VorgangTestFactory.ID, organisationseinheitenIds);
		}
	}

	@Nested
	class TestDeleteVorgang {

		private final Vorgang vorgang = VorgangTestFactory.create();
		private final VorgangStub stub = VorgangStubTestFactory.create();

		@BeforeEach
		void init() {
			when(repository.findById(any())).thenReturn(Optional.of(vorgang));
			when(stubMapper.fromVorgang(any())).thenReturn(stub);
		}

		@Test
		void shouldLoadVorgang() {
			service.deleteVorgang(VorgangTestFactory.ID);

			verify(service).loadById(VorgangTestFactory.ID);
		}

		@Test
		void shouldMapToStub() {
			service.deleteVorgang(VorgangTestFactory.ID);

			verify(stubMapper).fromVorgang(vorgang);
		}

		@Test
		void shouldPersistStub() {
			service.deleteVorgang(VorgangTestFactory.ID);

			verify(repository).saveStub(stub);
		}
	}

	@Nested
	class TestSetAktenzeichen {

		@Test
		void shouldCallRepository() {
			var command = CommandTestFactory.createBuilder()
					.bodyObject(Map.of(VorgangService.BODY_OBJECT_AKTENZEICHEN, VorgangTestFactory.AKTENZEICHEN)).build();

			service.setAktenzeichen(command);

			verify(repository).setAktenzeichen(VorgangTestFactory.ID, CommandTestFactory.RELATION_VERSION, VorgangTestFactory.AKTENZEICHEN);
		}

		@Test
		void shouldTrimEmptyStringToNull() {
			var command = CommandTestFactory.createBuilder().bodyObject(Map.of(VorgangService.BODY_OBJECT_AKTENZEICHEN, "")).build();

			service.setAktenzeichen(command);

			verify(repository).setAktenzeichen(VorgangTestFactory.ID, CommandTestFactory.RELATION_VERSION, null);
		}
	}

	@Nested
	class TestLockVorgang {

		private static final Command LOCK_COMMAND = CommandTestFactory.create();
		private static final Map<String, Object> LOCK_PATCH = Map.of("key", "value");

		@Captor
		private ArgumentCaptor<VorgangLockedEvent> eventCaptor;

		@BeforeEach
		void init() {
			doNothing().when(service).validateLockCommand(any());
			doReturn(LOCK_PATCH).when(service).buildLockPatch(any());
		}

		@Test
		void shouldCallValidateLockCommand() {
			service.lockVorgang(LOCK_COMMAND);

			verify(service).validateLockCommand(LOCK_COMMAND);
		}

		@Test
		void shouldCallBuildLockPatch() {
			service.lockVorgang(LOCK_COMMAND);

			verify(service).buildLockPatch(LOCK_COMMAND);
		}

		@Test
		void shouldCallRepository() {
			service.lockVorgang(LOCK_COMMAND);

			verify(repository).patch(VorgangTestFactory.ID, CommandTestFactory.RELATION_VERSION, LOCK_PATCH);
		}

		@Test
		void shouldCallPublishEvent() {
			service.lockVorgang(LOCK_COMMAND);

			verify(publisher).publishEvent(eventCaptor.capture());
			assertThat(eventCaptor.getValue().getCommand()).isSameAs(LOCK_COMMAND);
		}
	}

	@Nested
	class TestValidateLockCommand {

		@Test
		void shouldThrowExceptionOnMissingClient() {
			var command = CommandTestFactory.createBuilder().createdByClientName(null).build();

			assertThrows(TechnicalException.class, () -> service.validateLockCommand(command));
		}

		@Test
		void shouldThrowExceptionOnMissingReason() {
			var command = CommandTestFactory.createBuilder().bodyObject(Map.of()).build();

			assertThrows(TechnicalException.class, () -> service.validateLockCommand(command));
		}

		@Test
		void shouldAcceptValidCommand() {
			var command = CommandTestFactory.createBuilder().bodyObject(Map.of(Lock.FIELD_REASON, LockTestFactory.REASON)).build();

			assertDoesNotThrow(() -> service.validateLockCommand(command));
		}
	}

	@Nested
	class TestBuildLockPatch {

		private final Command command = CommandTestFactory.createBuilder().bodyObject(Map.of(Lock.FIELD_REASON, LockTestFactory.REASON)).build();

		@Test
		void shouldSetClientName() {
			var result = service.buildLockPatch(command);

			assertThat(result).extractingByKey(VorgangService.KEY_HEADER_LOCK, MAP)
					.containsEntry(Lock.FIELD_CLIENT_NAME, CommandTestFactory.CREATED_BY_CLIENT);
		}

		@Test
		void shouldSetLockedSince() {
			var result = service.buildLockPatch(command);

			assertThat(result).extractingByKey(VorgangService.KEY_HEADER_LOCK, MAP)
					.extractingByKey(Lock.FIELD_LOCKED_SINCE, InstanceOfAssertFactories.ZONED_DATE_TIME)
					.isCloseTo(ZonedDateTime.now(), within(1, ChronoUnit.SECONDS));
		}

		@Test
		void shouldSetReason() {
			var result = service.buildLockPatch(command);

			assertThat(result).extractingByKey(VorgangService.KEY_HEADER_LOCK, MAP).containsEntry(Lock.FIELD_REASON, LockTestFactory.REASON);
		}
	}

	@Nested
	class TestIsVorgangLocked {

		@Test
		void shouldCallRepository() {
			service.isVorgangLocked(VorgangTestFactory.ID);

			verify(repository).isVorgangLocked(VorgangTestFactory.ID);
		}

		@Test
		void shouldReturnResult() {
			when(repository.isVorgangLocked(anyString())).thenReturn(true);

			var result = service.isVorgangLocked(VorgangTestFactory.ID);

			assertThat(result).isTrue();
		}
	}

	@Nested
	class TestUnlockVorgang {

		private final Command command = CommandTestFactory.create();

		@BeforeEach
		void mock() {
			doNothing().when(service).validateUnlockingClient(any());
		}

		@Test
		void shouldCallValidateUnlockingClient() {
			unlockVorgang();

			verify(service).validateUnlockingClient(command);
		}

		@Test
		void shouldCallRepository() {
			unlockVorgang();

			verify(repository).removeLock(VorgangTestFactory.ID, CommandTestFactory.RELATION_VERSION);
		}

		private void unlockVorgang() {
			service.unlockVorgang(command);
		}
	}

	@Nested
	class TestValidateUnlockingClient {

		private final Vorgang vorgang = VorgangTestFactory.createBuilder().header(VorgangHeadTestFactory.create()).build();
		private final Command command = CommandTestFactory.create();

		@Test
		void shouldGetVorgangFromRepository() {
			givenRepositoryReturnsVorgang();

			validateUnlockingClient();

			verify(repository).findById(VorgangTestFactory.ID);
		}

		@Test
		void shouldThrowNotFoundException() {
			when(repository.findById(VorgangTestFactory.ID)).thenReturn(Optional.empty());

			assertThrows(NotFoundException.class, () -> validateUnlockingClient());
		}

		@Test
		void shouldCallValidateClientName() {
			givenRepositoryReturnsVorgang();

			validateUnlockingClient();

			verify(service).validateClientName(CommandTestFactory.CREATED_BY_CLIENT, Optional.of(VorgangHeadTestFactory.LOCK));
		}

		private void givenRepositoryReturnsVorgang() {
			when(repository.findById(VorgangTestFactory.ID)).thenReturn(Optional.of(vorgang));
		}

		private void validateUnlockingClient() {
			service.validateUnlockingClient(command);
		}
	}

	@Nested
	class TestValidateClientName {

		private final Lock lock = LockTestFactory.create();

		@Test
		void shouldCallHasNoPermissionToUnlock() {
			doReturn(false).when(service).hasNoPermissionToUnlock(any(), any());

			validateClientName();

			verify(service).hasNoPermissionToUnlock(CommandTestFactory.CREATED_BY_CLIENT, Optional.of(lock));
		}

		@Test
		void shouldThrowAccessDeniedException() {
			doReturn(true).when(service).hasNoPermissionToUnlock(any(), any());

			assertThrows(AccessDeniedException.class, () -> validateClientName());
		}

		@Test
		void shouldNotThrowException() {
			doReturn(false).when(service).hasNoPermissionToUnlock(any(), any());

			assertDoesNotThrow(() -> validateClientName());
		}

		private void validateClientName() {
			service.validateClientName(CommandTestFactory.CREATED_BY_CLIENT, Optional.of(lock));
		}
	}

	@Nested
	class TestHasNoPermissionToUnlock {

		@Test
		void shouldReturnFalseIfLockEmpty() {
			var result = service.hasNoPermissionToUnlock(CommandTestFactory.CREATED_BY_CLIENT, Optional.empty());

			assertThat(result).isFalse();
		}

		@Test
		void shouldReturnTrueIfLockPresentAndClientNameNotMatching() {
			var result = service.hasNoPermissionToUnlock("wrong name", Optional.of(LockTestFactory.create()));

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnFalseIfLockPresentAndClientNameMatching() {
			var result = service.hasNoPermissionToUnlock(LockTestFactory.CLIENT_NAME, Optional.of(LockTestFactory.create()));

			assertThat(result).isFalse();
		}
	}


	@Nested
	class TestFindDeleted {

		@Test
		void shouldCallRepository() {
			service.findDeleted();

			verify(repository).findDeleted();
		}

		@Test
		void shouldReturnRepositoryResult() {
			var repositoryResult = List.of(VorgangStubTestFactory.create());
			when(repository.findDeleted()).thenReturn(repositoryResult.stream());

			var serviceResult = service.findDeleted().toList();

			assertThat(serviceResult).isEqualTo(repositoryResult);
		}
	}
}