/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang.redirect;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.command.Command;
import de.ozgcloud.nachrichten.email.MailRecipient;
import de.ozgcloud.nachrichten.email.MailSendRequest;
import de.ozgcloud.nachrichten.email.MailSendRequest.MailAttachment;
import de.ozgcloud.nachrichten.email.MailService;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.status.StatusService;
import de.ozgcloud.vorgang.vorgang.AntragstellerTestFactory;
import de.ozgcloud.vorgang.vorgang.EingangHeaderTestFactory;
import de.ozgcloud.vorgang.vorgang.EingangTestFactory;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangService;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import de.ozgcloud.vorgang.vorgang.ZustaendigeStelleTestFactory;

class ForwardingServiceTest {

	@Spy
	@InjectMocks
	private ForwardingService service;
	@Mock
	private ForwardingRepository repository;

	@Mock
	private ZipBuilderService zipService;

	@Mock
	private MailService mailService;

	@Mock
	private VorgangService vorgangService;

	@Mock
	private StatusService statusService;

	@Mock
	private ApplicationEventPublisher publisher;

	@Nested
	class TestBuildForwarding {

		@Test
		void shouldBuildFromCommand() {
			var forwarding = service.buildForwarding(CommandTestFactory.createBuilder().redirectRequest(RedirectRequestTestFactory.create()).build());

			assertThat(forwarding.getStatus()).isEqualTo(ForwardingTestFactory.STATUS);
		}
	}

	@Nested
	class TestFillingTemplate {

		@Nested
		class TestBuildValues {

			private Vorgang vorgang = VorgangTestFactory.create();

			@Test
			void shouldCreateMap() {
				var map = service.buildValues(vorgang);

				assertThat(map).isNotNull().isInstanceOf(Map.class);
			}

			@Test
			void shouldFillAntragsName() {
				var map = service.buildValues(vorgang);

				assertThat(map).containsEntry(ForwardingService.FIELD_ANTRAGS_NAME, EingangHeaderTestFactory.FORM_NAME);
			}

			@Test
			void shouldFillEingangsKennzeichen() {
				var map = service.buildValues(vorgang);

				assertThat(map).containsEntry(ForwardingService.FIELD_EINGANGS_KENNZEICHEN, EingangHeaderTestFactory.REQUEST_ID);
			}

			@Test
			void shouldFillAktenzeichen() {
				var map = service.buildValues(vorgang);

				assertThat(map).containsEntry(ForwardingService.FIELD_AKTENZEICHEN, vorgang.getAktenzeichen());
			}

			@Test
			void shouldFillZustaendigeStelle() {
				var map = service.buildValues(vorgang);

				assertThat(map).containsEntry(ForwardingService.FIELD_ZUSTAENDIGE_STELLE, ZustaendigeStelleTestFactory.BEZEICHNUNG);
			}
		}
	}

	@DisplayName("Create forward mail")
	@Nested
	class TestCreateForwardMail {

		private static final String MAIL_FROM = "FROM ME WITH L";
		private static final String BODY = "FILLED BODY";
		private final Optional<String> replyTo = Optional.of("TestReplyTo");

		private Vorgang vorgang = VorgangTestFactory.create();
		private Forwarding forwarding = ForwardingTestFactory.create();

		@BeforeEach
		void initTest() {
			ReflectionTestUtils.setField(service, "mailFrom", MAIL_FROM);
			ReflectionTestUtils.setField(service, "replyTo", replyTo);

			doReturn(BODY).when(service).fillMailTemplate(any());
			doReturn(mock(MailAttachment.class)).when(service).buildZipAttachment(any(Vorgang.class),
					any(), anyBoolean());
		}

		@Test
		void shouldSetFromAddress() {
			var mailSendRequest = service.createForwardMail(vorgang, forwarding, RedirectRequestTestFactory.PASSWORD.toCharArray());

			assertThat(mailSendRequest.getFromAddress()).isEqualTo(MAIL_FROM);
		}

		@Test
		void shouldSetToAddress() {
			var mailSendRequest = service.createForwardMail(vorgang, forwarding, RedirectRequestTestFactory.PASSWORD.toCharArray());

			var toEmails = mailSendRequest.getToAddresses().stream().map(MailRecipient::email).toList();
			assertThat(toEmails).containsExactly(RedirectRequestTestFactory.EMAIL);
		}

		@Test
		void shouldSetBody() {
			var mailSendRequest = service.createForwardMail(vorgang, forwarding, RedirectRequestTestFactory.PASSWORD.toCharArray());

			verify(service).fillMailTemplate(vorgang);
			assertThat(mailSendRequest.getBody()).isEqualTo(BODY);
		}

		@Test
		void shouldEntcrypt() {
			createForwardMail();

			verify(service).shouldEntcrypt(RedirectRequestTestFactory.PASSWORD.toCharArray());
		}

		private MailSendRequest createForwardMail() {
			return service.createForwardMail(vorgang, forwarding, RedirectRequestTestFactory.PASSWORD.toCharArray());
		}
	}

	@DisplayName("entcrypt")
	@Nested
	class TestEntcrypt {

		@Test
		void shouldBeTrueOnExistingPassword() {
			var entcrypt = service.shouldEntcrypt(RedirectRequestTestFactory.PASSWORD.toCharArray());

			assertThat(entcrypt).isTrue();
		}

		@Test
		void shouldBeFalseOnEmptyPassword() {
			var entcrypt = service.shouldEntcrypt(new char[0]);

			assertThat(entcrypt).isFalse();
		}

		@Test
		void shouldBeFalseOnNull() {
			var entcrypt = service.shouldEntcrypt(null);

			assertThat(entcrypt).isFalse();
		}
	}

	@Nested
	class TestOnRedirectMailSent {

		@Captor
		private ArgumentCaptor<VorgangRedirectedEvent> eventCaptor;

		private Forwarding forwarding = ForwardingTestFactory.create();

		private static final String BODY = "FILLED BODY";

		@BeforeEach
		void mockVorgang() {
			ReflectionTestUtils.setField(service, "replyTo", Optional.empty());
			when(vorgangService.getById(any())).thenReturn(VorgangTestFactory.create());

			doReturn(BODY).when(service).fillConfirmationTemplate(any());
		}

		@Test
		void shouldPublishEvent() {
			service.onRedirectMailSent(forwarding);

			verify(publisher).publishEvent(eventCaptor.capture());
			assertThat(eventCaptor.getValue()).isInstanceOf(VorgangRedirectedEvent.class)
					.extracting(VorgangRedirectedEvent::getSource).isSameAs(forwarding);
		}

		@Test
		void shouldSendConfirmationMail() {
			service.onRedirectMailSent(forwarding);

			verify(service).sendRedirectConfirmationMail(forwarding);
		}
	}

	@Nested
	class TestSendRedirectConfirmationMail {

		private Vorgang vorgang = VorgangTestFactory.create();
		private Forwarding forwarding = ForwardingTestFactory.create();

		@Nested
		class TestCreateConfirmationMail {
			private static final String MAIL_FROM = "FROM ME WITH L";
			private static final String BODY = "FILLED BODY";
			private final Optional<String> replyTo = Optional.of("TestReplyTo");

			@BeforeEach
			void mockService() {
				ReflectionTestUtils.setField(service, "mailFrom", MAIL_FROM);
				ReflectionTestUtils.setField(service, "replyTo", replyTo);

				doReturn(BODY).when(service).fillConfirmationTemplate(any());
			}

			@Test
			void shouldSetBody() {
				var result = service.createConfirmationMail(VorgangTestFactory.create());

				assertThat(result.getBody()).isEqualTo(BODY);
			}

			@Test
			void shouldSetFromAdresses() {
				var result = service.createConfirmationMail(VorgangTestFactory.create());

				assertThat(result.getFromAddress()).isEqualTo(MAIL_FROM);
			}

			@Test
			void shouldSetSubject() {
				var result = service.createConfirmationMail(VorgangTestFactory.create());

				assertThat(result.getSubject())
						.isEqualTo("Ihr Antrag auf " + EingangHeaderTestFactory.FORM_NAME + " - Az: " + EingangHeaderTestFactory.REQUEST_ID);
			}

			@Nested
			class TestConfirmationMailReplyTo {

				@Test
				void shouldCreate() {
					ReflectionTestUtils.setField(service, "replyTo", replyTo);

					var result = service.createConfirmationMail(VorgangTestFactory.create());

					assertThat(result.getReplyTo()).isEqualTo(replyTo.get());
				}

				@DisplayName("Create email while 'replyTo' is not set in the properties")
				@Test
				void shouldCreateWithoutReplyTo() {
					ReflectionTestUtils.setField(service, "replyTo", Optional.empty());

					var result = service.createConfirmationMail(VorgangTestFactory.create());

					assertThat(result.getReplyTo()).isNull();
				}
			}
		}

		@Test
		void shouldCallVorgangService() {
			doReturn(false).when(service).antragstellerEMailExists(any());

			service.sendRedirectConfirmationMail(forwarding);

			verify(vorgangService).getById(forwarding.getVorgangId());
		}

		@DisplayName("should NOT send mail if antragsteller email does not exists")
		@Nested
		class TestNotSendMail {

			@BeforeEach
			void mockService() {
				Vorgang vorgang = VorgangTestFactory.createBuilder().clearEingangs()
						.eingang(EingangTestFactory.createBuilder()
								.antragsteller(AntragstellerTestFactory.createBuilder().email("").build())
								.build())
						.build();

				when(vorgangService.getById(any())).thenReturn(vorgang);
			}

			@Test
			void shouldNotSendMail() {
				service.sendRedirectConfirmationMail(forwarding);

				verify(mailService, times(0)).sendMail(any());
			}
		}

		@Nested
		class TestSendMail {

			private static final String MAIL_FROM = "FROM ME WITH L";
			private static final String BODY = "FILLED BODY";

			@BeforeEach
			void mock() {
				ReflectionTestUtils.setField(service, "mailFrom", MAIL_FROM);
				ReflectionTestUtils.setField(service, "replyTo", Optional.empty());

				when(vorgangService.getById(any())).thenReturn(vorgang);
				doReturn(BODY).when(service).fillConfirmationTemplate(any());
			}

			@DisplayName("should send mail if antragsteller email exists")
			@Test
			void shouldSendMail() {
				service.sendRedirectConfirmationMail(forwarding);

				verify(mailService).sendMail(any());
			}
		}
	}

	@Nested
	class TestCreateConfirmationMailRequest {

		private static final String MAIL_FROM = "FROM ME WITH L";
		private static final String BODY = "FILLED BODY";
		private Vorgang vorgang = VorgangTestFactory.create();

		@BeforeEach
		void mock() {
			ReflectionTestUtils.setField(service, "mailFrom", MAIL_FROM);
			ReflectionTestUtils.setField(service, "replyTo", Optional.empty());

			doReturn(BODY).when(service).fillConfirmationTemplate(any());
		}

		@Test
		void shouldSetFromAddress() {
			var mailSendRequest = service.createConfirmationMail(vorgang);

			assertThat(mailSendRequest.getFromAddress()).isEqualTo(MAIL_FROM);
		}

		@Test
		void shouldSetToAddress() {
			var mailSendRequest = service.createConfirmationMail(vorgang);

			var toEmails = mailSendRequest.getToAddresses().stream().map(MailRecipient::email).toList();
			assertThat(toEmails).containsExactly(AntragstellerTestFactory.EMAIL);
		}

		@Test
		void shouldSetSubject() {
			String expectedSubject = String.format("Ihr Antrag auf %s - Az: %s", EingangHeaderTestFactory.FORM_NAME,
					EingangHeaderTestFactory.REQUEST_ID);

			var mailSendRequest = service.createConfirmationMail(vorgang);

			assertThat(mailSendRequest.getSubject()).isEqualTo(expectedSubject);
		}

		@Test
		void shouldSetEALogoAsAttachment() {
			var mailSendRequest = service.createConfirmationMail(vorgang);

			assertThat(mailSendRequest.getAttachments()).hasSize(2);
			MailAttachment eaLogoAttachment = (MailAttachment) mailSendRequest.getAttachments().toArray()[0];
			assertThat(eaLogoAttachment.getContentId()).isEqualTo(ForwardingService.CID_LOGO_EA);
		}

		@Test
		void shouldSetItvshLogoAsAttachment() {
			var mailSendRequest = service.createConfirmationMail(vorgang);

			assertThat(mailSendRequest.getAttachments()).hasSize(2);
			MailAttachment itvshLogoAttachment = (MailAttachment) mailSendRequest.getAttachments().toArray()[1];
			assertThat(itvshLogoAttachment.getContentId()).isEqualTo(ForwardingService.CID_LOGO_ITVSH);
		}
	}

	@Nested
	class TestBuildConfirmationValues {

		private Vorgang vorgang = VorgangTestFactory.create();

		@Nested
		class TestAnrede {

			@Test
			void shouldFillByHerr() {
				Vorgang vorgangWithAnrede = createVorgangWithAntragstellerAnrede("Herr");

				var map = service.buildValues(vorgangWithAnrede);

				assertThat(map.get(ForwardingService.FIELD_ANTRAGSTELLER_ANREDE)).isNotNull().isEqualTo("Herr");
			}

			@Test
			void shouldFillByFrau() {
				Vorgang vorgangWithAnrede = createVorgangWithAntragstellerAnrede("Frau");

				var map = service.buildValues(vorgangWithAnrede);

				assertThat(map.get(ForwardingService.FIELD_ANTRAGSTELLER_ANREDE)).isNotNull().isEqualTo("Frau");
			}
		}

		@Test
		void shouldFillAntragstellerName() {
			var map = service.buildValues(vorgang);

			assertThat(map.get(ForwardingService.FIELD_ANTRAGSTELLER_NAME)).isNotNull().isEqualTo(AntragstellerTestFactory.NACHNAME);
		}

		@Test
		void shouldFillAfmId() {
			var map = service.buildValues(vorgang);

			assertThat(map.get(ForwardingService.FIELD_EINGANGS_KENNZEICHEN)).isNotNull().isEqualTo(EingangHeaderTestFactory.REQUEST_ID);
		}

		@Test
		void shouldContainsEALogo() {
			var map = service.buildValues(vorgang);

			assertThat(map.get(ForwardingService.FIELD_CID_LOGO_EA)).isNotNull().isEqualTo(ForwardingService.CID_LOGO_EA);
		}

		@Test
		void shouldContainsItvshLogo() {
			var map = service.buildValues(vorgang);

			assertThat(map.get(ForwardingService.FIELD_CID_LOGO_ITVSH)).isNotNull().isEqualTo(ForwardingService.CID_LOGO_ITVSH);
		}

		private Vorgang createVorgangWithAntragstellerAnrede(String anrede) {
			return VorgangTestFactory.createBuilder().clearEingangs()
					.eingang(EingangTestFactory.createBuilder()
							.antragsteller(AntragstellerTestFactory.createBuilder().anrede(anrede).build())
							.build())
					.build();
		}
	}

	@Nested
	class TestMarkAsSuccessfull {

		@Captor // NOSONAR
		private ArgumentCaptor<Map<String, Object>> mapCaptor;

		private Command command = CommandTestFactory.createBuilder().relationId(ForwardingTestFactory.ID).build();

		@Test
		void shouldCAllRepository() {

			service.markAsSuccessfull(command);

			verify(repository).patch(eq(ForwardingTestFactory.ID), mapCaptor.capture());
			assertThat(mapCaptor.getValue()).hasSize(1).containsEntry(Forwarding.FIELD_STATUS, Forwarding.Status.SUCCESSFULL);
		}

		@Test
		void shouldDispatchEvent() {
			service.markAsSuccessfull(command);

			verify(publisher).publishEvent(any(VorgangForwardSuccessfullEvent.class));
		}
	}

	@Nested
	class TestMarkAsFailed {

		@Captor // NOSONAR
		private ArgumentCaptor<Map<String, Object>> mapCaptor;

		private Command command = CommandTestFactory.createBuilder().relationId(ForwardingTestFactory.ID).build();

		@Test
		void shouldCallRepository() {
			service.markAsFailed(command);

			verify(repository).patch(eq(ForwardingTestFactory.ID), mapCaptor.capture());
			assertThat(mapCaptor.getValue()).hasSize(1).containsEntry(Forwarding.FIELD_STATUS, Forwarding.Status.FAILED);
		}

		@Test
		void shouldDispatchEvent() {
			service.markAsFailed(command);

			verify(publisher).publishEvent(any(VorgangForwardFailedEvent.class));
		}
	}

	@Nested
	class TestForwardVorgang {

		private Command command = CommandTestFactory.create();
		private Vorgang vorgang = VorgangTestFactory.create();

		@Test
		void shouldDoRedirect() {
			when(vorgangService.getById(any())).thenReturn(vorgang);
			doNothing().when(service).doRedirect(any(), any());

			service.forwardByCommand(command);

			verify(service).doRedirect(command, vorgang);
		}

		@DisplayName("on version missmatch")
		@Nested
		class VersionMissmatch {

			@BeforeEach
			void initTest() {
				when(vorgangService.getById(any())).thenReturn(vorgang);
			}

			@Test
			void shouldDoNothing() {
				service.forwardByCommand(CommandTestFactory.createBuilder().relationVersion(5L).build());

				verify(service, never()).doRedirect(any(), any());
			}

			@Test
			void shouldMarkAsFailed() {
				var command = CommandTestFactory.createBuilder().relationVersion(5L).build();

				service.forwardByCommand(command);

				verify(service).markAsFailed(command);
			}
		}

		@Nested
		class doRedirect {

			@BeforeEach
			void initTest() {
				doNothing().when(service).redirectVorgang(any(), any());
			}

			@DisplayName("should update vorgang by the given status 'IN_BEARBEITUNG")
			@Test
			void shouldCallService() {
				service.doRedirect(command, vorgang);

				verify(statusService).setStatusToInBearbeitung(CommandTestFactory.RELATION_ID, CommandTestFactory.RELATION_VERSION);
			}

			@Test
			void shouldForwardVorgang() {
				service.doRedirect(command, vorgang);

				verify(service).redirectVorgang(vorgang, command);
			}

			@Test
			void shouldNOTDispatchEvent() {
				service.doRedirect(command, vorgang);

				verifyNoInteractions(publisher);
			}
		}

	}
}